/*****************************************************************************************
 * File name     : INFORECD.C
 * Author        : Han Zhi Qiang
 * Module        : 
 * Purpose       : 
 * Create Date   : October 18, 2001 - 22:06
 * Modify history: 
 *****************************************************************************************
 */

#include <windows.h>
#include "inforecd.h"

/*****************************************************************************************
 * MACRO functions declare
 ****************************************************************************************/

/*****************************************************************************************
 * Identifier constant declare
 ****************************************************************************************/
#define CHARS_ON_ROW                    74
#define INDENT                          4

#define WNDNAME                         "Information Recorder"

/*****************************************************************************************
 * Data structure declare
 ****************************************************************************************/

/*****************************************************************************************
 * Local variable declare
 ****************************************************************************************/
HANDLE            ghInfoRecordWnd = NULL;

/*****************************************************************************************
 * Local functions declare
 ****************************************************************************************/

/****************************************************************************************/
int  SpyWindow( void )
{
  if (!ghInfoRecordWnd)
    ghInfoRecordWnd = FindWindow (WNDNAME, WNDNAME);
  return ghInfoRecordWnd != NULL;
}
/****************************************************************************************/
unsigned char ScanEndHalfHz( const unsigned char *ptr, int len )
{
  unsigned char  c2 = 0;
  int            i;

  for( i = 0; i < len; i++ )
  {
    unsigned char  c = *ptr++;
    if( c2 )
    { c2 = 0;
      continue;
    }

    c2 = ( c > 0xA0 ) ? c : 0;
  }

  if( c2 )
  { c2 = *ptr;
    if( c2 < 0x80 )
      c2 = 0;
  }
  return c2;
}
/****************************************************************************************/
void  x_memmove( char *ptr, const char *buf, int len )
{
  int            i;
  unsigned char *pSrc = (unsigned char *)buf;

  for( i = 0; i < len; i++ )
  {
    if( *pSrc >= ' ' /* && *pSrc < 0x80 */ )
      *ptr++ = *pSrc++;
    else
    { *ptr++ = 0x01;
      pSrc++;
    }
  }
}
/****************************************************************************************/
int  GetNewLineNum( const char *p )
{
  int   n;

  for( n = 0; *p; p++ )
  {
    if( *p == '\n' )
      n++;
  }
  return n;
}
/****************************************************************************************/
void    SpyShowData( const char *title, const char *buf, int len, BOOL bAscii )
{
  COPYDATASTRUCT  cds;
  char           *p_buf, *ptr;
  int             i, iCnt, str_len, delta;

  if (ghInfoRecordWnd == NULL) return;
  
  str_len = title ? strlen( title ) : 0;
  iCnt  = ( ( str_len + CHARS_ON_ROW - 1 ) / CHARS_ON_ROW ) * CHARS_ON_ROW;

  if( str_len )              // =======
    iCnt += CHARS_ON_ROW;    // =======

  if( bAscii )
  {
    iCnt += ( ( len + ( CHARS_ON_ROW - INDENT - 1 ) ) / 
              ( CHARS_ON_ROW - INDENT ) ) * CHARS_ON_ROW;
    iCnt += CHARS_ON_ROW * GetNewLineNum( buf );
    p_buf = (char *)GlobalAlloc( GMEM_FIXED, iCnt + 32 );
    if( ! p_buf )
      return;
  }
  else
  { 
    //[O] 00 01 02 03 04 05 06 07 - 08 09 0A 0B 0C 0D 0E 0F    0123456789ABCDEF 
    iCnt += ( ( len + 15 ) / 16 ) * CHARS_ON_ROW;
    p_buf = (char *)GlobalAlloc( GMEM_FIXED, iCnt + 32 );
    if( ! p_buf )
      return;
  }

  iCnt = 0;
  ptr  = p_buf;
  if( title && str_len > 0 )
  {
    memset( ptr, ' ', CHARS_ON_ROW );  // =======
    ptr  += CHARS_ON_ROW;              // =======
    iCnt += CHARS_ON_ROW;              // =======

    strcpy( ptr, title );
    ptr += str_len;
    delta = str_len % CHARS_ON_ROW;
    
    if( delta )
    { delta = CHARS_ON_ROW - delta;
      memset( ptr, ' ', delta );
      ptr += delta;
    }
    iCnt += str_len + delta;
  }

  if( bAscii )
  {
    while( len > 0 )
    { int   n;
      memset( ptr, ' ', INDENT );
      ptr  += INDENT;
      iCnt += 4;

      for( n = 0; len > 0 && n < ( CHARS_ON_ROW - INDENT ); n++ )
      { if( *buf == '\n' )
        { buf++;
          len--;
          break;
        }
        *ptr++ = *buf++;
        len--;
        iCnt++;
      }
      
      for( ; n < ( CHARS_ON_ROW - INDENT ); n++ )
      { *ptr++ = ' ';
        iCnt++;
      }
      
    }
  }
  else
  {
    unsigned char cLastHalfHz = 0;
    while( len > 0 )
    { const unsigned char  *p_tmp;

      memset( ptr, ' ', INDENT );
      ptr += INDENT;
    
      p_tmp = buf;
      for( i = 0; i < 8; i++ )
      { if( i < len )
          wsprintf( ptr, "%02X ", *p_tmp++ );
        else
          memset( ptr, ' ', 3 );
        ptr += 3;
      }
      
      if( len > 8 )
        *ptr++ = '-';
      else
        *ptr++ = ' ';
      *ptr++ = ' ';

      for( i = 8; i < 16; i++ )
      { if( i < len )
          wsprintf( ptr, "%02X ", *p_tmp++ );
        else
          memset( ptr, ' ', 3 );
        ptr += 3;
      }
      
      memset( ptr, ' ', 3 );
      ptr += 3;

      if( len >= 16 )
      {
        if( cLastHalfHz == 0 )
        { x_memmove( ptr, buf, 16 );
          cLastHalfHz = ScanEndHalfHz( buf, 16 );
        }
        else
        { *ptr = ' ';
          x_memmove( ptr+1, buf+1, 16-1 );
          cLastHalfHz = ScanEndHalfHz( buf+1, 16-1 );
        }
      }
      else
      { x_memmove( ptr, buf, len );
        memset( ptr + len, ' ', (16-len) );
        cLastHalfHz = 0;
      }

      len  -= 16;
      buf  += 16;
      ptr  += 16;
      
      if( cLastHalfHz )
        *ptr++ = cLastHalfHz;
      else
        *ptr++ = ' ';

      iCnt += CHARS_ON_ROW;
    }
  }
  
  cds.cbData = iCnt;
  cds.dwData = ( ('I'<<24) | ('F'<<16) | ('R'<<8) | 'C' );
  cds.lpData = p_buf;
  if( ghInfoRecordWnd )
  { if( ! SendMessage( ghInfoRecordWnd, WM_COPYDATA, 0, (LPARAM)&cds ) )
      ghInfoRecordWnd = NULL;
  }

  GlobalFree( (HGLOBAL)p_buf );
}
/****************************************************************************************/
