#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "bmptool.h"

/*************************************************************************/
/* IP:BmpFile - Point to a string of file name
   OP:Fill the information into a sturcture of BMPINFO
*/
BMPINFO * GetBmpInfo (const char *BmpFile)
{
  BMPINFO *BmpInfo;
  FILE *BitmapFile;
  RGBQUAD Pallette;
  int i;

  BitmapFile = fopen (BmpFile, "rb");
  if (BitmapFile == NULL) return NULL;
  if (fgetc(BitmapFile) != 'B' || fgetc(BitmapFile) != 'M')
  {
    fclose (BitmapFile);
    return NULL;
  }
  BmpInfo = malloc (sizeof (BMPINFO));
  if (BmpInfo == NULL)
  {
    fclose (BitmapFile);
    return NULL;
  }

  BmpInfo->BitmapFile = BitmapFile;
  fseek (BmpInfo->BitmapFile, 0l, SEEK_SET);
  fread (&BmpInfo->FileHeader, sizeof(BITMAPFILEHEADER), 1,
          BmpInfo->BitmapFile);
  fread (&BmpInfo->BmpHeader, sizeof (BITMAPINFOHEADER), 1,
          BmpInfo->BitmapFile);

  if (BmpInfo->BmpHeader.biClrUsed == 0)
  {
    switch (BmpInfo->BmpHeader.biBitCount)
    {
      case  1: BmpInfo->PalletteCount =   2; break;
      case  4: BmpInfo->PalletteCount =  16; break;
      case  8: BmpInfo->PalletteCount = 256; break;
      default: BmpInfo->PalletteCount = 256; break;
    }
  }
  else
    BmpInfo->PalletteCount = BmpInfo->BmpHeader.biClrUsed;

  fseek (BmpInfo->BitmapFile,
         sizeof (BITMAPFILEHEADER)+BmpInfo->BmpHeader.biSize, SEEK_SET);
  for (i=0; i<BmpInfo->PalletteCount; i++)
  {
    fread (&Pallette, sizeof (RGBQUAD), 1, BmpInfo->BitmapFile);
    BmpInfo->Pallette[i] = RGB ( Pallette.rgbRed,
                                 Pallette.rgbGreen,
                                 Pallette.rgbBlue);
  }

  return BmpInfo;
}

/*************************************************************************/
/* IP:BmpInfo - Point to a structure of BMPINFO
   OP:Release all resource
*/
void ReleaseBmpInfo (BMPINFO *BmpInfo)
{
  fclose (BmpInfo->BitmapFile);
  free (BmpInfo);
}

/*************************************************************************/
/* IP:BmpInfo - Point to a structure of BMPINFO
      x - specify the coordinate x
      y - specify the coordinate y
   OP:Return the specified pixel's color
*/
COLORREF GetBitmapColor (BMPINFO *BmpInfo, int x, int y)
{
  int LineBytes;
  COLORREF Value;
  unsigned char color[3];

  if (x<0 || x>=BmpInfo->BmpHeader.biWidth ||
      y<0 || y>=BmpInfo->BmpHeader.biHeight)
    return 0;

  LineBytes = BmpInfo->BmpHeader.biWidth * BmpInfo->BmpHeader.biBitCount;
  LineBytes = LineBytes % 32 ? (LineBytes/32+1) * 32 : LineBytes;
  LineBytes /= 8;

  switch (BmpInfo->BmpHeader.biBitCount)
  {
    case 1:
      fseek (BmpInfo->BitmapFile,
             BmpInfo->FileHeader.bfOffBits +
             (BmpInfo->BmpHeader.biHeight-y-1) * LineBytes + x / 8,
             SEEK_SET);
      if ((fgetc(BmpInfo->BitmapFile)<<(x%8))&0x80)
        Value = BmpInfo->Pallette[1];
      else
        Value = BmpInfo->Pallette[0];
      break;
    case 4:
      fseek (BmpInfo->BitmapFile,
             BmpInfo->FileHeader.bfOffBits +
             (BmpInfo->BmpHeader.biHeight-y-1) * LineBytes + x / 2,
             SEEK_SET);
      if (x % 2)
        Value = BmpInfo->Pallette[fgetc(BmpInfo->BitmapFile)&0xf];
      else
        Value = BmpInfo->Pallette[(fgetc(BmpInfo->BitmapFile)>>4)&0xf];
      break;
    case 8:
      fseek (BmpInfo->BitmapFile,
             BmpInfo->FileHeader.bfOffBits +
             (BmpInfo->BmpHeader.biHeight-y-1) * LineBytes + x,
             SEEK_SET);
      Value = BmpInfo->Pallette[fgetc(BmpInfo->BitmapFile)];
      break;
    case 24:
      fseek (BmpInfo->BitmapFile,
             BmpInfo->FileHeader.bfOffBits +
             (BmpInfo->BmpHeader.biHeight-y-1) * LineBytes + x * 3,
             SEEK_SET);
      fread (color, 3, 1, BmpInfo->BitmapFile);
      Value = RGB (color[2], color[1], color[0]);
      break;
  }

  return Value;
}

/*************************************************************************/
/* IP:BmpInfo - Point to a structure of BMPINFO
   OP:Allocate a row color buffer
*/
COLORREF *GetLineColorBuf (BMPINFO *BmpInfo)
{
  return (COLORREF*) malloc (sizeof(COLORREF)*BmpInfo->BmpHeader.biWidth);
}

/*************************************************************************/
/* IP:BmpInfo - Point to a structure of BMPINFO
      ColorBuf - Point the color buffer
   OP:free color buffer
*/
void FreeLineColorBuf (BMPINFO *BmpInfo, COLORREF *ColorBuf)
{
  free (ColorBuf);
}

/*************************************************************************/
/* IP:BmpInfo - Point to a structure of BMPINFO
      y - specify the row number
      ColorBuf - specify color buffer
   OP:The colors in the color buffer
*/
int GetBitmapLineColor (BMPINFO *BmpInfo, int y, COLORREF *ColorBuf)
{
  int LineBytes;
  unsigned char color[3];
  int i, TempChar;

  if (y<0 || y>=BmpInfo->BmpHeader.biHeight) return 0;

  LineBytes = BmpInfo->BmpHeader.biWidth * BmpInfo->BmpHeader.biBitCount;
  LineBytes = LineBytes % 32 ? (LineBytes/32+1) * 32 : LineBytes;
  LineBytes /= 8;

  switch (BmpInfo->BmpHeader.biBitCount)
  {
    case 1:
      fseek (BmpInfo->BitmapFile,
             BmpInfo->FileHeader.bfOffBits +
             (BmpInfo->BmpHeader.biHeight-y-1) * LineBytes,
             SEEK_SET);
      for (i=0; i<BmpInfo->BmpHeader.biWidth; i++)
      {
        if (i%8 == 0)
          TempChar = fgetc(BmpInfo->BitmapFile);
        ColorBuf[i] = BmpInfo->Pallette[TempChar&(0x80>>(i%8)) ? 1 : 0];
      }
      break;
    case 4:
      fseek (BmpInfo->BitmapFile,
             BmpInfo->FileHeader.bfOffBits +
             (BmpInfo->BmpHeader.biHeight-y-1) * LineBytes,
             SEEK_SET);
      for (i=0; i<BmpInfo->BmpHeader.biWidth; i++)
      {
        if (i % 2)
          ColorBuf[i] = BmpInfo->Pallette[TempChar&0xf];
        else
        {
          TempChar = fgetc(BmpInfo->BitmapFile);
          ColorBuf[i] = BmpInfo->Pallette[(TempChar>>4)&0xf];
        }
      }
      break;
    case 8:
      fseek (BmpInfo->BitmapFile,
             BmpInfo->FileHeader.bfOffBits +
             (BmpInfo->BmpHeader.biHeight-y-1) * LineBytes,
             SEEK_SET);
      for (i=0; i<BmpInfo->BmpHeader.biWidth; i++)
        ColorBuf[i] = BmpInfo->Pallette[fgetc(BmpInfo->BitmapFile)];
      break;
    case 24:
      fseek (BmpInfo->BitmapFile,
             BmpInfo->FileHeader.bfOffBits +
             (BmpInfo->BmpHeader.biHeight-y-1) * LineBytes,
             SEEK_SET);
      for (i=0; i<BmpInfo->BmpHeader.biWidth; i++)
      {
        fread (color, 3, 1, BmpInfo->BitmapFile);
        ColorBuf[i] = RGB (color[2], color[1], color[0]);
      }
      break;
  }

  return BmpInfo->BmpHeader.biWidth;
}
