/*****************************************************************************************
 * File name     : COMM.C
 * Author        : Han Zhi Qiang
 * Module        : 
 * Purpose       : 
 * Create Date   : June 30, 2001 - 23:42
 * Modify history: 
 *****************************************************************************************
 */

#define ALLOCATE_COMM_VARIABLE_MEMORY

#include <windows.h>
#include <stdarg.h>
#include "resource.h"
#include "ui.h"
#include "comm.h"

/*****************************************************************************************
 * MACRO functions declare
 ****************************************************************************************/
#define mfIsEscapeChar(c)               ((c) == PACKETBEGINNER || (c) == '+' ||       \
                                         (c) == ESCAPECHAR || (c) == '3')
#define mfExpandEscapeChar(a, i, c)     {                                             \
                                          if (mfIsEscapeChar (c))                     \
                                          {                                           \
                                            (a)[(i)++] = ESCAPECHAR;                  \
                                            (a)[(i)++] = ~(c);                        \
                                          }                                           \
                                          else                                        \
                                            (a)[(i)++] = (c);                         \
                                        }

/*****************************************************************************************
 * Identifier constant declare
 ****************************************************************************************/
#define PACKETBEGINNER                  0xa5
#define ESCAPECHAR                      0x1b

#define MAXRESENDTIMES                  3

#define IDLERETRYTIME                   10000

/*****************************************************************************************
 * Data structure declare
 ****************************************************************************************/
typedef enum
{
  MSC_DATAMODE,
  MSC_DIRECTCONNECT,
  MSC_NOMODEM,
  MSC_READY,
  MSC_INIT,
  MSC_DIAL,
  MSC_ANSWER,
  MSC_HANGUP,
  
  MSC_BUSY,
  MSC_NODIALTONE,
  MSC_NOCARRIER,
  // Error
  MSC_COMMERR,
} MODEMSTATUS;

/*****************************************************************************************
 * Local variable declare
 ****************************************************************************************/
static HANDLE     ghCommDev = INVALID_HANDLE_VALUE;

static MODEMSTATUS gModemStatus;

// Send packet control variable
static BYTE       gabLastPacket[512];
static BYTE       gbLastPacketSize;
static BYTE       gbLastCommand;
static WORD       gwSendPacketRetryTimes;

static DWORD      gdwSendTickCount;
static DWORD      gdwIdleTickCount;

// Thread control variable
static HANDLE     ghCommunicationThreadHandle = NULL;
static BOOL       gfStopCommunicationThread;

// Task control variable
static PZFZPARAM  gpTaskZFZParam;

/*****************************************************************************************
 * Local functions declare
 ****************************************************************************************/

/****************************************************************************************/
static BOOL OpenCommPort(int PortNo)
{
  char          Buf[100];
  DCB           dcb;
  COMMTIMEOUTS  CommTimeouts;


  if (PortNo < 1 || PortNo > 2) return FALSE;

  wsprintf (Buf, "COM%d", PortNo);

  ghCommDev = CreateFile (Buf, GENERIC_WRITE|GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
  if (ghCommDev != INVALID_HANDLE_VALUE)
  {
    wsprintf (Buf, "COM%d:baud=9600 parity=N data=8 stop=1", PortNo);
    if  (GetCommState (ghCommDev, &dcb))
    {
      //dcb.fDtrControl = RTS_CONTROL_ENABLE;
      //dcb.fRtsControl = RTS_CONTROL_DISABLE;
      if (BuildCommDCB (Buf, &dcb))
      {
        if (SetCommState (ghCommDev, &dcb))
        {
          memset (&CommTimeouts, 0, sizeof(COMMTIMEOUTS));
          CommTimeouts.ReadTotalTimeoutMultiplier = 2;
          if (SetCommTimeouts (ghCommDev, &CommTimeouts))
          {
            return TRUE;
          }
        }
      }
    }
    CloseHandle (ghCommDev);
  }
  return FALSE;
}
/****************************************************************************************/
static int CloseCommPort(void)
{
  HANDLE hHandle = ghCommDev;
  ghCommDev = INVALID_HANDLE_VALUE;
  return CloseHandle (hHandle);
}
/****************************************************************************************/
/****************************************************************************************/
#ifdef LOOPBACKDEBUG
/****************************************************************************************/
/****************************************************************************************/
#define ReadFile(hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped) \
        LoopBackReadFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped)
#define WriteFile(hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped) \
        LoopBackWriteFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped)
/****************************************************************************************/
BYTE    abLoopBackBuf[1024];
DWORD   dwLoopBackBufSize = 0;
/****************************************************************************************/
BOOL LoopBackReadFile(HANDLE hFile, PBYTE pbBuf, DWORD dwBytes, PDWORD pdwActBytes,
                      LPOVERLAPPED lpOverlapped)
{
  DWORD           i;

  if (dwLoopBackBufSize == 0)
    *pdwActBytes = 0;
  else if (dwLoopBackBufSize <= dwBytes)
  {
    memcpy (pbBuf, abLoopBackBuf, dwLoopBackBufSize);
    *pdwActBytes = dwLoopBackBufSize;
    if (ghDebugWnd)
    {
      for (i=0; i<dwLoopBackBufSize; i++)
        gpfunSTATWND_StreamByte (ghDebugWnd, OMPT_DOWNLOAD, pbBuf[i]);
    }
    dwLoopBackBufSize = 0;
  }
  else
  {
    memcpy (pbBuf, abLoopBackBuf, dwBytes);
    *pdwActBytes = dwBytes;
    dwLoopBackBufSize -= dwBytes;
    memcpy (abLoopBackBuf, abLoopBackBuf+dwBytes, dwLoopBackBufSize);
    if (ghDebugWnd)
    {
      for (i=0; i<dwBytes; i++)
        gpfunSTATWND_StreamByte (ghDebugWnd, OMPT_DOWNLOAD, pbBuf[i]);
    }
  }
  return TRUE;
}
/****************************************************************************************/
BOOL LoopBackWriteFile(HANDLE hFile, PBYTE pbBuf, DWORD dwBytes, PDWORD pdwActBytes,
                       LPOVERLAPPED lpOverlapped)
{
  BYTE            bLen, bCheckSum;
  PBYTE           pbTemp;
  DWORD           i;

  if (ghDebugWnd)
  {
    for (i=0; i<dwBytes; i++)
      gpfunSTATWND_StreamByte (ghDebugWnd, OMPT_UPLOAD, pbBuf[i]);
  }

  if ( (gModemStatus == MSC_DATAMODE || gModemStatus == MSC_DIRECTCONNECT) &&
       pbBuf[0] == PACKETBEGINNER
     )
  {
    switch (gbLastCommand)
    {
    case 4:
#ifdef CDMA
      pbTemp = "\x82\xcc\xcd\xc8\x00"
               "\x00\x81\x00\x08\x09\xfd\xfc\x6d\xc0"
               "\x00\x95\x00\x0a\x0b\xf9\xf8\x6d\xc0";
#else
      pbTemp = "\x82\xcc\xcd\xc8\x00"
               "\x00\x81\x01\x08\x09\xfd\xfc\x6d\xc0"
               "\x00\x95\x01\x0a\x0b\xf9\xf8\x6d\xc0";
#endif
      bLen = 20;
      break;
    case 5:
      pbTemp = "\x83\x02\x00\x1c\x14\x00\x1c\x3c";
      bLen = 8;
      break;
    case 9:
      pbTemp = "\x84""13600417512";
      bLen = 12;
      break;
    default:
      pbTemp = "\x80";
      bLen = 1;
      break;
    }
    bCheckSum = 0;
    abLoopBackBuf[dwLoopBackBufSize++] = PACKETBEGINNER;
    mfExpandEscapeChar (abLoopBackBuf, dwLoopBackBufSize, bLen);
    for (i=0; i<bLen; i++)
    {
      bCheckSum += pbTemp[i];
      mfExpandEscapeChar (abLoopBackBuf, dwLoopBackBufSize, pbTemp[i]);
    }
    mfExpandEscapeChar (abLoopBackBuf, dwLoopBackBufSize, bCheckSum);
  }
  else if (pbBuf[0] == 'A' && pbBuf[1] == 'T')
  {
    if (gModemStatus == MSC_DIAL || gModemStatus == MSC_ANSWER)
    {
      memcpy (abLoopBackBuf+dwLoopBackBufSize, "\r\nCONNECT 9600\r\n", 16);
      dwLoopBackBufSize += 16;
    }
    else 
    {
      memcpy (abLoopBackBuf+dwLoopBackBufSize, "\r\nOK\r\n", 6);
      dwLoopBackBufSize += 6;
    }
  }
  Sleep (1000);
  *pdwActBytes = dwBytes;
  return TRUE;
}
/****************************************************************************************/
#else
/****************************************************************************************/
BOOL DebugReadFile(HANDLE hFile, PBYTE pbBuf, DWORD dwBytes, PDWORD pdwActBytes,
                      LPOVERLAPPED lpOverlapped)
{
  BOOL            fResult;
  DWORD           i;
  
  fResult = ReadFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped);
  if (ghDebugWnd)
  {
    for (i=0; i<*pdwActBytes; i++)
      gpfunSTATWND_StreamByte (ghDebugWnd, OMPT_DOWNLOAD, pbBuf[i]);
  }
  return fResult;
}
/****************************************************************************************/
BOOL DebugWriteFile(HANDLE hFile, PBYTE pbBuf, DWORD dwBytes, PDWORD pdwActBytes,
                        LPOVERLAPPED lpOverlapped)
{
  BOOL            fResult;
  DWORD           i;
  
  fResult = WriteFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped);
  if (ghDebugWnd)
  {
    for (i=0; i<*pdwActBytes; i++)
      gpfunSTATWND_StreamByte (ghDebugWnd, OMPT_UPLOAD, pbBuf[i]);
  }
  return fResult;
}
/****************************************************************************************/
#define ReadFile(hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped) \
        DebugReadFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped)
#define WriteFile(hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped) \
        DebugWriteFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped)
/****************************************************************************************/
#endif
/****************************************************************************************/
/****************************************************************************************/
static BOOL SendPacket(const BYTE *pcbPacket, BYTE bPacketSize)
{
  BYTE            abBuf[512];
  BYTE            bCheckSum;
  BYTE            bPos = 0, i;
  DWORD           dwActualLen;

  bCheckSum = 0;
  abBuf[bPos++] = PACKETBEGINNER;
  mfExpandEscapeChar (abBuf, bPos, bPacketSize);
  for (i=0; i<bPacketSize; i++)
  {
    bCheckSum += pcbPacket[i];
    mfExpandEscapeChar (abBuf, bPos, pcbPacket[i]);
  }
  mfExpandEscapeChar (abBuf, bPos, bCheckSum);
  
  memcpy (gabLastPacket, abBuf, bPos);
  gbLastPacketSize = bPos;
  gbLastCommand = pcbPacket[0];
  
  gwSendPacketRetryTimes = 0;
  gdwSendTickCount = GetTickCount ();

  return WriteFile (ghCommDev, abBuf, bPos, &dwActualLen, NULL) && dwActualLen == bPos;
}
/****************************************************************************************/
static BOOL ResendPacket(void)
{
  DWORD           dwActualLen;

  gdwSendTickCount = GetTickCount ();

  return WriteFile (ghCommDev, gabLastPacket, gbLastPacketSize, &dwActualLen, NULL) &&
         dwActualLen == gbLastPacketSize;
}
/****************************************************************************************/
/****************************************************************************************/
static BOOL SendModemString(MODEMSTATUS ModemStatus, const CHAR *pcszCommand, ...)
{
  CHAR            aszBuf[512];
  DWORD           dwActualLen;
  BYTE            bCmdLen;
  va_list         Arguments;

  va_start (Arguments, pcszCommand);
  wvsprintf (aszBuf, pcszCommand, Arguments);
  va_end (Arguments);
  strcat (aszBuf, "\r");
  bCmdLen = strlen (aszBuf);

  memcpy (gabLastPacket, aszBuf, bCmdLen);
  gbLastPacketSize = bCmdLen;
  gwSendPacketRetryTimes = 0;
  gdwSendTickCount = GetTickCount ();

  gModemStatus = ModemStatus;
  return WriteFile (ghCommDev, aszBuf, bCmdLen, &dwActualLen, NULL) &&
         bCmdLen == dwActualLen;
}
/****************************************************************************************/
static BOOL SetModemToCmdMode(void)
{
  DWORD           dwActualLen;
  DWORD           dwLen;
  Sleep (1000);
  gModemStatus = MSC_HANGUP;
  dwLen = strlen (gModemParam.aszToCMD);
  if (!WriteFile (ghCommDev, gModemParam.aszToCMD, dwLen, &dwActualLen, NULL) ||
      dwLen<dwActualLen)
    return FALSE;
  Sleep (1000);
  return TRUE;
}
/****************************************************************************************/
static BOOL Dial(PCSTR pcszPhoneNumber)
{
  CHAR            aszBuf[MAXATCMDLEN+3];

  if (gModemStatus == MSC_DATAMODE) return TRUE;
  strcpy (aszBuf, gModemParam.aszDial);
  strcat (aszBuf, "%s");

  SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_DIAL, (LPARAM)pcszPhoneNumber);
  return SendModemString (MSC_DIAL, aszBuf, pcszPhoneNumber);
}
/****************************************************************************************/
static BOOL Hangup(void)
{
  SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_HANGUP, 0);
  if (gModemStatus == MSC_DATAMODE)
    if (!SetModemToCmdMode ()) return FALSE;
  return SendModemString (MSC_HANGUP, gModemParam.aszHangup);
}
/****************************************************************************************/
/****************************************************************************************/
BOOL SendZFZCommand(void)
{
  BYTE            aszCMD[256];
  BYTE            bCMDSize;
  WORD            i;
  TASKID          TaskID;

  do
    TaskID = gpTaskZFZParam->TaskID;
  while (TaskID == TASKID_SYNCHRONIZE);
  SendMessage (ghMainWnd, WM_TASKSTATUS, (WPARAM)gpTaskZFZParam->aszName, (LPARAM)TaskID);

  if (TaskID & TASKID_SETPHONE1)
  {
    aszCMD[0] = 0x01;
    strcpy (aszCMD+1, gaszCentrePhone);
    bCMDSize = strlen (gaszCentrePhone) + 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_SETPHONE1;
  }
  else if (TaskID & TASKID_SETPHONE2)
  {
    aszCMD[0] = 0x02;
    strcpy (aszCMD+1, gpTaskZFZParam->aszPhone);
    bCMDSize = strlen (gpTaskZFZParam->aszPhone) + 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_SETPHONE2;
  }
  else if (TaskID & TASKID_GETPHONE2)
  {
    aszCMD[0] = 0x09;
    bCMDSize = 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_GETPHONE2;
  }
  else if (TaskID & TASKID_SETPARAM)
  {
    aszCMD[0] = 0x03;
    for (i=0; i<gpTaskZFZParam->bChannels; i++)
    {
      aszCMD[i*6+1] = gpTaskZFZParam->ChannelData[i].wFrequency % 0xff;
      aszCMD[i*6+2] = gpTaskZFZParam->ChannelData[i].wFrequency / 0xff;
      aszCMD[i*6+3] = gpTaskZFZParam->ChannelData[i].bUpGain;
      aszCMD[i*6+4] = gpTaskZFZParam->ChannelData[i].bDownGain;
      aszCMD[i*6+5] = gpTaskZFZParam->ChannelData[i].bUpPower;
      aszCMD[i*6+6] = gpTaskZFZParam->ChannelData[i].bDownPower;
    }
    bCMDSize = i*6+1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_SETPARAM;
  }
  else if (TaskID & TASKID_GETPROPERTIES)
  {
    aszCMD[0] = 0x05;
    bCMDSize = 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_GETPROPERTIES;
  }
  else if (TaskID & TASKID_ALARM)
  {
    if (gpTaskZFZParam->fDisableAlarm)
      aszCMD[0] = 0x07;
    else
      aszCMD[0] = 0x08;
    bCMDSize = 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_ALARM;
  }
  else if (TaskID & TASKID_TURNON)
  {
    aszCMD[0] = 0x0B;
    bCMDSize = 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_TURNON;
  }
  else if (TaskID & TASKID_TURNOFF)
  {
    aszCMD[0] = 0x0A;
    bCMDSize = 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_TURNOFF;
  }
  else if (TaskID & TASKID_GETSTATUS)
  {
    aszCMD[0] = 0x04;
    bCMDSize = 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_GETSTATUS;
  }
  else if (TaskID & TASKID_LOCK)
  {
    aszCMD[0] = 0xfe;
    bCMDSize = 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_LOCK;
  }
  else if (TaskID & TASKID_UNLOCK)
  {
    aszCMD[0] = 0xff;
    bCMDSize = 1;
    gpTaskZFZParam->ExecutedTaskID |= TASKID_UNLOCK;
  }
  else if (gModemStatus == MSC_DIRECTCONNECT)
  {
    if (gpTaskZFZParam->fPopupStatusDialog)
    {
      SendMessage (ghMainWnd, WM_AUTOPOPUPSTATUSBOX, 0, (LPARAM)gpTaskZFZParam);
      gpTaskZFZParam->fPopupStatusDialog = FALSE;
    }
    gdwSendTickCount = 0;
    gdwIdleTickCount = GetTickCount ();
    return TRUE;
  }
  else
  {
    aszCMD[0] = 0x06;
    bCMDSize = 1;
  }

  return SendPacket (aszCMD, bCMDSize);
}
/****************************************************************************************/
BOOL RecvPacket(const BYTE *pcbPacket, WORD wPacketSize)
{
  TASKID          TaskID;
  BOOLEAN         fResult;
  WORD            i;
  SYSTEMTIME      LocalTime;
  CHAR            aszPhone[ZFZPHONELEN+1];

  switch (pcbPacket[0])
  {
  case 0x80:
    do
      TaskID=InterlockedExchange (&gpTaskZFZParam->TaskID, TASKID_SYNCHRONIZE);
    while (TaskID == TASKID_SYNCHRONIZE);
    switch (gbLastCommand)
    {
    case 1:    TaskID &= ~TASKID_SETPHONE1; break;
    case 2:    TaskID &= ~TASKID_SETPHONE2; break;
    case 3:    TaskID &= ~TASKID_SETPARAM;  Sleep (2000); break;
    case 6:    TaskID &= ~TASKID_LOCKNODE;  break;
    case 7:
    case 8:    TaskID &= ~TASKID_ALARM;     break;
    case 0x0a:
      TaskID &= ~TASKID_TURNOFF;
      SendMessage (ghMainWnd, WM_TURNOFF, 0, 0);
      break;
    case 0x0b:
      TaskID &= ~TASKID_TURNON;
      SendMessage (ghMainWnd, WM_TURNON, 0, 0);
      break;
    case 0xfe: TaskID &= ~TASKID_LOCK;      break;
    case 0xff: TaskID &= ~TASKID_UNLOCK;    break;
    }
    InterlockedExchange (&gpTaskZFZParam->TaskID, TaskID);

    if (gbLastCommand == 6)
    {
      if (gpTaskZFZParam->fPopupStatusDialog)
      {
        SendMessage (ghMainWnd, WM_AUTOPOPUPSTATUSBOX, 0, (LPARAM)gpTaskZFZParam);
        gpTaskZFZParam->fPopupStatusDialog = FALSE;
      }
      SendMessage (ghMainWnd, WM_UPDATETASKNODE, 0, (LPARAM)gpTaskZFZParam);
      Sleep (4);
      fResult = Hangup ();
    }
    else
      fResult = SendZFZCommand ();
    break;
  case 0x81:
    fResult = ResendPacket ();
    break;
  case 0x82:
    fResult = FALSE;
    gpTaskZFZParam->bVoltage     = pcbPacket[1];
    gpTaskZFZParam->bCurrent     = pcbPacket[2];
    gpTaskZFZParam->bTemperature = pcbPacket[3];
    gpTaskZFZParam->bStatusFlag  = pcbPacket[4];
    if (gpTaskZFZParam->bStatusFlag & 7) fResult = TRUE;
/*
    if (gpTaskZFZParam->bVoltage<194 || gpTaskZFZParam->bVoltage>214) fResult = TRUE;
    if (gpTaskZFZParam->bTemperature>204) fResult = TRUE;
*/
    for (i=0; i<gpTaskZFZParam->bChannels; i++)
    {
      gpTaskZFZParam->ChannelData[i].bErrorCode       = pcbPacket[i*9+ 5];
      if (gpTaskZFZParam->ChannelData[i].bErrorCode & 0x30) fResult = TRUE;
      gpTaskZFZParam->ChannelData[i].wFrequency       = pcbPacket[i*9+ 6] +
                                                        pcbPacket[i*9+ 7] * 256;
      gpTaskZFZParam->ChannelData[i].bUpGain          = pcbPacket[i*9+ 8];
      gpTaskZFZParam->ChannelData[i].bDownGain        = pcbPacket[i*9+ 9];
      gpTaskZFZParam->ChannelData[i].bUpPower         = pcbPacket[i*9+10];
      gpTaskZFZParam->ChannelData[i].bDownPower       = pcbPacket[i*9+11];
      gpTaskZFZParam->ChannelData[i].bActualUpPower   = pcbPacket[i*9+12];
      gpTaskZFZParam->ChannelData[i].bActualDownPower = pcbPacket[i*9+13];
    }
    GetLocalTime (&LocalTime);
    gpTaskZFZParam->LastSamplingDate.wYear   = LocalTime.wYear;
    gpTaskZFZParam->LastSamplingDate.bMonth  = (BYTE)LocalTime.wMonth;
    gpTaskZFZParam->LastSamplingDate.bDay    = (BYTE)LocalTime.wDay;
    gpTaskZFZParam->LastSamplingDate.bHour   = (BYTE)LocalTime.wHour;
    gpTaskZFZParam->LastSamplingDate.bMinute = (BYTE)LocalTime.wMinute;
    if (fResult)
    {
      gpTaskZFZParam->LastAlarmDate = gpTaskZFZParam->LastSamplingDate;
      gpTaskZFZParam->ErrorCode = ERRORCODE_ERRORS;
    }
    else
      gpTaskZFZParam->ErrorCode = ERRORCODE_NONE;
    do
      TaskID=InterlockedExchange (&gpTaskZFZParam->TaskID, TASKID_SYNCHRONIZE);
    while (TaskID == TASKID_SYNCHRONIZE);
    TaskID &= ~TASKID_GETSTATUS;
    InterlockedExchange (&gpTaskZFZParam->TaskID, TaskID);
    fResult = SendZFZCommand ();
    break;
  case 0x83:
    if (gModemStatus == MSC_DIRECTCONNECT)
      gpTaskZFZParam = (PZFZPARAM)SendMessage (ghMainWnd, WM_DIRECTCONNECT, pcbPacket[1], 0);
    else if (gpTaskZFZParam->bChannels != pcbPacket[1])
      gpTaskZFZParam = (PZFZPARAM)SendMessage (ghMainWnd, WM_REMALLOCNODE, pcbPacket[1], 0);
    gpTaskZFZParam->bChannels      = pcbPacket[1];
    gpTaskZFZParam->bRatedVoltage  = pcbPacket[3];
    gpTaskZFZParam->bRatedPower1   = pcbPacket[4];
    gpTaskZFZParam->bRatedPower2   = pcbPacket[5];
    gpTaskZFZParam->bRatedCurrent  = pcbPacket[6];
    gpTaskZFZParam->bRatedTempture = pcbPacket[7];
    gpTaskZFZParam->bRatedGain     = pcbPacket[8];
    do
      TaskID=InterlockedExchange (&gpTaskZFZParam->TaskID, TASKID_SYNCHRONIZE);
    while (TaskID == TASKID_SYNCHRONIZE);
    TaskID &= ~TASKID_GETPROPERTIES;
    InterlockedExchange (&gpTaskZFZParam->TaskID, TaskID);
    fResult = SendZFZCommand ();
    break;
  case 0x84:
    for (i=1; i<wPacketSize; i++) aszPhone[i-1] = pcbPacket[i];
    aszPhone[i-1] = '\0';
    gpTaskZFZParam = (PZFZPARAM)SendMessage (ghMainWnd, WM_SEARCHTASKNODE,
                                             0, (LPARAM)aszPhone);
    if (gpTaskZFZParam)
    {
      gpTaskZFZParam->ExecutedTaskID = 0;
      fResult = SendZFZCommand ();
    }
    else
      fResult = Hangup ();
    break;
  default:
    fResult = TRUE;
    break;
  }
  return fResult;
}
/****************************************************************************************/
BOOL RecvModemResponse(PCSTR pszResponse)
{
  static WORD     wRingTimes;
  const struct
  {
    PCSTR         pcszResponse;
    WORD          wResponseSize;
  } aModemResponse[] =
  {
     { "OK",            2 }
    ,{ "ERROR",         5 }
    ,{ "CONNECT",       7 }
    ,{ "BUSY",          4 }
    ,{ "NO DIAL TONE", 12 }
    ,{ "NO DIALTONE",  11 }
    ,{ "NO CARRIER",   10 }
    ,{ "RING",          4 }
    ,{ NULL,            0 }
  };
  WORD            i;
  BOOL            fResult = TRUE;

  for (i=0; aModemResponse[i].pcszResponse; i++)
    if (memcmp (pszResponse, aModemResponse[i].pcszResponse,
                aModemResponse[i].wResponseSize) == 0)
      break;
  if (aModemResponse[i].pcszResponse == NULL) return fResult;

  if (i != 7) wRingTimes = 0;
  switch (i)
  {
  case 0: // OK
    SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_READY, 0);
    if (gModemStatus == MSC_INIT || gModemStatus == MSC_DIAL ||
        gModemStatus == MSC_HANGUP)
      gdwIdleTickCount = GetTickCount () - IDLERETRYTIME + 500;
    gModemStatus = MSC_READY;
    break;
  case 1: // ERROR
    // ???
    break;
  case 2: // CONNECT
    SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_DATAMODE, 0);
    fResult = (gModemStatus == MSC_DIAL);
    gModemStatus = MSC_DATAMODE;
    if (fResult)
    {
      gpTaskZFZParam->ExecutedTaskID = 0;
      fResult = SendZFZCommand ();
    }
    else
      fResult = SendPacket ("\x09", 1);
    break;
  case 3: // BUSY
    // ???
    SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_BUSY, 0);
    gModemStatus = MSC_READY;
    break;
  case 4: // NO DIAL TONE
  case 5: // NO DIALTONE
    // ???
    SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_NODIALTONE, 0);
    gModemStatus = MSC_READY;
    break;
  case 6: // NO CARRIER
    // ???
    SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_NOCARRIER, 0);
    gModemStatus = MSC_READY;
    break;
  case 7: // RING
    wRingTimes ++;
    if (wRingTimes == 2)
    {
      SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_ANSWER, 0);
      fResult = SendModemString (MSC_ANSWER, gModemParam.aszAnswer);
    }
    break;
  }

  return fResult;
}
/****************************************************************************************/
/****************************************************************************************/
static DWORD WINAPI CommunicationThread(LPVOID lpParameter)
{
  const BYTE aszNoCarrier[] = "\r\nNO CARRIER\r\n";
  const BYTE aszAT___[] = "AT___\r";
  WORD            wNoCarrierPos = 0;
  WORD            wAT___Pos = 0;
  BYTE            abBuf[512], bData, bCheckSum;
  WORD            wBufPos, wPacketSize;
  DWORD           dwActualBytes;
  BOOL            fEscapeCharMode;
  DWORD           dwTickDuration;
  DWORD           dwConnectIdle;
#ifdef LOOPBACKDEBUG
  DWORD           dwTickQuery = GetTickCount () - 10000;
#endif

  wBufPos = 0;

  while (1)
  {
#ifdef LOOPBACKDEBUG
    if (gfEnableAT___ && GetTickCount () - dwTickQuery > 10000)
    {
      dwTickQuery = GetTickCount ();
      memcpy (abLoopBackBuf+dwLoopBackBufSize, aszAT___, sizeof aszAT___-1);
      dwLoopBackBufSize += sizeof aszAT___-1;
    }
    if (gfPutRing)
    {
      memcpy (abLoopBackBuf+dwLoopBackBufSize, "\r\nRING\r\n", 8);
      dwLoopBackBufSize += 8;
      gfPutRing = FALSE;
    }
#endif
    if ((gModemStatus == MSC_READY || gModemStatus == MSC_NOMODEM ||
         gModemStatus == MSC_DIRECTCONNECT) &&
        gfStopCommunicationThread)
      goto QuitThread;

    if (gModemStatus==MSC_DIRECTCONNECT && GetTickCount ()-dwConnectIdle>20000)
    {
      SendMessage (ghMainWnd, WM_QUITDIRECTCONNECT, MSC_INIT, 0);
      SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_INIT, 0);
      if (!SendModemString (MSC_INIT, gModemParam.aszInit)) break;
    }
    if (!ReadFile (ghCommDev, &bData, 1, &dwActualBytes, NULL)) break;
    if (dwActualBytes == 1)
    {
      // There are data ready.

      if (aszAT___[wAT___Pos] == bData)
      {
        wAT___Pos ++;
        if (!aszAT___[wAT___Pos])
        {
          if (!WriteFile (ghCommDev, "9\r", 2, &dwActualBytes, NULL)) break;
          dwConnectIdle = GetTickCount ();
          if (gModemStatus != MSC_DIRECTCONNECT)
          {
            gModemStatus = MSC_DIRECTCONNECT;
            if (!SendPacket ("\x05", 1)) break;
          }
        }
      }
      else if (aszAT___[0] == bData)
        wAT___Pos = 1;
      else
        wAT___Pos = 0;

      if (gModemStatus == MSC_DATAMODE || gModemStatus == MSC_DIRECTCONNECT)
      {
        // Modem data mode
       
        // Detect the remote hangup
        if (aszNoCarrier[wNoCarrierPos] == bData)
        {
          wNoCarrierPos ++;
          if (!aszNoCarrier[wNoCarrierPos])
          {
            if (!Hangup ()) break; //??? Remote hangup
          }
        }
        else if (aszNoCarrier[0] == bData)
          wNoCarrierPos = 1;
        else
          wNoCarrierPos = 0;
        
        // Detect escape sequence
        if (bData == PACKETBEGINNER)
        {
          wPacketSize = 0;
          fEscapeCharMode = FALSE;
        }
        else
        {
          if (fEscapeCharMode)
          {
            bData = ~bData;
            fEscapeCharMode = FALSE;
          }
          else if (bData == ESCAPECHAR)
          {
            fEscapeCharMode = TRUE;
            continue;
          }
          
          if (wPacketSize == 0)
          {
            wPacketSize = bData;
            wBufPos     = 0;
            bCheckSum   = 0;
          }
          else
          {
            if (wBufPos == wPacketSize)
            {
              if (bCheckSum == bData)
              {
                gdwSendTickCount = 0;
                if (!RecvPacket (abBuf, wBufPos)) break;
              }
              else
                wBufPos ++;             
            }
            else if (wBufPos < wPacketSize)
              bCheckSum += abBuf[wBufPos++] = bData;
          }
        }
      }
      else
      {
        // Modem command mode
        if (bData == '\n')
        {
          if (wBufPos>0 && abBuf[wBufPos-1] == '\r')
            abBuf[wBufPos-1] = 0;
          else
            abBuf[wBufPos] = 0;
          if (abBuf[0] && (abBuf[0] != 'A' || abBuf[1] != 'T'))
          {
            gdwSendTickCount = 0;
            if (!RecvModemResponse (abBuf)) break;
          }
          wBufPos = 0;
        }
        else
          abBuf[wBufPos++] = bData;
      }
    }
    else if (gdwSendTickCount)
    {
      dwTickDuration = GetTickCount ()-gdwSendTickCount;

      if ((gModemStatus==MSC_DIAL || gModemStatus==MSC_ANSWER))
      {
        if (dwTickDuration>60000l)
        {
          // ??? Dialing timeout
          if (!SendModemString (MSC_INIT, gModemParam.aszInit)) break;
        }
      }
      else if (dwTickDuration>1000l && gModemStatus != MSC_DATAMODE ||
               dwTickDuration>4000l && gModemStatus == MSC_DATAMODE)
      {
        gwSendPacketRetryTimes ++;
        if (gwSendPacketRetryTimes >= MAXRESENDTIMES)
        {
          if (gModemStatus==MSC_DATAMODE)
          {
            // ???
            if (!Hangup ()) break;
          }
          else
          {
            SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_NOMODEM, 0);
            gModemStatus = MSC_NOMODEM;
            gdwIdleTickCount = GetTickCount ();
          }
          gdwSendTickCount = 0;
        }
        else
        {
          if (!ResendPacket ()) break;
        }
      }
    }
    else if (gdwIdleTickCount)
    {
      if (gfSkipIdleTickCount || GetTickCount ()-gdwIdleTickCount > IDLERETRYTIME)
      {
        if (gModemStatus == MSC_NOMODEM)
        {
          SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_INIT, 0);
          if (!SendModemString (MSC_INIT, gModemParam.aszInit)) break;
        }
        else
        {
          gpTaskZFZParam = (PZFZPARAM)SendMessage (ghMainWnd, WM_GETNEXTTASKNODE, 0, 0);
          if (gpTaskZFZParam)
          {
            if (gModemStatus == MSC_DIRECTCONNECT)
            {
              if (!SendZFZCommand ()) break;
            }
            else
            {
              if (!Dial (gpTaskZFZParam->aszPhone)) break;
            }
            gdwIdleTickCount = 0;
          }
          else
            gdwIdleTickCount = GetTickCount ();
        }
        gfSkipIdleTickCount = FALSE;
      }
    }
  }
  SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_COMMERR, 0);
QuitThread:
  CloseCommPort ();
  return 0;
}
/****************************************************************************************/
/****************************************************************************************/
STEC StartCommunicationThread(WORD wPort)
{
  DWORD           dwThreadID;
  if (ghCommunicationThreadHandle) return STEC_ALREADYSTART;
  if (!OpenCommPort (wPort)) return STEC_OPENCOMMPORT;

  SendMessage (ghMainWnd, WM_MODEMSTATUS, MSC_INIT, 0);
  if (!SendModemString (MSC_INIT, gModemParam.aszInit))
  {
    CloseCommPort ();
    return STEC_COMMPROTERR;
  }
  gfSkipIdleTickCount = FALSE;
  gfStopCommunicationThread = FALSE;
#ifdef LOOPBACKDEBUG
  gfEnableAT___ = FALSE;
#endif
  ghCommunicationThreadHandle = CreateThread (NULL, 0, CommunicationThread,
                                              NULL, 0, &dwThreadID);
  if (ghCommunicationThreadHandle)
    return STEC_SUCCESSFUL;
  else
  {
    CloseCommPort ();
    return STEC_STARTTHREAD;
  }
}
/****************************************************************************************/
void StopCommunicationThread(void)
{
  gfStopCommunicationThread = TRUE;
}
/****************************************************************************************/
BOOL IsStopCommunicationThread(void)
{
  DWORD           dwExitCode;
  if (ghCommunicationThreadHandle == NULL) return TRUE;
  GetExitCodeThread (ghCommunicationThreadHandle, &dwExitCode);
  return dwExitCode!=STILL_ACTIVE;
}
/****************************************************************************************/
void ReleaseCommunicationThread(void)
{
  CloseHandle (ghCommunicationThreadHandle);
  ghCommunicationThreadHandle = NULL;
  ghCommDev = INVALID_HANDLE_VALUE;
}
/****************************************************************************************/
void KillCommuicationThread(void)
{
  DWORD           dwTickCount;
  
  StopCommunicationThread ();
  dwTickCount = GetTickCount ();
  while (GetTickCount () - dwTickCount < 5000l)
    if (IsStopCommunicationThread ()) break;
  if (!IsStopCommunicationThread ())
  {
    TerminateThread (ghCommunicationThreadHandle, 0);
    CloseCommPort ();
  }
  ReleaseCommunicationThread ();
}
/****************************************************************************************/
