/*****************************************************************************************
 * File name     : PASSWD.C
 * Author        : Han Zhi Qiang
 * Module        : 
 * Purpose       : 
 * Create Date   : November 9, 2001 - 20:58
 * Modify history: 
 *****************************************************************************************
 */

#include <windows.h>
#include "ui.h"
#include "passwd.h"
#include "resource.h"

/*****************************************************************************************
 * MACRO functions declare
 ****************************************************************************************/
#define REGKEYYBDREPEATOR               "SOFTWARE\\YinBoDa\\Repeater"
#define REGKEYUSERDATABASE              REGKEYYBDREPEATOR "\\USER"

/*****************************************************************************************
 * Identifier constant declare
 ****************************************************************************************/
#define MAXUSERINFOS                    50

/*****************************************************************************************
 * Data structure declare
 ****************************************************************************************/

/*****************************************************************************************
 * Local variable declare
 ****************************************************************************************/
static CHAR       gaszString[250];
static CHAR       gaszString2[250];

static const char gaszCharSet[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                                   "0123456789"
                                   "abcdefghijklmnopqrstuvwxyz"
                                   "_!";

static WORD       gwUsers;
static URM_USERINFO gUserInfo[MAXUSERINFOS];

/*****************************************************************************************
 * Local functions declare
 ****************************************************************************************/
/****************************************************************************************/
void Shift(unsigned char *pbBuf, int iBufSize, int iBits)
{
  int             i;
  unsigned char   bTemp;

  if (iBits==0)
    return;
  else if (iBits<0)
  {
    iBits = -iBits;
    bTemp = pbBuf[0] >> (8-iBits);
    for (i=0; i<iBufSize-1; i++)
    {
      pbBuf[i] <<= iBits;
      pbBuf[i] |= pbBuf[i+1] >> (8-iBits);
    }
    pbBuf[i] <<= iBits;
      pbBuf[i] |= bTemp;
  }
  else
  {
    bTemp = (pbBuf[iBufSize-1] << (8-iBits)) & 0xff;
    for (i=iBufSize-1; i>0; i--)
    {
      pbBuf[i] >>= iBits;
      pbBuf[i] |= (pbBuf[i-1] << (8-iBits)) & 0xff;
    }
    pbBuf[i] >>= iBits;
    pbBuf[i] |= bTemp;
  }
}
/****************************************************************************************/
void EncryptPass(unsigned char *pszBuf,
                 const char *pcszUser, const char *pcszPass,
                 unsigned long dwTime)
{
  unsigned char   abBuf[2+URM_MAXNAMELEN+URM_MAXPASSLEN+4+2];
  int             iUserLen, iPassLen, i, j, k;

  iUserLen = strlen (pcszUser);
  iPassLen = strlen (pcszPass);

  // Length
  abBuf[0] = 'a' + iUserLen - 1;
  abBuf[1] = 'a' + iPassLen - 1;
  // User & Password
  for (i=0, k=2; i<iUserLen || i<iPassLen; i++)
  {
    if (i<iUserLen) abBuf[k++] = pcszUser[i];
    if (i<iPassLen) abBuf[k++] = pcszPass[i];
  }
  for (; k<URM_MAXNAMELEN+URM_MAXPASSLEN+2; )
    abBuf[k++] = gaszCharSet[rand () % sizeof gaszCharSet];
  // Shift right 3 bits
  Shift (abBuf, k, 3);
  // Time
  abBuf[k  ] = (unsigned char)(((dwTime>>24) ^ dwTime) & 0xff);
  abBuf[k+1] = (unsigned char)(((dwTime>>16) ^ dwTime) & 0xff);
  abBuf[k+2] = (unsigned char)(((dwTime>> 8) ^ dwTime) & 0xff);
  abBuf[k+3] = (unsigned char)(  dwTime                & 0xff);
  // Different by Time
  for (i=0; i<k; i++) abBuf[i] ^= abBuf[k+i%4];
  k += 4;
  Shift (abBuf, k, 2);
  // Different by Random 
  abBuf[k] = 0;
  for (i=0; i<k; i++) abBuf[k] += abBuf[i];
  abBuf[k+1] = rand () & 0xff;
  for (i=0; i<k; i++) abBuf[i] ^= abBuf[k+i%2];
  k += 2;
  Shift (abBuf, k, 2);
    
  for (i=0, j=0; i<k; i+=3)
  {
    pszBuf[j++] = gaszCharSet[abBuf[i]>>2];
    pszBuf[j++] = gaszCharSet[(abBuf[i]<<4)&0x3f|(abBuf[i+1]>>4)];
    if (i+1>=k) break;
    pszBuf[j++] = gaszCharSet[(abBuf[i+1]<<2)&0x3f|(abBuf[i+2]>>6)];
    if (i+2>=k) break;
    pszBuf[j++] = gaszCharSet[abBuf[i+2]&0x3f];
  }
  pszBuf[j] = '\0';
}
/****************************************************************************************/
int DecryptPass(const unsigned char *pcszEncryptToken,
                char *pszUser, char *pszPass, unsigned long *pdwTime)
{
  unsigned char   abBuf[2+URM_MAXNAMELEN+URM_MAXPASSLEN+4+2];
  int             iPackLen, iTokenLen;
  int             iUserLen, iPassLen, i, k;
  unsigned char   bCheckSum;

  iTokenLen = strlen (pcszEncryptToken);
  for (i=0, iPackLen=0; i<iTokenLen; i+=4)
  {
    for (k=0; k<sizeof gaszCharSet && gaszCharSet[k]!=pcszEncryptToken[i]; k++);
    abBuf[iPackLen] = k << 2;
    for (k=0; k<sizeof gaszCharSet && gaszCharSet[k]!=pcszEncryptToken[i+1]; k++);
    abBuf[iPackLen++] |= k >> 4;
    if (i+2>=iTokenLen) break;
    abBuf[iPackLen] = k << 4;
    for (k=0; k<sizeof gaszCharSet && gaszCharSet[k]!=pcszEncryptToken[i+2]; k++);
    abBuf[iPackLen++] |= k >> 2;
    if (i+3>=iTokenLen) break;
    abBuf[iPackLen] = k << 6;
    for (k=0; k<sizeof gaszCharSet && gaszCharSet[k]!=pcszEncryptToken[i+3]; k++);
    abBuf[iPackLen++] |= k;
  }
  Shift (abBuf, iPackLen, -2);
  if (iPackLen<2) return 0;
  iPackLen -= 2;
  for (i=0, bCheckSum=0; i<iPackLen; i++)
  {
    abBuf[i] ^= abBuf[iPackLen+i%2];
    bCheckSum += abBuf[i];
  }
  if (bCheckSum != abBuf[iPackLen]) return 0;
  Shift (abBuf, iPackLen, -2);
  iPackLen -= 4;
  for (i=0; i<iPackLen; i++) abBuf[i] ^= abBuf[iPackLen+i%4];
  Shift (abBuf, iPackLen, -3);
  iUserLen = abBuf[0] + 1 - 'a';
  iPassLen = abBuf[1] + 1 - 'a';
  if (iUserLen>URM_MAXNAMELEN) return 0;
  if (iPassLen>URM_MAXPASSLEN) return 0;

  for (i=0, k=2; i<iUserLen || i<iPassLen; i++)
  {
    if (i<iUserLen) pszUser[i] = abBuf[k++];
    if (i<iPassLen) pszPass[i] = abBuf[k++];
  }
  pszUser[iUserLen] = pszPass[iPassLen] = '\0';

  abBuf[iPackLen  ] ^= abBuf[iPackLen+3];
  abBuf[iPackLen+1] ^= abBuf[iPackLen+3];
  abBuf[iPackLen+2] ^= abBuf[iPackLen+3];
  *pdwTime = abBuf[iPackLen  ] * 16777216l +
             abBuf[iPackLen+1] * 65536l +
             abBuf[iPackLen+2] * 256l +
             abBuf[iPackLen+3];
  return 1;
}
/****************************************************************************************/
BOOL LoadUserList(void)
{
  HKEY            hKey;
  DWORD           Result, dwLen, dwType;
  WORD            i;
  CHAR            aszEncryptBuf[256];
  CHAR            aszKeyName[100];

  if (ERROR_SUCCESS == RegOpenKey (HKEY_LOCAL_MACHINE, REGKEYUSERDATABASE, &hKey))
  {
    dwLen = sizeof gwUsers;
    Result = RegQueryValueEx (hKey, "Users", 0, &dwType, (PBYTE)&gwUsers, &dwLen);
    if (Result != ERROR_SUCCESS) goto SetDefault;

    for (i=0; i<gwUsers; i++)
    {
      wsprintf (aszKeyName, "User_%04d", i);
      dwLen = sizeof aszEncryptBuf;
      Result = RegQueryValueEx (hKey, aszKeyName, 0, &dwType, aszEncryptBuf, &dwLen);
      if (Result != ERROR_SUCCESS) goto SetDefault;
      if (!DecryptPass (aszEncryptBuf,
                        gUserInfo[i].aszUserName, gUserInfo[i].aszPassword,
                        (PDWORD)&gUserInfo[i].RightMask)) goto SetDefault;
    }
    return TRUE;
  }
SetDefault:
  gwUsers = 0;
  return FALSE;
}
/****************************************************************************************/
BOOL SaveUserList(void)
{
  HKEY            hKey;
  DWORD           Result;
  WORD            i;
  CHAR            aszEncryptBuf[256];
  CHAR            aszKeyName[100];

  if (ERROR_SUCCESS == RegCreateKey (HKEY_LOCAL_MACHINE, REGKEYUSERDATABASE, &hKey))
  {
    Result = RegSetValueEx (hKey, "Users", 0, REG_BINARY, (PBYTE)&gwUsers, sizeof gwUsers);
    if (Result != ERROR_SUCCESS) return FALSE;
    for (i=0; i<gwUsers; i++)
    {
      wsprintf (aszKeyName, "User_%04d", i);
      EncryptPass (aszEncryptBuf,
                   gUserInfo[i].aszUserName, gUserInfo[i].aszPassword,
                   (unsigned long)gUserInfo[i].RightMask);
      Result = RegSetValueEx (hKey, aszKeyName, 0, REG_SZ,
                              aszEncryptBuf, strlen (aszEncryptBuf)+1);
      if (Result != ERROR_SUCCESS) return FALSE;
    }
  }

  return TRUE;
}
/****************************************************************************************/
BOOL CheckUser(URM_PUSERINFO pUserInfo)
{
  WORD            i;
  for (i=0; i<gwUsers; i++)
  {
    if (stricmp (gUserInfo[i].aszUserName, pUserInfo->aszUserName) == 0)
    {
      if (strcmp (gUserInfo[i].aszPassword, pUserInfo->aszPassword) == 0)
      {
        *pUserInfo = gUserInfo[i];
        return TRUE;
      }
      break;
    }
  }

  return FALSE;
}
/****************************************************************************************/
BOOL ChangeUserPass(PCSTR pcszUserName, URM_PPASSINFO pPassInfo)
{
  WORD            i;
  
  for (i=0; i<gwUsers; i++)
  {
    if (stricmp (gUserInfo[i].aszUserName, pcszUserName) == 0)
    {
      if (strcmp (gUserInfo[i].aszPassword, pPassInfo->aszOldPass) == 0)
      {
        strcpy (gUserInfo[i].aszPassword, pPassInfo->aszNewPass);
        return SaveUserList ();
      }
      break;
    }
  }

  return FALSE;
}
/****************************************************************************************/
/****************************************************************************************/
static void CentreDialog(HWND hWnd)
{
  RECT            WndRect;
  int             x, y, w, h;

  GetWindowRect (hWnd, &WndRect);
  w = WndRect.right - WndRect.left;
  h = WndRect.bottom - WndRect.top;
  x = (GetSystemMetrics (SM_CXFULLSCREEN)-w)/2;
  y = (GetSystemMetrics (SM_CYFULLSCREEN)+20-h)/2;
  MoveWindow (hWnd, x, y, w, h, TRUE);
}
/****************************************************************************************/
static BOOL CALLBACK UserEditProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  URM_PUSERINFO   pUserRight;
  CHAR            aszUserName[URM_MAXNAMELEN+1];
  CHAR            aszPassword1[URM_MAXPASSLEN+1];
  CHAR            aszPassword2[URM_MAXPASSLEN+1];

  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    SetWindowLong (hDlg, GWL_USERDATA, lParam);
    pUserRight = (URM_PUSERINFO)lParam;
    if (pUserRight->aszUserName[0] != '\0') SetWindowText (hDlg, "༭û");
    SetWindowText (GetDlgItem (hDlg, IDC_USERNAME), pUserRight->aszUserName);
    SetWindowText (GetDlgItem (hDlg, IDC_PASSWORD1), pUserRight->aszPassword);
    SetWindowText (GetDlgItem (hDlg, IDC_PASSWORD2), pUserRight->aszPassword);
    SendDlgItemMessage (hDlg, IDC_USERNAME, EM_LIMITTEXT, URM_MAXNAMELEN, 0);
    SendDlgItemMessage (hDlg, IDC_PASSWORD1, EM_LIMITTEXT, URM_MAXPASSLEN, 0);
    SendDlgItemMessage (hDlg, IDC_PASSWORD2, EM_LIMITTEXT, URM_MAXPASSLEN, 0);
    if (pUserRight->RightMask & UR_USERMANAGER)
    {
      SendDlgItemMessage (hDlg, IDC_USERRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
      SendDlgItemMessage (hDlg, IDC_MODIFYRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
      EnableWindow (GetDlgItem (hDlg, IDC_MODIFYRIGHT), FALSE);
      SendDlgItemMessage (hDlg, IDC_VIEWRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
      EnableWindow (GetDlgItem (hDlg, IDC_VIEWRIGHT), FALSE);
      SendDlgItemMessage (hDlg, IDC_SHUTRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
      EnableWindow (GetDlgItem (hDlg, IDC_SHUTRIGHT), FALSE);
    }
    else if (pUserRight->RightMask & UR_MODIFY)
    {
      SendDlgItemMessage (hDlg, IDC_MODIFYRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
      SendDlgItemMessage (hDlg, IDC_VIEWRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
      EnableWindow (GetDlgItem (hDlg, IDC_VIEWRIGHT), FALSE);
      SendDlgItemMessage (hDlg, IDC_SHUTRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
      EnableWindow (GetDlgItem (hDlg, IDC_SHUTRIGHT), FALSE);
    }
    else if (pUserRight->RightMask & UR_VIEW)
    {
      SendDlgItemMessage (hDlg, IDC_VIEWRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
      SendDlgItemMessage (hDlg, IDC_SHUTRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
      EnableWindow (GetDlgItem (hDlg, IDC_SHUTRIGHT), FALSE);
    }
    else if (pUserRight->RightMask & UR_SHUTDOWN)
      SendDlgItemMessage (hDlg, IDC_SHUTRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
    
    if (pUserRight->RightMask == UR_ADMIN)
      EnableWindow (GetDlgItem (hDlg, IDC_USERRIGHT), FALSE);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDC_USERRIGHT:
      if (SendDlgItemMessage (hDlg, IDC_USERRIGHT, BM_GETCHECK, 0, 0l) == BST_CHECKED)
      {
        SendDlgItemMessage (hDlg, IDC_MODIFYRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
        EnableWindow (GetDlgItem (hDlg, IDC_MODIFYRIGHT), FALSE);
        SendDlgItemMessage (hDlg, IDC_VIEWRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
        EnableWindow (GetDlgItem (hDlg, IDC_VIEWRIGHT), FALSE);
        SendDlgItemMessage (hDlg, IDC_SHUTRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
        EnableWindow (GetDlgItem (hDlg, IDC_SHUTRIGHT), FALSE);
      }
      else
        EnableWindow (GetDlgItem (hDlg, IDC_MODIFYRIGHT), TRUE);
      return TRUE;
    case IDC_MODIFYRIGHT:
      if (SendDlgItemMessage (hDlg, IDC_MODIFYRIGHT, BM_GETCHECK, 0, 0l) == BST_CHECKED)
      {
        SendDlgItemMessage (hDlg, IDC_VIEWRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
        EnableWindow (GetDlgItem (hDlg, IDC_VIEWRIGHT), FALSE);
        SendDlgItemMessage (hDlg, IDC_SHUTRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
        EnableWindow (GetDlgItem (hDlg, IDC_SHUTRIGHT), FALSE);
      }
      else
        EnableWindow (GetDlgItem (hDlg, IDC_VIEWRIGHT), TRUE);
      return TRUE;
    case IDC_VIEWRIGHT:
      if (SendDlgItemMessage (hDlg, IDC_VIEWRIGHT, BM_GETCHECK, 0, 0l) == BST_CHECKED)
      {
        SendDlgItemMessage (hDlg, IDC_SHUTRIGHT, BM_SETCHECK, BST_CHECKED, 0l);
        EnableWindow (GetDlgItem (hDlg, IDC_SHUTRIGHT), FALSE);
      }
      else
        EnableWindow (GetDlgItem (hDlg, IDC_SHUTRIGHT), TRUE);
      return TRUE;
    case IDOK:
      GetWindowText (GetDlgItem (hDlg, IDC_USERNAME), aszUserName, sizeof aszUserName);
      GetWindowText (GetDlgItem (hDlg, IDC_PASSWORD1), aszPassword1, sizeof aszPassword1);
      GetWindowText (GetDlgItem (hDlg, IDC_PASSWORD2), aszPassword2, sizeof aszPassword2);
      if (strlen (aszUserName) == 0)
      {
        SetFocus (GetDlgItem (hDlg, IDC_USERNAME));
        return TRUE;
      }
      if (strlen (aszPassword1) == 0 || strcmp (aszPassword1, aszPassword2) != 0)
      {
        SetWindowText (GetDlgItem (hDlg, IDC_PASSWORD1), "");
        SetWindowText (GetDlgItem (hDlg, IDC_PASSWORD2), "");
        SetFocus (GetDlgItem (hDlg, IDC_PASSWORD1));
        return TRUE;
      }
      pUserRight = (URM_PUSERINFO)GetWindowLong (hDlg, GWL_USERDATA);
      strcpy (pUserRight->aszUserName, aszUserName);
      strcpy (pUserRight->aszPassword, aszPassword1);
      if (pUserRight->RightMask != UR_ADMIN)
      {
        pUserRight->RightMask = UR_NONE;
        if (SendDlgItemMessage (hDlg, IDC_USERRIGHT, BM_GETCHECK, 0, 0l) == BST_CHECKED)
          pUserRight->RightMask |= UR_USERMANAGER;
        if (SendDlgItemMessage (hDlg, IDC_MODIFYRIGHT, BM_GETCHECK, 0, 0l) == BST_CHECKED)
          pUserRight->RightMask |= UR_MODIFY;
        if (SendDlgItemMessage (hDlg, IDC_VIEWRIGHT, BM_GETCHECK, 0, 0l) == BST_CHECKED)
          pUserRight->RightMask |= UR_VIEW;
        if (SendDlgItemMessage (hDlg, IDC_SHUTRIGHT, BM_GETCHECK, 0, 0l) == BST_CHECKED)
          pUserRight->RightMask |= UR_SHUTDOWN;
      }
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
BOOL UserEditDlg(HWND hParentWnd, URM_PUSERINFO pUserRight)
{
  URM_USERINFO    UserRight = *pUserRight;
  if (DialogBoxParam (ghInst, MAKEINTRESOURCE (IDD_USEREDIT), hParentWnd, UserEditProc,
                      (LPARAM)&UserRight) == IDOK)
  {
    *pUserRight = UserRight;
    return TRUE;
  }
  else
    return FALSE;
}
/****************************************************************************************/
static BOOL CALLBACK UserLoginProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  URM_PUSERINFO   pUserRight;

  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    SetWindowLong (hDlg, GWL_USERDATA, lParam);
    SendDlgItemMessage (hDlg, IDC_USERNAME, EM_LIMITTEXT, URM_MAXNAMELEN, 0);
    SendDlgItemMessage (hDlg, IDC_PASSWORD, EM_LIMITTEXT, URM_MAXPASSLEN, 0);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDOK:
      pUserRight = (URM_PUSERINFO)GetWindowLong (hDlg, GWL_USERDATA);
      GetWindowText (GetDlgItem (hDlg, IDC_USERNAME),
                     pUserRight->aszUserName, sizeof pUserRight->aszUserName);
      if (strlen (pUserRight->aszUserName) == 0)
      {
        SetFocus (GetDlgItem (hDlg, IDC_USERNAME));
        return TRUE;
      }
      GetWindowText (GetDlgItem (hDlg, IDC_PASSWORD),
                     pUserRight->aszPassword, sizeof pUserRight->aszPassword);
      if (strlen (pUserRight->aszPassword) == 0)
      {
        SetFocus (GetDlgItem (hDlg, IDC_PASSWORD));
        return TRUE;
      }
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  }
  return FALSE;
}
/****************************************************************************************/
BOOL UserLoginDlg(HWND hParantWnd, URM_PUSERINFO pUserRight)
{
  return DialogBoxParam (ghInst, MAKEINTRESOURCE (IDD_USERLOGIN), hParantWnd, UserLoginProc,
                         (LPARAM)pUserRight) == IDOK;
}
/****************************************************************************************/
static BOOL CALLBACK ChangePasswordProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  URM_PPASSINFO   pPassInfo;
  CHAR            aszRetypePass[URM_MAXPASSLEN+1];

  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    SetWindowLong (hDlg, GWL_USERDATA, lParam);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDOK:
      pPassInfo = (URM_PPASSINFO)GetWindowLong (hDlg, GWL_USERDATA);
      GetWindowText (GetDlgItem (hDlg, IDC_OLDPASS),
                     pPassInfo->aszOldPass, sizeof pPassInfo->aszOldPass);
      if (strlen (pPassInfo->aszOldPass) == 0)
      {
        SetFocus (GetDlgItem (hDlg, IDC_OLDPASS));
        return TRUE;
      }
      GetWindowText (GetDlgItem (hDlg, IDC_NEWPASS1),
                     pPassInfo->aszNewPass, sizeof pPassInfo->aszNewPass);
      if (strlen (pPassInfo->aszNewPass) == 0)
      {
        SetFocus (GetDlgItem (hDlg, IDC_NEWPASS1));
        return TRUE;
      }
      GetWindowText (GetDlgItem (hDlg, IDC_NEWPASS2),
                     aszRetypePass, sizeof aszRetypePass);
      if (strcmp (pPassInfo->aszNewPass, aszRetypePass) != 0)
      {
        LoadString (ghInst, IDS_PASSERR_TITLE, gaszString, sizeof gaszString);
        LoadString (ghInst, IDS_PASSERR_DISMISSPASS2, gaszString2, sizeof gaszString2);
        MessageBox (hDlg, gaszString2, gaszString, MB_OK);
        SetWindowText (GetDlgItem (hDlg, IDC_NEWPASS1), NULL);
        SetWindowText (GetDlgItem (hDlg, IDC_NEWPASS2), NULL);
        SetFocus (GetDlgItem (hDlg, IDC_NEWPASS1));
        return TRUE;
      }
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  }
  return FALSE;
}
/****************************************************************************************/
BOOL ChangePasswordDlg(HWND hParentWnd, URM_PPASSINFO pPassInfo)
{
  return DialogBoxParam (ghInst, MAKEINTRESOURCE (IDD_USERPASSCHANGE), hParentWnd,
                         ChangePasswordProc, (LPARAM)pPassInfo) == IDOK;
}
/****************************************************************************************/
static BOOL CALLBACK UserManagerProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  int             i, j;
  URM_USERINFO    UserInfo;

  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    for (i=0; i<gwUsers; i++)
      SendDlgItemMessage (hDlg, IDC_USERLIST, LB_INSERTSTRING,
                          i, (LPARAM)gUserInfo[i].aszUserName);
    SendDlgItemMessage (hDlg, IDC_USERLIST, LB_SETCURSEL, 0, 0);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDC_ADDUSER:
      if (gwUsers>=MAXUSERINFOS)
      {
        LoadString (ghInst, IDS_PASSERR_MAXUSERS, gaszString, sizeof gaszString);
        LoadString (ghInst, IDS_PASSERR_TITLE, gaszString2, sizeof gaszString2);
        MessageBox (hDlg, gaszString, gaszString2, MB_OK);
        return TRUE;
      }
      UserInfo.aszUserName[0] = '\0';
      UserInfo.aszPassword[0] = '\0';
      UserInfo.RightMask = UR_VIEW | UR_SHUTDOWN;
      if (UserEditDlg (hDlg, &UserInfo))
      {
        for (i=0; i<gwUsers; i++)
        {
          j = stricmp (UserInfo.aszUserName, gUserInfo[i].aszUserName);
          if (j<0)
            break;
          if (j == 0)
          {
            LoadString (ghInst, IDS_PASSERR_DUPUSER, gaszString, sizeof gaszString);
            LoadString (ghInst, IDS_PASSERR_TITLE, gaszString2, sizeof gaszString2);
            MessageBox (hDlg, gaszString, gaszString2, MB_OK);
            return TRUE;
          }
        }
        for (j=gwUsers; j>i; j--)
        {
          gUserInfo[j] = gUserInfo[j-1];
        }
        gUserInfo[i] = UserInfo;
        SendDlgItemMessage (hDlg, IDC_USERLIST, LB_INSERTSTRING,
                            i, (LPARAM)UserInfo.aszUserName);
        SendDlgItemMessage (hDlg, IDC_USERLIST, LB_SETCURSEL, i, 0);
        gwUsers ++;
        SaveUserList ();
      }
      return TRUE;
    case IDC_EDITUSER:
      i = SendDlgItemMessage (hDlg, IDC_USERLIST, LB_GETCURSEL, 0, 0);
      UserInfo = gUserInfo[i];
      if (UserEditDlg (hDlg, &UserInfo))
      {
        for (j=0; j<gwUsers; j++)
        {
          if (j == i) continue;
          if (stricmp (UserInfo.aszUserName, gUserInfo[j].aszUserName) == 0)
          {
            LoadString (ghInst, IDS_PASSERR_DUPUSER, gaszString, sizeof gaszString);
            LoadString (ghInst, IDS_PASSERR_TITLE, gaszString2, sizeof gaszString2);
            MessageBox (hDlg, gaszString, gaszString2, MB_OK);
            break;
          }
        }
        if (j>=gwUsers)
        {
          SendDlgItemMessage (hDlg, IDC_USERLIST, LB_DELETESTRING, i, 0);
          gwUsers --;
          for (; i<gwUsers; i++) gUserInfo[i] = gUserInfo[i+1];

          for (i=0; i<gwUsers; i++)
            if (stricmp (UserInfo.aszUserName, gUserInfo[i].aszUserName) <0)
              break;
          for (j=gwUsers; j>i; j--)
            gUserInfo[j] = gUserInfo[j-1];
          gUserInfo[i] = UserInfo;
          SendDlgItemMessage (hDlg, IDC_USERLIST, LB_INSERTSTRING,
                              i, (LPARAM)UserInfo.aszUserName);
          SendDlgItemMessage (hDlg, IDC_USERLIST, LB_SETCURSEL, i, 0);
          gwUsers ++;
          SaveUserList ();
        }
      }
      return TRUE;
    case IDC_DELUSER:
      i = SendDlgItemMessage (hDlg, IDC_USERLIST, LB_GETCURSEL, 0, 0);
      if (i != LB_ERR)
      {
        if (gUserInfo[i].RightMask == UR_ADMIN)
        {
          LoadString (ghInst, IDS_PASSERR_DELADMIN, gaszString, sizeof gaszString);
          LoadString (ghInst, IDS_PASSERR_TITLE, gaszString2, sizeof gaszString2);
          MessageBox (hDlg, gaszString, gaszString2, MB_OK);
        }
        else
        {
          LoadString (ghInst, IDS_PASS_DELCONFIRM, gaszString, sizeof gaszString);
          wsprintf (gaszString2, gaszString, gUserInfo[i].aszUserName);
          if (MessageBox (hDlg, gaszString2, "", MB_YESNO) == IDYES)
          {
            gwUsers --;
            SendDlgItemMessage (hDlg, IDC_USERLIST, LB_DELETESTRING, i, 0);
            if (i < gwUsers)
              SendDlgItemMessage (hDlg, IDC_USERLIST, LB_SETCURSEL, i, 0);
            else
              SendDlgItemMessage (hDlg, IDC_USERLIST, LB_SETCURSEL, gwUsers-1, 0);
            for (; i<gwUsers; i++) gUserInfo[i] = gUserInfo[i+1];
            SaveUserList ();
          }
        }
      }
      return TRUE;
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
void UserManagerDlg(HINSTANCE hInst, HWND hWnd)
{
  DialogBox (hInst, MAKEINTRESOURCE (IDD_USERMANAGER), hWnd, UserManagerProc);
}
/****************************************************************************************/
BOOL AddAdmin (void)
{
  strcpy (gUserInfo[0].aszUserName, "admin");
  gUserInfo[0].aszPassword[0] = '\0';
  gUserInfo[0].RightMask = UR_ADMIN;
  if (UserEditDlg (NULL, &gUserInfo[0]))
  {
    gwUsers = 1;
    return SaveUserList ();
  }
  else
  {
    gwUsers = 0;
    return FALSE;
  }
}
/****************************************************************************************/
