/*****************************************************************************************
 * File name     : STATWND.C
 * Author        : Han Zhi Qiang
 * Module        : SOCKTOOL
 * Purpose       : The status window
 * Create Date   : May 16, 2001 - 11:26
 * Modify history: 
 *****************************************************************************************
 */

#include <windows.h>
#include <socktool.h>
#include <stdio.h>
#include "resource.h"
#include "dllmain.h"

/*****************************************************************************************
 * MACRO functions declare
 ****************************************************************************************/

/*****************************************************************************************
 * Identifier constant declare
 ****************************************************************************************/
#define STATUSWNDCLASSNAME              "SocketTestStatusWndClass"
#define WHEELDELTAPERLINE               40

/*****************************************************************************************
 * Data structure declare
 ****************************************************************************************/
typedef struct tagINFOITEM
{
  struct tagINFOITEM *pPrevItem;
  struct tagINFOITEM *pNextItem;
  OMPT                Type;
  PCHAR               pszContent;
  DWORD               dwLineNo;
} INFOITEM;

typedef struct tagSTATWNDINFO
{
  INFOITEM       *pStatusInfo;
  INFOITEM       *pCurStatusInfo;

  HBITMAP         hInfoIconSuccess;
  HBITMAP         hInfoIconError;
  HBITMAP         hInfoIconInfo;
  HBITMAP         hInfoIconUpload;
  HBITMAP         hInfoIconDownload;

  HFONT           hFont;
  WORD            wFontHeight;

  WORD            wStatusWndWidth, wStatusWndHeight;
  WORD            wStatusWndTextLines;
  DWORD           dwTotalLine;
  STFLAG          Flags;

  OMPT            LineType;
  BYTE            bCacheSize;
  BYTE            abStreamCache[16];
} STATWNDINFO, *PSTATWNDINFO;

/*****************************************************************************************
 * Local variable declare
 ****************************************************************************************/

/*****************************************************************************************
 * Local functions declare
 ****************************************************************************************/

/****************************************************************************************/
static void SetStatusWndTextPos(HWND hWnd, PSTATWNDINFO pStatusWndInfo, DWORD dwPos)
{
  INFOITEM       *pStatusInfo;
  SCROLLINFO      ScrollInfo;

  pStatusInfo = pStatusWndInfo->pCurStatusInfo;

  while (pStatusInfo->dwLineNo != dwPos)
  {
    if (pStatusInfo->dwLineNo > dwPos)
      pStatusInfo = pStatusInfo->pPrevItem;
    else
      pStatusInfo = pStatusInfo->pNextItem;
  }

  if (pStatusInfo != pStatusWndInfo->pCurStatusInfo)
  {
    pStatusWndInfo->pCurStatusInfo = pStatusInfo;
    ScrollInfo.cbSize = sizeof (SCROLLINFO);
    ScrollInfo.fMask  = SIF_PAGE | SIF_POS | SIF_RANGE;
    ScrollInfo.nMin   = 0;
    ScrollInfo.nMax   = pStatusWndInfo->dwTotalLine - pStatusWndInfo->wStatusWndTextLines;
    ScrollInfo.nPage  = 1;
    ScrollInfo.nPos   = dwPos;
    SetScrollInfo (hWnd, SB_VERT, &ScrollInfo, TRUE);
    InvalidateRect (hWnd, NULL, FALSE);
  }
}
/****************************************************************************************/
static void DrawStatusIcon(HDC hDC, int x, int y, HBITMAP hBitmap)
{
  HDC             hMemDC = CreateCompatibleDC (hDC);
  HBITMAP         hOldBitmap = SelectObject (hMemDC, hBitmap);
  BitBlt (hDC, x, y, 16, 16, hMemDC, 0, 0, SRCCOPY);
  SelectObject (hMemDC, hOldBitmap);
  DeleteDC (hMemDC);
}
/****************************************************************************************/
static LRESULT CALLBACK StatusWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  PSTATWNDINFO    pStatusWndInfo;
  WORD            i;
  SHORT           iDelta;
  HDC             hDC;
  PAINTSTRUCT     ps;
  CREATESTRUCT   *pcs;
  INFOITEM       *pStatusInfo;
  HFONT           hOldFont;
  HPEN            hPen, hOldPen;
  HBRUSH          hBrush, hOldBrush;
  RECT            TextRect;
  COLORREF        TextBkColor, TextFgColor;
  SCROLLINFO      ScrollInfo;
  HINSTANCE       hInst;
  LOGFONT         LogicalFont;
  TEXTMETRIC      TextMetric;
  BOOLEAN         fFocusWnd;

  switch (uMsg)
  {
  case WM_PAINT:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    hDC = BeginPaint (hWnd, &ps);
    if (pStatusWndInfo->pCurStatusInfo)
    {
      if (pStatusWndInfo->hFont) hOldFont = SelectObject (hDC, pStatusWndInfo->hFont);
      hPen = CreatePen (PS_SOLID, 0, RGB (192, 192, 192));
      hOldPen = SelectObject (hDC, hPen);
    
      pStatusInfo = pStatusWndInfo->pCurStatusInfo;
      fFocusWnd = (pStatusWndInfo->Flags & STFLAG_ALWAYSFOCUS) || (GetFocus () == hWnd);
      for (i=0; i<pStatusWndInfo->wStatusWndHeight; i+=pStatusWndInfo->wFontHeight+1)
      {
        if (pStatusInfo)
        {
          if (pStatusWndInfo->Flags & STFLAG_IGNORETYPE)
          {
            TextBkColor = fFocusWnd ? RGB (223, 239, 255) : RGB (255, 255, 255);
            TextFgColor = RGB (0, 0, 0);
          }
          else
          {
            switch (pStatusInfo->Type)
            {
            case OMPT_SUCCESS:
              DrawStatusIcon (hDC, 0, i, pStatusWndInfo->hInfoIconSuccess);
              TextBkColor = fFocusWnd ? RGB (223, 255, 223) : RGB (200, 200, 200);
              TextFgColor = RGB (0, 0, 0);
              break;
            case OMPT_ERROR:
              DrawStatusIcon (hDC, 0, i, pStatusWndInfo->hInfoIconError);
              TextBkColor = fFocusWnd ? RGB (255, 223, 223) : RGB (200, 200, 200);
              TextFgColor = RGB (0, 0, 0);
              break;
            case OMPT_INFO:
              DrawStatusIcon (hDC, 0, i, pStatusWndInfo->hInfoIconInfo);
              TextBkColor = fFocusWnd ? RGB (223, 239, 255) : RGB (255, 255, 255);
              TextFgColor = RGB (0, 0, 0);
              break;
            case OMPT_UPLOAD:
              DrawStatusIcon (hDC, 0, i, pStatusWndInfo->hInfoIconUpload);
              TextBkColor = fFocusWnd ? RGB (233, 255, 255) : RGB (218, 218, 218);
              TextFgColor = RGB (0, 0, 0);
              break;
            case OMPT_DOWNLOAD:
              DrawStatusIcon (hDC, 0, i, pStatusWndInfo->hInfoIconDownload);
              TextBkColor = fFocusWnd ? RGB (255, 255, 239) : RGB (236, 236, 236);
              TextFgColor = RGB (0, 0, 0);
              break;
            }
          }
          SetTextColor (hDC, TextFgColor);
          SetBkColor (hDC, TextBkColor);
          TextRect.top    = i;
          TextRect.left   = pStatusWndInfo->Flags & STFLAG_IGNORETYPE ? 0 : 18;
          TextRect.bottom = i + pStatusWndInfo->wFontHeight;
          TextRect.right  = pStatusWndInfo->wStatusWndWidth;
          DrawText (hDC, pStatusInfo->pszContent, -1, &TextRect, DT_SINGLELINE|DT_NOPREFIX);
          DrawText (hDC, pStatusInfo->pszContent, -1, &TextRect, 
                    DT_SINGLELINE|DT_CALCRECT|DT_NOPREFIX);
          TextRect.left = TextRect.right;
          TextRect.right = pStatusWndInfo->wStatusWndWidth;
          
          hBrush = CreateSolidBrush (TextBkColor);
          hOldBrush = SelectObject (hDC, hBrush);
          FillRect (hDC, &TextRect, hBrush);
          SelectObject (hDC, hOldBrush);
          DeleteObject (hBrush);

          pStatusInfo = pStatusInfo->pNextItem;
          MoveToEx (hDC, 0, i+pStatusWndInfo->wFontHeight, NULL);
          LineTo (hDC, pStatusWndInfo->wStatusWndWidth, i+pStatusWndInfo->wFontHeight);
        }
        else
          break;
      }
      hBrush = CreateSolidBrush (RGB (0xff, 0xff, 0xff));
      hOldBrush = SelectObject (hDC, hBrush);
      TextRect.top    = i;
      TextRect.left   = 0;
      TextRect.bottom = pStatusWndInfo->wStatusWndHeight;
      TextRect.right  = pStatusWndInfo->wStatusWndWidth;
      FillRect (hDC, &TextRect, hBrush);
      SelectObject (hDC, hOldBrush);
      DeleteObject (hBrush);

      SelectObject (hDC, hOldPen);
      DeleteObject (hPen);
      if (pStatusWndInfo->hFont) SelectObject (hDC, hOldFont);
    }
    else
    {
      hBrush = CreateSolidBrush (RGB (0xff, 0xff, 0xff));
      hOldBrush = SelectObject (hDC, hBrush);
      TextRect.top    = 0;
      TextRect.left   = 0;
      TextRect.bottom = pStatusWndInfo->wStatusWndHeight;
      TextRect.right  = pStatusWndInfo->wStatusWndWidth;
      FillRect (hDC, &TextRect, hBrush);
      SelectObject (hDC, hOldBrush);
      DeleteObject (hBrush);
    }
    EndPaint (hWnd, &ps);
    return 0;
  case WM_VSCROLL:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    ScrollInfo.cbSize = sizeof (SCROLLINFO);
    ScrollInfo.fMask  = SIF_POS | SIF_RANGE;
    GetScrollInfo (hWnd, SB_VERT, &ScrollInfo);
    switch (LOWORD(wParam))
    {
    case SB_LINEUP:
      if (ScrollInfo.nPos > ScrollInfo.nMin)
      {
        ScrollInfo.nPos --;
        SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
      }
      break;
    case SB_LINEDOWN:
      if (ScrollInfo.nPos < ScrollInfo.nMax)
      {
        ScrollInfo.nPos ++;
        SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
      }
      break;
    case SB_PAGEUP:
      if (ScrollInfo.nPos > ScrollInfo.nMin)
      {
        ScrollInfo.nPos -= pStatusWndInfo->wStatusWndTextLines;
        if (ScrollInfo.nPos < ScrollInfo.nMin) ScrollInfo.nPos = ScrollInfo.nMin;
        SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
      }
      break;
    case SB_PAGEDOWN:
      if (ScrollInfo.nPos < ScrollInfo.nMax)
      {
        ScrollInfo.nPos += pStatusWndInfo->wStatusWndTextLines;
        if (ScrollInfo.nPos > ScrollInfo.nMax) ScrollInfo.nPos = ScrollInfo.nMax;
        SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
      }
      break;
    case SB_THUMBTRACK:
      SetStatusWndTextPos (hWnd, pStatusWndInfo, HIWORD(wParam));
      break;
    }
    return 0;
  case WM_SIZE:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    pStatusWndInfo->wStatusWndWidth     = LOWORD(lParam);
    pStatusWndInfo->wStatusWndHeight    = HIWORD(lParam);
    if (pStatusWndInfo->pCurStatusInfo)
      i = pStatusWndInfo->dwTotalLine - pStatusWndInfo->wStatusWndTextLines ==
          pStatusWndInfo->pCurStatusInfo->dwLineNo;
    else
      i = 0;
    pStatusWndInfo->wStatusWndTextLines = pStatusWndInfo->wStatusWndHeight /
                                          (pStatusWndInfo->wFontHeight+1);
    if (pStatusWndInfo->wStatusWndTextLines<1) pStatusWndInfo->wStatusWndTextLines = 10;
    if (i)
    {
      if (pStatusWndInfo->dwTotalLine <= pStatusWndInfo->wStatusWndTextLines)
        i = 0;
      else
        i = (WORD)(pStatusWndInfo->dwTotalLine - pStatusWndInfo->wStatusWndTextLines);
      if (i>=pStatusWndInfo->pCurStatusInfo->dwLineNo)
      {
        while (i>pStatusWndInfo->pCurStatusInfo->dwLineNo)
          pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pCurStatusInfo->pNextItem;
      }
      else
      {
        while (i<pStatusWndInfo->pCurStatusInfo->dwLineNo)
          pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pCurStatusInfo->pPrevItem;
      }
    }
    if (pStatusWndInfo->wStatusWndHeight<0) pStatusWndInfo->wStatusWndHeight = 0;
    ScrollInfo.cbSize = sizeof (SCROLLINFO);
    ScrollInfo.fMask  = SIF_PAGE | SIF_POS | SIF_RANGE;
    ScrollInfo.nMin   = 0;
    ScrollInfo.nMax   = pStatusWndInfo->dwTotalLine-pStatusWndInfo->wStatusWndTextLines;
    ScrollInfo.nPage  = 1;
    ScrollInfo.nPos   = pStatusWndInfo->pCurStatusInfo ? 
                          pStatusWndInfo->pCurStatusInfo->dwLineNo : 0;
    SetScrollInfo (hWnd, SB_VERT, &ScrollInfo, TRUE);
    return 0;
  case WM_KEYDOWN:
    switch (wParam)
    {
    case VK_UP:    PostMessage (hWnd, WM_VSCROLL, SB_LINEUP, 0);   break;
    case VK_DOWN:  PostMessage (hWnd, WM_VSCROLL, SB_LINEDOWN, 0); break;
    case VK_PRIOR: PostMessage (hWnd, WM_VSCROLL, SB_PAGEUP, 0);   break;
    case VK_NEXT:  PostMessage (hWnd, WM_VSCROLL, SB_PAGEDOWN, 0); break;
    }
    return 0;
  case WM_MOUSEWHEEL:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    ScrollInfo.cbSize = sizeof (SCROLLINFO);
    ScrollInfo.fMask  = SIF_POS | SIF_RANGE;
    GetScrollInfo (hWnd, SB_VERT, &ScrollInfo);
    iDelta = HIWORD(wParam);

    if (iDelta < 0)
    {
      iDelta /= -WHEELDELTAPERLINE;
      if (iDelta == 0) iDelta = 1;
      ScrollInfo.nPos += iDelta;
      if (ScrollInfo.nPos > ScrollInfo.nMax) ScrollInfo.nPos = ScrollInfo.nMax;
      SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
    }
    else
    {
      iDelta /= WHEELDELTAPERLINE;
      if (iDelta == 0) iDelta = 1;
      ScrollInfo.nPos -= iDelta;
      if (ScrollInfo.nPos < ScrollInfo.nMin) ScrollInfo.nPos = ScrollInfo.nMin;
      SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
    }
    return 0;
  case WM_CREATE:
    pStatusWndInfo = malloc (sizeof *pStatusWndInfo);
    if (pStatusWndInfo == NULL) return -1;
    SetWindowLong (hWnd, GWL_USERDATA, (LONG)pStatusWndInfo);

    memset (pStatusWndInfo, 0, sizeof *pStatusWndInfo);
    pcs = (CREATESTRUCT *)lParam;
    pStatusWndInfo->Flags = (STFLAG)pcs->lpCreateParams;
    
    hInst = GetDLLInst ();
    pStatusWndInfo->hInfoIconSuccess  = LoadBitmap (hInst, MAKEINTRESOURCE(IDB_SUCCESS));
    pStatusWndInfo->hInfoIconError    = LoadBitmap (hInst, MAKEINTRESOURCE(IDB_ERROR));
    pStatusWndInfo->hInfoIconInfo     = LoadBitmap (hInst, MAKEINTRESOURCE(IDB_INFO));
    pStatusWndInfo->hInfoIconUpload   = LoadBitmap (hInst, MAKEINTRESOURCE(IDB_UPLOAD));
    pStatusWndInfo->hInfoIconDownload = LoadBitmap (hInst, MAKEINTRESOURCE(IDB_DOWNLOAD));

    LogicalFont.lfHeight         = 0;
    LogicalFont.lfWidth          = 0;
    LogicalFont.lfEscapement     = 0;
    LogicalFont.lfOrientation    = 0;
    LogicalFont.lfWeight         = FW_NORMAL;
    LogicalFont.lfItalic         = FALSE;
    LogicalFont.lfUnderline      = FALSE;
    LogicalFont.lfStrikeOut      = FALSE;
    LogicalFont.lfCharSet        = ANSI_CHARSET;
    LogicalFont.lfOutPrecision   = OUT_DEFAULT_PRECIS;
    LogicalFont.lfClipPrecision  = CLIP_DEFAULT_PRECIS;
    LogicalFont.lfQuality        = DEFAULT_QUALITY;
    LogicalFont.lfPitchAndFamily = FIXED_PITCH;
    strcpy (LogicalFont.lfFaceName, "Fixedsys");
    pStatusWndInfo->hFont = CreateFontIndirect (&LogicalFont);
    if (pStatusWndInfo->hFont)
    {
      hDC = GetDC (hWnd);
      hOldFont = SelectObject (hDC, pStatusWndInfo->hFont);
      GetTextMetrics (hDC, &TextMetric);
      pStatusWndInfo->wFontHeight = (WORD)TextMetric.tmHeight;
      SelectObject (hDC, hOldFont);
      ReleaseDC (hWnd, hDC);
    }
    else
      pStatusWndInfo->wFontHeight = 16;
    break;
  case WM_DESTROY:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    while (pStatusWndInfo->pStatusInfo)
    {
      pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pStatusInfo;
      pStatusWndInfo->pStatusInfo = pStatusWndInfo->pStatusInfo->pNextItem;
      free (pStatusWndInfo->pCurStatusInfo->pszContent);
      free (pStatusWndInfo->pCurStatusInfo);
    }
    if (pStatusWndInfo->hFont) DeleteObject (pStatusWndInfo->hFont);
    free (pStatusWndInfo);
    break;
  case WM_SETFOCUS:
  case WM_KILLFOCUS:
    InvalidateRect (hWnd, NULL, FALSE);
    return 0;
  case WM_GETDLGCODE:
    return DLGC_WANTARROWS;
  case WM_LBUTTONDOWN:
    SetFocus (hWnd);
    return 0;
  }
  return DefWindowProc (hWnd, uMsg, wParam, lParam);
}
/****************************************************************************************/
SOCKTOOLAPI HWND PASCAL STATWND_CreateWnd(HINSTANCE hInst, HWND hWndParent, STFLAG Flags)
{
  HWND            hWnd;
  WNDCLASSEX      WndClass;

  WndClass.cbSize        = sizeof (WNDCLASSEX);
  WndClass.style         = 0;
  WndClass.lpfnWndProc   = StatusWndProc;
  WndClass.cbClsExtra    = 0;
  WndClass.cbWndExtra    = 0;
  WndClass.hInstance     = hInst;
  WndClass.hIcon         = NULL;
  WndClass.hCursor       = LoadCursor (NULL, IDC_ARROW);
  WndClass.hbrBackground = GetStockObject(WHITE_BRUSH);
  WndClass.lpszMenuName  = NULL;
  WndClass.lpszClassName = STATUSWNDCLASSNAME;
  WndClass.hIconSm       = NULL;
  if (!RegisterClassEx (&WndClass)) return NULL;

  hWnd = CreateWindowEx (0, STATUSWNDCLASSNAME, NULL,
                         WS_CHILD|WS_VSCROLL|(Flags&STFLAG_TABSTOP ? WS_TABSTOP : 0),
                         0, 0, 0, 0, hWndParent, NULL, hInst, (LPVOID)Flags);
  if (hWnd == NULL) return NULL;
  ShowWindow (hWnd, SW_SHOW);

  return hWnd;
}
/****************************************************************************************/
SOCKTOOLAPI DWORD PASCAL STATWND_SetFlag(HWND hWnd, STFLAG Flags)
{
  PSTATWNDINFO    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);

  if (!(pStatusWndInfo->Flags&STFLAG_INSPECTSTREAM) && (Flags&STFLAG_INSPECTSTREAM))
    pStatusWndInfo->bCacheSize = 0;
  pStatusWndInfo->Flags |= Flags;

  return pStatusWndInfo->Flags;
}
/****************************************************************************************/
SOCKTOOLAPI DWORD PASCAL STATWND_ClearFlag(HWND hWnd, STFLAG Flags)
{
  PSTATWNDINFO    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
  pStatusWndInfo->Flags &= ~Flags;
  return pStatusWndInfo->Flags;
}
/****************************************************************************************/
static void AppendStatusInfoSub(HWND hWnd, PSTATWNDINFO pStatusWndInfo, OMPT Type,
                                PCCHAR pcszStatusInfo)
{
  INFOITEM       *pStatusInfo, *pTempInfo;
  PCHAR           pszContent;
  SCROLLINFO      ScrollInfo;

  pStatusInfo = malloc (sizeof (INFOITEM));
  if (pStatusInfo == NULL) return;
  pszContent = malloc (strlen (pcszStatusInfo)+1);
  if (pszContent == NULL)
  {
    free (pStatusInfo);
    return;
  }
  strcpy (pszContent, pcszStatusInfo);

  pStatusInfo->pNextItem  = NULL;
  pStatusInfo->pPrevItem  = NULL;
  pStatusInfo->Type       = Type;
  pStatusInfo->pszContent = pszContent;

  if (pStatusWndInfo->pStatusInfo == NULL)
  {
    pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pStatusInfo = pStatusInfo;
    pStatusInfo->dwLineNo = 0;
    pStatusWndInfo->dwTotalLine = 1;
  }
  else
  {
    pTempInfo = pStatusWndInfo->pCurStatusInfo;
    while (pTempInfo->pNextItem) pTempInfo = pTempInfo->pNextItem;
    pTempInfo->pNextItem = pStatusInfo;
    pStatusInfo->pPrevItem = pTempInfo;
    pStatusInfo->dwLineNo = pTempInfo->dwLineNo + 1;
    pStatusWndInfo->dwTotalLine ++;
  }

  if (pStatusInfo->dwLineNo - pStatusWndInfo->pCurStatusInfo->dwLineNo ==
      pStatusWndInfo->wStatusWndTextLines)
    pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pCurStatusInfo->pNextItem;
  ScrollInfo.cbSize = sizeof (SCROLLINFO);
  ScrollInfo.fMask  = SIF_PAGE | SIF_POS | SIF_RANGE;
  ScrollInfo.nMin   = 0;
  ScrollInfo.nMax   = pStatusWndInfo->dwTotalLine - pStatusWndInfo->wStatusWndTextLines;
  ScrollInfo.nPage  = 1;
  ScrollInfo.nPos   = pStatusWndInfo->pCurStatusInfo->dwLineNo;
  SetScrollInfo (hWnd, SB_VERT, &ScrollInfo, TRUE);
}
/****************************************************************************************/
SOCKTOOLAPI WORD PASCAL STATWND_AppendInfoV(HWND hWnd, OMPT Type, PCCHAR pcszFormatString,
                                            va_list ArgList)
{
  PSTATWNDINFO    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
  CHAR            aszBuf[1024], aszBuf2[10];
  DWORD           dwSize, i, j;
  PCBYTE          pcbBuf;
  PCCHAR          pcszIndicator;
  CHAR            aszASCIIView[17];

  if (pStatusWndInfo->Flags & STFLAG_INSPECTSTREAM) return 0;

  if (pcszFormatString)
  {
    vsprintf (aszBuf, pcszFormatString, ArgList);
    if (pStatusWndInfo->Flags & STFLAG_EXPANDNEWLINE)
    {
      for (i=0, j=0; aszBuf[j]; j++)
      {
        if (aszBuf[j] == '\n')
        {
          if (j>i && aszBuf[j] == '\r') aszBuf[j-1] = '\0'; else aszBuf[j] = '\0';
          AppendStatusInfoSub (hWnd, pStatusWndInfo, Type, aszBuf+i);
          i = j + 1;
        }
      }
      if (i < j)
        AppendStatusInfoSub (hWnd, pStatusWndInfo, Type, aszBuf+i);
    }
    else
      AppendStatusInfoSub (hWnd, pStatusWndInfo, Type, aszBuf);
  }
  else
  {
    // OutMsgProc(HWND hWnd, OMPT Type, NULL, PCCHAR pcszIndicator,
    //            PCBYTE pcbBuf, DWORD dwSize);
    pcszIndicator = va_arg (ArgList, PCCHAR);
    pcbBuf        = va_arg (ArgList, PCBYTE);
    dwSize        = va_arg (ArgList, DWORD);

    for (i=0; i<dwSize; i++)
    {
      if (pcbBuf[i]>=0x20 && pcbBuf[i]<=0x7f)
        aszASCIIView[i%16] = pcbBuf[i];
      else
        aszASCIIView[i%16] = '.';

      if ((i%16) == 0)
        sprintf (aszBuf, "%s%02X", pcszIndicator, pcbBuf[i]);
      else if ((i%16) == 8)
      {
        sprintf (aszBuf2, " - %02X", pcbBuf[i]);
        strcat (aszBuf, aszBuf2);
      }
      else
      {
        sprintf (aszBuf2, " %02X", pcbBuf[i]);
        strcat (aszBuf, aszBuf2);
      }
      if ((i%16) == 15)
      {
        strcat (aszBuf, "    ");
        aszASCIIView[16] = 0;
        strcat (aszBuf, aszASCIIView);
        AppendStatusInfoSub (hWnd, pStatusWndInfo, Type, aszBuf);
      }
    }
    i %= 16;
    if (i)
    {
      for (j=i; j<16; j++)
      {
        if ((j%16) == 8)
          strcat (aszBuf, "     ");
        else
          strcat (aszBuf, "   ");
      }
      strcat (aszBuf, "    ");
      aszASCIIView[i] = '\0';
      strcat (aszBuf, aszASCIIView);
      AppendStatusInfoSub (hWnd, pStatusWndInfo, Type, aszBuf);
    }
  }
  InvalidateRect (hWnd, NULL, FALSE);
  return 0;
}
/****************************************************************************************/
SOCKTOOLAPI WORD STATWND_AppendInfo(HWND hWnd, OMPT Type, PCCHAR pcszFormatString, ...)
{
  WORD            wResult;
  va_list         ArgList;
  
  va_start (ArgList, pcszFormatString);   
  wResult = STATWND_AppendInfoV(hWnd, Type, pcszFormatString, ArgList);
  va_end (ArgList);

  return wResult;
}
/****************************************************************************************/
SOCKTOOLAPI WORD PASCAL STATWND_StreamByte(HWND hWnd, OMPT Type, BYTE bVal)
{
  PSTATWNDINFO    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
  INFOITEM       *pTempInfo;
  DWORD           i, j;

  if (!(pStatusWndInfo->Flags & STFLAG_INSPECTSTREAM)) return 0;
  if (Type != OMPT_INFO && Type != OMPT_UPLOAD && Type != OMPT_DOWNLOAD) return 0;

  if (pStatusWndInfo->bCacheSize == 0 || pStatusWndInfo->LineType != Type)
  {
    AppendStatusInfoSub (hWnd, pStatusWndInfo, Type,
                         "                                   "
                         "                                  ");
    pStatusWndInfo->bCacheSize = 0;
  }
  pStatusWndInfo->abStreamCache[pStatusWndInfo->bCacheSize++] = bVal;
  pStatusWndInfo->LineType = Type;
  
  if (pStatusWndInfo->pStatusInfo == NULL) return 0;
  pTempInfo = pStatusWndInfo->pStatusInfo;
  while (pTempInfo->pNextItem) pTempInfo = pTempInfo->pNextItem;
  if (pTempInfo == NULL) return 0;
  for (i=0, j=0; i<16; i++)
  {
    if (i != 0)
      pTempInfo->pszContent[j++] = ' ';
    if (i == 8)
    {
      pTempInfo->pszContent[j++] = '-';
      pTempInfo->pszContent[j++] = ' ';
    }
    if (i<pStatusWndInfo->bCacheSize)
      j += wsprintf (pTempInfo->pszContent+j, "%02X", pStatusWndInfo->abStreamCache[i]);
    else
      j += wsprintf (pTempInfo->pszContent+j, "  ");
  }
  pTempInfo->pszContent[j++] = ' ';
  pTempInfo->pszContent[j++] = ' ';
  pTempInfo->pszContent[j++] = ' ';
  pTempInfo->pszContent[j++] = ' ';
  for (i=0; i<16; i++)
  {
    if (i<pStatusWndInfo->bCacheSize)
    {
      if (pStatusWndInfo->abStreamCache[i]>=0x20 && pStatusWndInfo->abStreamCache[i]<=0x7f)
        pTempInfo->pszContent[j++] = pStatusWndInfo->abStreamCache[i];
      else
        pTempInfo->pszContent[j++] = '.';
    }
    else
      break;
  }
  pTempInfo->pszContent[j++] = '\0';

  i = pStatusWndInfo->pCurStatusInfo->dwLineNo;
  j = pStatusWndInfo->pCurStatusInfo->dwLineNo+pStatusWndInfo->wStatusWndTextLines;
  if (pTempInfo->dwLineNo >= i && pTempInfo->dwLineNo<j)
    InvalidateRect (hWnd, NULL, FALSE);

  if (pStatusWndInfo->bCacheSize == 16) pStatusWndInfo->bCacheSize = 0;

  return 0;
}
/****************************************************************************************/
SOCKTOOLAPI void PASCAL STATWND_ClearInfo(HWND hWnd)
{
  PSTATWNDINFO    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
  SCROLLINFO      ScrollInfo;

  while (pStatusWndInfo->pStatusInfo)
  {
    pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pStatusInfo;
    pStatusWndInfo->pStatusInfo = pStatusWndInfo->pStatusInfo->pNextItem;
    free (pStatusWndInfo->pCurStatusInfo->pszContent);
    free (pStatusWndInfo->pCurStatusInfo);
  }

  ScrollInfo.cbSize = sizeof (SCROLLINFO);
  ScrollInfo.fMask  = SIF_PAGE | SIF_POS | SIF_RANGE;
  ScrollInfo.nMin   = 0;
  ScrollInfo.nMax   = pStatusWndInfo->dwTotalLine - pStatusWndInfo->wStatusWndTextLines;
  ScrollInfo.nPage  = 1;
  ScrollInfo.nPos   = 0;
  SetScrollInfo (hWnd, SB_VERT, &ScrollInfo, TRUE);
  InvalidateRect (hWnd, NULL, FALSE);

  pStatusWndInfo->dwTotalLine = 0;
  pStatusWndInfo->pStatusInfo = pStatusWndInfo->pCurStatusInfo = NULL;
}
/****************************************************************************************/
SOCKTOOLAPI void PASCAL STATWND_CopyInfoToClipboard(HWND hWnd)
{
  PSTATWNDINFO    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
  INFOITEM       *pInfoItem;
  DWORD           dwInfoSize;
  HGLOBAL         hInfoMem;
  PCHAR           pInfoMem;

  pInfoItem = pStatusWndInfo->pStatusInfo;
  dwInfoSize = 0;
  while (pInfoItem)
  {
    dwInfoSize += strlen (pInfoItem->pszContent)+
                  (pStatusWndInfo->Flags & STFLAG_IGNORETYPE ? 2 : 4);
    pInfoItem = pInfoItem->pNextItem;
  }
  if (dwInfoSize == 0) return;
  dwInfoSize += 1;
   
  hInfoMem = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, dwInfoSize);
  if (hInfoMem)
  {
    pInfoMem = (PCHAR)GlobalLock (hInfoMem);
    pInfoItem = pStatusWndInfo->pStatusInfo;
    dwInfoSize = 0;
    while (pInfoItem)
    {
      if ( !(pStatusWndInfo->Flags & STFLAG_IGNORETYPE))
      {
        switch (pInfoItem->Type)
        {
        case OMPT_SUCCESS:  pInfoMem[dwInfoSize] = 'S'; break;
        case OMPT_ERROR:    pInfoMem[dwInfoSize] = 'E'; break;
        case OMPT_INFO:     pInfoMem[dwInfoSize] = 'I'; break;
        case OMPT_UPLOAD:   pInfoMem[dwInfoSize] = 'U'; break;
        case OMPT_DOWNLOAD: pInfoMem[dwInfoSize] = 'D'; break;
        }
        pInfoMem[dwInfoSize+1] = '>';
        dwInfoSize += 2;
      }
      strcpy (pInfoMem+dwInfoSize, pInfoItem->pszContent);
      dwInfoSize += 2 + strlen (pInfoItem->pszContent);
      pInfoMem[dwInfoSize-2] = '\r';
      pInfoMem[dwInfoSize-1] = '\n';
      pInfoItem = pInfoItem->pNextItem;
    }
    pInfoMem[dwInfoSize] = '\0';
    GlobalUnlock (pInfoMem);

    OpenClipboard (NULL);
    SetClipboardData (CF_TEXT, hInfoMem);
    CloseClipboard ();
  }
}
/****************************************************************************************/
