/*****************************************************************************************
 * File name     : UI.C
 * Author        : Skeleton Key
 * Module        : 
 * Purpose       : 
 * Create Date   : June 8, 2001 - 10:10
 * Modify history: 
 *****************************************************************************************
 */
#define ALLOCATE_UI_VARIABLE_MEMORY

#include <windows.h>
#include <commctrl.h>
#include <stdarg.h>
#include <htmlhelp.h>
#include <time.h>
#include "passwd.h"
#include "io.h"
#include "resource.h"
#include "ui.h"
#include "hiswnd.h"
#include "comm.h"
#include "stdio.h"
#include "bmptool.h"

/*****************************************************************************************
 * MACRO functions declare
 ****************************************************************************************/

/*****************************************************************************************
 * Identifier constant declare
 ****************************************************************************************/
#define MAINWNDCLASSNAME                "YBD Repeater controller"
#define HISTORYWNDCLASSNAME             "YBD Access History"
#define REGKEYYBDREPEATOR               "SOFTWARE\\YinBoDa\\Repeater"
#define REGKEYMODEMPARAM                REGKEYYBDREPEATOR "\\MODEM"
#define REGKEYZFZLIST                   REGKEYYBDREPEATOR "\\Repeater List"
#define VIEWWNDCOL                      7
#define STATUSBARPARTS                  3

#define MAXZFZPARAMS                    16384

#define IDS_MODEM_FIRST                 IDS_MODEM_DATAMODE
#define IDS_HEADERCOL_FIRST             IDS_HEADERCOL_NAME
#define IDS_TASK_FIRST                  IDS_TASK_SETPHONE1

#define HISTORYHEADER                   "This is the log file for repeater control centre."\
                                        "\x1a""YBD0001"
#define HISTORYFILENAME                 "history.log"

/*****************************************************************************************
 * Data structure declare
 ****************************************************************************************/

/*****************************************************************************************
 * Local variable declare
 ****************************************************************************************/
static HWND       ghTerminateWnd;
static HWND       ghViewWnd;
static HWND       ghStatusWnd;
static HWND       ghActivateDialog;

static HWND       ghStatusDialog[MAXZFZPARAMS];
static WORD       gwStatusDialogs;
static BOOL       gfStatusDialogEnable = TRUE;

static CHAR       gaszString[250];
static CHAR       gaszString2[250];

static CHAR       gaszDlgRAMPhone[ZFZPHONELEN+1];
static CHAR       gaszDlgZFZName[ZFZNAMELEN+1];
static CHAR       gaszDlgZFZRemark[ZFZREMARKLEN+1];
static PZFZPARAM  gpDlgZFZParam;

static DWORD      gdwZFZParams;
static PZFZPARAM  gapZFZParam[MAXZFZPARAMS];
static PZFZPARAM  gpDirectConnectZFZParam;

static PZFZPARAM  gpAutoPopupParam;
static DWORD      gdwTaskNodeIndex;

static HMODULE    ghDebugWndDll;

static URM_USERRIGHT gUserRight = UR_NONE;
static CHAR       gaszLoginUser[URM_MAXNAMELEN+1];

static const int STATUSBARPARTWIDTH[STATUSBARPARTS] = { 150, 250, -1 };
static const int VIEWWNDCOLWIDTH[VIEWWNDCOL] = { 180, 150, 50, 120, 120, 60, 100 };
static const int VIEWWNDCOLFORMAT[VIEWWNDCOL] = { LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_RIGHT,
                                                  LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_LEFT,
                                                  LVCFMT_LEFT
                                                };

/*****************************************************************************************
 * Local functions declare
 ****************************************************************************************/
DWORD ZFZPARAM_SizeOf(BYTE bChannels);
PZFZPARAM ZFZPARAM_malloc(BYTE bChannels);

/****************************************************************************************/
/****************************************************************************************/
/****************************************************************************************/
void AddWindowHandle(HWND hWnd)
{
  WORD            i;

  for (i=0; i<gwStatusDialogs && ghStatusDialog[i]!=hWnd; i++);
  if (i<gwStatusDialogs) return;
  if (i<MAXZFZPARAMS)
    ghStatusDialog[gwStatusDialogs++] = hWnd;
}
/****************************************************************************************/
void RemoveWindowHandle(HWND hWnd)
{
  WORD            i;

  for (i=0; i<gwStatusDialogs && ghStatusDialog[i]!=hWnd; i++);
  if (i<gwStatusDialogs)
  {
    gwStatusDialogs --;
    for (; i<gwStatusDialogs; i++)
      ghStatusDialog[i] = ghStatusDialog[i+1];
  }
}
/****************************************************************************************/
void EnableAllWindow(BOOL fEnable)
{
  WORD            i;

  gfStatusDialogEnable = fEnable;
  for (i=0; i<gwStatusDialogs; i++)
  {
    ShowWindow (ghStatusDialog[i], SW_SHOWNORMAL);
    EnableWindow (ghStatusDialog[i], fEnable);
  }
}
/****************************************************************************************/
/****************************************************************************************/
BOOL GetWindowPosition(PRECT pWndPos, BOOL *pfIsZoomed)
{
  HKEY            hKey;
  DWORD           Result, dwLen, dwType;

  if (pWndPos == NULL) return FALSE;

  if (ERROR_SUCCESS == RegOpenKey (HKEY_LOCAL_MACHINE, REGKEYYBDREPEATOR, &hKey))
  {
    dwLen = sizeof *pWndPos;
    Result = RegQueryValueEx (hKey, "Position", 0, &dwType, (PBYTE)pWndPos, &dwLen);
    if (Result != ERROR_SUCCESS) goto SetDefault;
    dwLen = sizeof *pfIsZoomed;
    Result = RegQueryValueEx (hKey, "Maximized", 0, &dwType, (PBYTE)pfIsZoomed, &dwLen);
    if (Result != ERROR_SUCCESS) goto SetDefault;
    return TRUE;
  }
SetDefault:
  pWndPos->top    = 0;
  pWndPos->left   = 0;
  pWndPos->right  = GetSystemMetrics (SM_CXFULLSCREEN);
  pWndPos->bottom = GetSystemMetrics (SM_CYFULLSCREEN)+20;
  *pfIsZoomed     = FALSE;
  return FALSE;
}
/****************************************************************************************/
BOOL SetWindowPosition(LPCRECT pWndPos, BOOL fIsZoomed)
{
  HKEY            hKey;
  DWORD           Result;

  if (pWndPos == NULL) return FALSE;

  if (ERROR_SUCCESS == RegCreateKey (HKEY_LOCAL_MACHINE, REGKEYYBDREPEATOR, &hKey))
  {
    if (!fIsZoomed)
    {
      Result = RegSetValueEx (hKey, "Position", 0, REG_BINARY,
                              (PBYTE)pWndPos, sizeof *pWndPos);
      if (Result != ERROR_SUCCESS) return FALSE;
    }
    Result = RegSetValueEx (hKey, "Maximized", 0, REG_BINARY,
                            (PBYTE)&fIsZoomed, sizeof fIsZoomed);
    if (Result != ERROR_SUCCESS) return FALSE;
  }

  return TRUE;
}
/****************************************************************************************/
/****************************************************************************************/
BOOL GetModemParam(PMODEMPARAM pModemParam)
{
  HKEY            hKey;
  DWORD           Result, dwLen, dwType;

  if (pModemParam == NULL) return FALSE;

  if (ERROR_SUCCESS == RegOpenKey (HKEY_LOCAL_MACHINE, REGKEYMODEMPARAM, &hKey))
  {
    dwLen = sizeof pModemParam->wCommPort;
    Result = RegQueryValueEx (hKey, "Serial Port", 0, &dwType, 
                              (PBYTE)&pModemParam->wCommPort, &dwLen);
    dwLen = sizeof pModemParam->aszInit;
    Result = RegQueryValueEx (hKey, "Initialize", 0, &dwType,
                              pModemParam->aszInit, &dwLen);
    if (Result != ERROR_SUCCESS) goto SetDefault;
    dwLen = sizeof pModemParam->aszDial;
    Result = RegQueryValueEx (hKey, "Dial", 0, &dwType,
                              pModemParam->aszDial, &dwLen);
    if (Result != ERROR_SUCCESS) goto SetDefault;
    dwLen = sizeof pModemParam->aszAnswer;
    Result = RegQueryValueEx (hKey, "Answer", 0, &dwType,
                              pModemParam->aszAnswer, &dwLen);
    if (Result != ERROR_SUCCESS) goto SetDefault;
    dwLen = sizeof pModemParam->aszHangup;
    Result = RegQueryValueEx (hKey, "Hangup", 0, &dwType,
                              pModemParam->aszHangup, &dwLen);
    if (Result != ERROR_SUCCESS) goto SetDefault;
    dwLen = sizeof pModemParam->aszToCMD;
    Result = RegQueryValueEx (hKey, "To Command Mode", 0, &dwType,
                              pModemParam->aszToCMD, &dwLen);
    if (Result != ERROR_SUCCESS) goto SetDefault;
    return TRUE;
  }
SetDefault:
  pModemParam->wCommPort = 1;
  strcpy (pModemParam->aszInit, "ATZE0V1");
  strcpy (pModemParam->aszDial, "ATDT");
  strcpy (pModemParam->aszAnswer, "ATA");
  strcpy (pModemParam->aszHangup, "ATH");
  strcpy (pModemParam->aszToCMD, "+++");
  return FALSE;
}
/****************************************************************************************/
BOOL SetModemParam(PCMODEMPARAM pcModemParam)
{
  HKEY            hKey;
  DWORD           Result;

  if (pcModemParam == NULL) return FALSE;

  if (ERROR_SUCCESS == RegCreateKey (HKEY_LOCAL_MACHINE, REGKEYMODEMPARAM, &hKey))
  {
    Result = RegSetValueEx (hKey, "Serial Port", 0, REG_BINARY,
                            (PBYTE)&pcModemParam->wCommPort,
                            sizeof pcModemParam->wCommPort);
    if (Result != ERROR_SUCCESS) return FALSE;
    Result = RegSetValueEx (hKey, "Initialize", 0, REG_SZ,
                            pcModemParam->aszInit, strlen (pcModemParam->aszInit)+1);
    if (Result != ERROR_SUCCESS) return FALSE;
    Result = RegSetValueEx (hKey, "Dial", 0, REG_SZ,
                            pcModemParam->aszDial, strlen (pcModemParam->aszDial)+1);
    if (Result != ERROR_SUCCESS) return FALSE;
    Result = RegSetValueEx (hKey, "Answer", 0, REG_SZ,
                            pcModemParam->aszAnswer, strlen (pcModemParam->aszAnswer)+1);
    if (Result != ERROR_SUCCESS) return FALSE;
    Result = RegSetValueEx (hKey, "Hangup", 0, REG_SZ,
                            pcModemParam->aszHangup, strlen (pcModemParam->aszHangup)+1);
    if (Result != ERROR_SUCCESS) return FALSE;
    Result = RegSetValueEx (hKey, "To Command Mode", 0, REG_SZ,
                            pcModemParam->aszToCMD, strlen (pcModemParam->aszToCMD)+1);
    if (Result != ERROR_SUCCESS) return FALSE;
  }

  return TRUE;
}
/****************************************************************************************/
/****************************************************************************************/
BOOL LoadZFZList(void)
{
  HKEY            hKey;
  DWORD           Result, dwLen, dwType, dwVersion;
  CHAR            aszBuf[100];
  DWORD           i;
  PZFZPARAM       pZFZParam = NULL;

  if (ERROR_SUCCESS == RegOpenKey (HKEY_LOCAL_MACHINE, REGKEYZFZLIST, &hKey))
  {
    dwLen = sizeof dwVersion;
    Result = RegQueryValueEx (hKey, "Version", 0, &dwType, (PBYTE)&dwVersion, &dwLen);
    if (Result != ERROR_SUCCESS) goto Error;
    if (dwVersion == 0x80000001)
    {
      pZFZParam = ZFZPARAM_malloc (255);
      if (pZFZParam == NULL) goto Error;
      dwLen = sizeof gaszCentrePhone;
      Result = RegQueryValueEx (hKey, "Centre phone", 0, &dwType, gaszCentrePhone, &dwLen);
      if (Result != ERROR_SUCCESS) goto Error;
      dwLen = sizeof gdwZFZParams;
      Result = RegQueryValueEx (hKey, "Repeaters", 0, &dwType, (PBYTE)&gdwZFZParams, &dwLen);
      if (Result != ERROR_SUCCESS) goto Error;
      for (i=0; i<gdwZFZParams; i++)
      {
        wsprintf (aszBuf, "Repeater%04X", i);
        dwLen = ZFZPARAM_SizeOf (255);
        Result = RegQueryValueEx (hKey, aszBuf, 0, &dwType, (PBYTE)pZFZParam, &dwLen);
        if (Result != ERROR_SUCCESS) { gdwZFZParams = i; goto Error; }
        gapZFZParam[i] = ZFZPARAM_malloc (pZFZParam->bChannels);
        if (gapZFZParam[i] == NULL) { gdwZFZParams = i; goto Error; }
        memcpy (gapZFZParam[i], pZFZParam, ZFZPARAM_SizeOf (pZFZParam->bChannels));
      }
      free (pZFZParam);
      return TRUE;
    }
  }
Error:
  if (pZFZParam) free (pZFZParam);
  return FALSE;
}
/****************************************************************************************/
BOOL SaveZFZList(void)
{
  HKEY            hKey;
  DWORD           Result;
  CHAR            aszBuf[100];
  DWORD           i;

  RegDeleteKey (HKEY_LOCAL_MACHINE, REGKEYZFZLIST);
  if (ERROR_SUCCESS == RegCreateKey (HKEY_LOCAL_MACHINE, REGKEYZFZLIST, &hKey))
  {
    Result = RegSetValueEx (hKey, "Centre phone", 0, REG_SZ,
                            gaszCentrePhone, strlen (gaszCentrePhone)+1);
    if (Result != ERROR_SUCCESS) return FALSE;
    Result = RegSetValueEx (hKey, "Repeaters", 0, REG_DWORD, (PBYTE)&gdwZFZParams, 4);
    if (Result != ERROR_SUCCESS) return FALSE;
    Result = 0x80000001l;
    Result = RegSetValueEx (hKey, "Version", 0, REG_DWORD, (PBYTE)&Result, 4);
    if (Result != ERROR_SUCCESS) return FALSE;

    for (i=0; i<gdwZFZParams; i++)
    {
      wsprintf (aszBuf, "Repeater%04X", i);
      Result = RegSetValueEx (hKey, aszBuf, 0, REG_BINARY, (PBYTE)gapZFZParam[i],
                              ZFZPARAM_SizeOf(gapZFZParam[i]->bChannels));
      if (Result != ERROR_SUCCESS) return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}
/****************************************************************************************/
void SaveControlHistory(DWORD dwTime, PZFZPARAM pZFZParam)
{
  FILE           *fp;
  BOOLEAN         fFileExist;
  WORD            wSize;

  fFileExist = (access (HISTORYFILENAME, 0) == -1);
  fp = fopen (HISTORYFILENAME, "ab");
  if (fp != NULL)
  {
    if (fFileExist)
      fwrite (HISTORYHEADER, 1, sizeof HISTORYHEADER, fp);

    fwrite (&dwTime, 1, sizeof dwTime, fp);
    wSize = (WORD)ZFZPARAM_SizeOf (pZFZParam->bChannels);
    fwrite (&wSize, 1, sizeof wSize, fp);
    fwrite (pZFZParam, 1, wSize, fp);
    fclose (fp);
  }
}
/****************************************************************************************/
/****************************************************************************************/
void GetTaskString(PCHAR pszBuf, TASKID TaskID)
{
  WORD            i, j;
  TASKID          TempTaskID;

  pszBuf[0] = 0;
  for (TempTaskID=1, i=0, j=0; i<11; i++)
  {
    if (TempTaskID & TaskID)
    {
      LoadString (ghInst, IDS_TASK_FIRST+i, gaszString2, sizeof gaszString2);
      if (j) strcat (pszBuf, ", ");
      j = 1;
      strcat (pszBuf, gaszString2);
    }
    TempTaskID <<= 1;
  }
}
/****************************************************************************************/
/****************************************************************************************/
void STATUSBAR_ShowModemStatus(WORD wOffset, ...)
{
  CHAR            aszModemMessage[200];
  va_list         Arguments;

  LoadString (ghInst, IDS_MODEM_FIRST+wOffset, gaszString, sizeof gaszString);
  Arguments = va_start (Arguments, wOffset);
  wvsprintf (aszModemMessage, gaszString, Arguments);
  va_end (Arguments);

  SendMessage (ghStatusWnd, SB_SETTEXT, 0, (LPARAM)aszModemMessage);
}
/****************************************************************************************/
void STATUSBAR_ShowTaskStatus(PCSTR pcszZFZName, TASKID ZFZTaskID)
{
  if (ZFZTaskID & ((1<<11)-1))
  {
    strcpy (gaszString, pcszZFZName);
    strcat (gaszString, ":");
    GetTaskString (gaszString+strlen (gaszString), ZFZTaskID);
  }
  else
    LoadString (ghInst, IDS_TASKNONE, gaszString, sizeof gaszString);
  SendMessage (ghStatusWnd, SB_SETTEXT, 2, (LPARAM)gaszString);
}
/****************************************************************************************/
/****************************************************************************************/
DWORD ZFZPARAM_SizeOf(BYTE bChannels)
{
  if (bChannels == 0)
    return sizeof (ZFZPARAM) - sizeof (ZFZCHANNEL);
  else if (bChannels == 1)
    return sizeof (ZFZPARAM);
  else
    return sizeof (ZFZPARAM) + (bChannels-1)*sizeof (ZFZCHANNEL);
}
/****************************************************************************************/
PZFZPARAM ZFZPARAM_malloc(BYTE bChannels)
{
  DWORD           dwSize;
  PZFZPARAM       pZFZParam;

  dwSize = ZFZPARAM_SizeOf (bChannels);
  pZFZParam = (PZFZPARAM)malloc (dwSize);
  if (pZFZParam) memset (pZFZParam, 0, dwSize);
  return pZFZParam;
}
/****************************************************************************************/
WORD ZFZPARAM_SearchByName(PCSTR pcszName)
{
  WORD             i;

  for (i=0; i<gdwZFZParams; i++)
    if (strcmp (pcszName, gapZFZParam[i]->aszName) == 0) break;
  if (i<gdwZFZParams)
    return i;
  else
    return 0xffff;
}
/****************************************************************************************/
PCSTR ZFZPARAM_GenerateName(void)
{
  WORD            i;

  LoadString (ghInst, IDS_DEFAULPREFIXNAME, gaszString, sizeof gaszString);
  for (i=1; ; i++)
  {
    wsprintf (gaszString2, "%s %u", gaszString, i);
    if (ZFZPARAM_SearchByName (gaszString2) == 0xffff) break;
  }
  return gaszString2;
}
/****************************************************************************************/
int CompareDate(PZFZDATE pDate1, PZFZDATE pDate2)
{
  int iResult;
  iResult = pDate1->wYear - pDate2->wYear;
  if (iResult != 0) return iResult;
  iResult = pDate1->bMonth - pDate2->bMonth;
  if (iResult != 0) return iResult;
  iResult = pDate1->bDay - pDate2->bDay;
  if (iResult != 0) return iResult;
  iResult = pDate1->bHour - pDate2->bHour;
  if (iResult != 0) return iResult;
  iResult = pDate1->bMinute - pDate2->bMinute;
  if (iResult != 0) return iResult;
  return 0;
}
/****************************************************************************************/
int CALLBACK ZFZPARAM_Compare(PZFZPARAM pZFZParam1, PZFZPARAM pZFZParam2, LPARAM iSubItem)
{
  switch (iSubItem)
  {
  case 0:  return strcmp (pZFZParam1->aszName, pZFZParam2->aszName);
  case 1:  return strcmp (pZFZParam1->aszPhone, pZFZParam2->aszPhone);
  case 2:  return pZFZParam1->bChannels - pZFZParam2->bChannels;
  case 3:  return CompareDate (&pZFZParam1->CreateDate, &pZFZParam2->CreateDate);
  case 4:  return CompareDate (&pZFZParam2->LastAlarmDate, &pZFZParam1->LastAlarmDate);
  case 5:  return pZFZParam2->fDisableAlarm - pZFZParam1->fDisableAlarm;
  case 6:  return strcmp (pZFZParam1->aszRemark, pZFZParam2->aszRemark);
  default: return 0;
  }
}
/****************************************************************************************/
/****************************************************************************************/
LPSTR LISTVIEW_ItemContent(int iSubItem, PZFZPARAM pZFZParam)
{
  switch (iSubItem)
  {
  case 0:
    return pZFZParam->aszName;
  case 1:
    return pZFZParam->aszPhone;
  case 2:
    wsprintf (gaszString, "%d", pZFZParam->bChannels);
    return gaszString;
  case 3:
    wsprintf (gaszString, "%4u-%u-%u %02u:%02u", pZFZParam->CreateDate.wYear,
                                                 pZFZParam->CreateDate.bMonth,
                                                 pZFZParam->CreateDate.bDay,
                                                 pZFZParam->CreateDate.bHour,
                                                 pZFZParam->CreateDate.bMinute);
    return gaszString;
  case 4:
    if (pZFZParam->LastAlarmDate.wYear)
      wsprintf (gaszString, "%4u-%u-%u %02u:%02u", pZFZParam->LastAlarmDate.wYear,
                                                   pZFZParam->LastAlarmDate.bMonth,
                                                   pZFZParam->LastAlarmDate.bDay,
                                                   pZFZParam->LastAlarmDate.bHour,
                                                   pZFZParam->LastAlarmDate.bMinute);
    else
      LoadString (ghInst, IDS_NEVERALARM, gaszString, sizeof gaszString);
    return gaszString;
  case 5:
    if (pZFZParam->fDisableAlarm)
      LoadString (ghInst, IDS_TURNOFFALARM, gaszString, sizeof gaszString);
    else
      LoadString (ghInst, IDS_TURNONALARM, gaszString, sizeof gaszString);
    return gaszString;
  case 6:
    return pZFZParam->aszRemark;
  }
  return NULL;
}
/****************************************************************************************/
int LISTVIEW_AddItem(HWND hViewWnd, int iItemIndex, PZFZPARAM pZFZParam)
{
  LVITEM          lvi;

  lvi.mask      = LVIF_TEXT|LVIF_IMAGE|LVIF_PARAM|LVIF_STATE;
  lvi.state     = 0;
  lvi.stateMask = 0;
  lvi.iItem     = iItemIndex;
  lvi.iSubItem  = 0; 
  lvi.iImage    = pZFZParam->VWImageIndex;
  lvi.pszText   = LPSTR_TEXTCALLBACK;
  lvi.lParam    = (LPARAM)pZFZParam;
  return ListView_InsertItem (hViewWnd, &lvi); 
}
/****************************************************************************************/
BOOL LISTVIEW_GetZFZParam(HWND hViewWnd, int iItemIndex, PZFZPARAM *ppZFZParam)
{
  LVITEM lvi;

  lvi.mask = LVIF_PARAM;
  lvi.iItem  = iItemIndex;
  lvi.iSubItem = 0;
  if (ListView_GetItem (ghViewWnd, &lvi))
  {
    *ppZFZParam = (PZFZPARAM)lvi.lParam;
    return TRUE;
  }
  else
    return FALSE;
}
/****************************************************************************************/
/****************************************************************************************/
static void CentreDialog(HWND hWnd)
{
  RECT            WndRect;
  int             x, y, w, h;

  GetWindowRect (hWnd, &WndRect);
  w = WndRect.right - WndRect.left;
  h = WndRect.bottom - WndRect.top;
  x = (GetSystemMetrics (SM_CXFULLSCREEN)-w)/2;
  y = (GetSystemMetrics (SM_CYFULLSCREEN)+20-h)/2;
  MoveWindow (hWnd, x, y, w, h, TRUE);
}
/****************************************************************************************/
BOOL ValidateField(HWND hDlg, WORD wCtrlID, LONG liMin, LONG liMax)
{
  CHAR            aszBuf[10];
  LONG            liValue;

  GetWindowText (GetDlgItem (hDlg, wCtrlID), aszBuf, sizeof aszBuf);
  liValue = atol (aszBuf);
  if (liValue>=liMin && liValue<=liMax)
    return TRUE;
  else
  {
    LoadString (ghInst, IDS_INVALIDVALUE, gaszString2, sizeof gaszString2);
    wsprintf (gaszString, gaszString2, liValue, liMin, liMax);
    LoadString (ghInst, IDS_PARAMETER_TITLE, gaszString2, sizeof gaszString2);
    MessageBox (hDlg, gaszString, gaszString2, MB_OK|MB_ICONERROR);
    SetFocus (GetDlgItem (hDlg, wCtrlID));
    SendDlgItemMessage (hDlg, wCtrlID, EM_SETSEL, 0, -1);
    return FALSE;
  }
}
/****************************************************************************************/
void EnableZFZMenu(void)
{
  HMENU           hMenu = GetSubMenu (GetMenu (ghMainWnd), 0);

  if (gpDirectConnectZFZParam)
  {
    EnableMenuItem (hMenu, ID_ZFZ_NEW, MF_GRAYED);
    EnableMenuItem (hMenu, ID_ZFZ_DELETE, MF_GRAYED);
    EnableMenuItem (hMenu, ID_ZFZ_GETSTATUS, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_DISPSTATUS, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_RUNPARAM, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_TURNON, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_TURNOFF, MF_ENABLED);
//    EnableMenuItem (hMenu, ID_ZFZ_HISTORY, MF_GRAYED);
    EnableMenuItem (hMenu, ID_ZFZ_PROPERTIES, MF_ENABLED);
  }
  else if (ListView_GetItemCount (ghViewWnd))
  {
    EnableMenuItem (hMenu, ID_ZFZ_NEW, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_DELETE, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_GETSTATUS, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_DISPSTATUS, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_RUNPARAM, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_TURNON, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_TURNOFF, MF_ENABLED);
//    EnableMenuItem (hMenu, ID_ZFZ_HISTORY, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_PROPERTIES, MF_ENABLED);
  }
  else
  {
    EnableMenuItem (hMenu, ID_ZFZ_NEW, MF_ENABLED);
    EnableMenuItem (hMenu, ID_ZFZ_DELETE, MF_GRAYED);
    EnableMenuItem (hMenu, ID_ZFZ_GETSTATUS, MF_GRAYED);
    EnableMenuItem (hMenu, ID_ZFZ_DISPSTATUS, MF_GRAYED);
    EnableMenuItem (hMenu, ID_ZFZ_RUNPARAM, MF_GRAYED);
    EnableMenuItem (hMenu, ID_ZFZ_TURNON, MF_GRAYED);
    EnableMenuItem (hMenu, ID_ZFZ_TURNOFF, MF_GRAYED);
//    EnableMenuItem (hMenu, ID_ZFZ_HISTORY, MF_GRAYED);
    EnableMenuItem (hMenu, ID_ZFZ_PROPERTIES, MF_GRAYED);
  }
}
/****************************************************************************************/
/****************************************************************************************/
static BOOL CALLBACK ZFZ_NewDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    SetWindowText (GetDlgItem (hDlg, IDC_ZFZ_NAME), ZFZPARAM_GenerateName ());
    SendDlgItemMessage (hDlg, IDC_ZFZ_NAME, EM_LIMITTEXT, ZFZNAMELEN, 0);
    SendDlgItemMessage (hDlg, IDC_ZFZ_PHONENO, EM_LIMITTEXT, ZFZPHONELEN, 0);
    SendDlgItemMessage (hDlg, IDC_ZFZ_REMARK, EM_LIMITTEXT, ZFZREMARKLEN, 0);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDOK:
      GetWindowText (GetDlgItem (hDlg, IDC_ZFZ_NAME), gaszDlgZFZName, sizeof gaszDlgZFZName);
      if (gaszDlgZFZName[0] == 0)
      {
        SetFocus (GetDlgItem (hDlg, IDC_ZFZ_NAME));
        return TRUE;
      }
      GetWindowText (GetDlgItem (hDlg, IDC_ZFZ_PHONENO),
                     gaszDlgRAMPhone, sizeof gaszDlgRAMPhone);
      if (gaszDlgRAMPhone[0] == 0)
      {
        SetFocus (GetDlgItem (hDlg, IDC_ZFZ_PHONENO));
        return TRUE;
      }
      GetWindowText (GetDlgItem (hDlg, IDC_ZFZ_REMARK),
                     gaszDlgZFZRemark, sizeof gaszDlgZFZRemark);
      if (ZFZPARAM_SearchByName (gaszDlgZFZName) != 0xffff)
      {
        LoadString (ghInst, IDS_DUPLICATEZFZNAME, gaszString, sizeof gaszString);
        LoadString (ghInst, IDS_PARAMETER_TITLE, gaszString2, sizeof gaszString2);
        MessageBox (NULL, gaszString, gaszString2, MB_OK|MB_ICONERROR);
        return TRUE;
      }
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
void ZFZ_NewDialog(void)
{
  SYSTEMTIME      LocalTime;

  if (DialogBox (ghInst, MAKEINTRESOURCE (IDD_NEW), ghMainWnd, ZFZ_NewDlgProc) == IDOK)
  {
    if (gdwZFZParams >= MAXZFZPARAMS) return;
    gapZFZParam[gdwZFZParams] = ZFZPARAM_malloc (0);
    if (gapZFZParam[gdwZFZParams] == NULL) return;
    strcpy (gapZFZParam[gdwZFZParams]->aszName, gaszDlgZFZName);
    strcpy (gapZFZParam[gdwZFZParams]->aszPhone, gaszDlgRAMPhone);
    strcpy (gapZFZParam[gdwZFZParams]->aszRemark, gaszDlgZFZRemark);
    GetLocalTime (&LocalTime);
    gapZFZParam[gdwZFZParams]->CreateDate.wYear   = LocalTime.wYear;
    gapZFZParam[gdwZFZParams]->CreateDate.bMonth  = (BYTE)LocalTime.wMonth;
    gapZFZParam[gdwZFZParams]->CreateDate.bDay    = (BYTE)LocalTime.wDay;
    gapZFZParam[gdwZFZParams]->CreateDate.bHour   = (BYTE)LocalTime.wHour;
    gapZFZParam[gdwZFZParams]->CreateDate.bMinute = (BYTE)LocalTime.wMinute;
    gapZFZParam[gdwZFZParams]->VWImageIndex       = VWI_INSTALL;
    gapZFZParam[gdwZFZParams]->TaskID = TASKID_SETPHONE1 | TASKID_SETPHONE2 |
                                        TASKID_GETPROPERTIES | TASKID_GETSTATUS |
                                        TASKID_ALARM;
    LISTVIEW_AddItem (ghViewWnd, gdwZFZParams, gapZFZParam[gdwZFZParams]);
    gdwZFZParams ++;
    gfSkipIdleTickCount = TRUE;
    if (gdwZFZParams == 1)
      ListView_SetItemState (ghViewWnd, 0, LVIS_FOCUSED, LVIS_FOCUSED);
    if (gdwTaskNodeIndex >= gdwZFZParams) gdwTaskNodeIndex = gdwZFZParams - 1;
    EnableZFZMenu ();
    SaveZFZList ();
  }
}
/****************************************************************************************/
static BOOL CALLBACK ZFZ_DelDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  DWORD            i;
  PZFZPARAM        pZFZParam;

  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    for (i=0; i<gdwZFZParams; i++)
      SendDlgItemMessage (hDlg, IDC_ZFZ_NAME, CB_ADDSTRING, 
                          0, (LPARAM)gapZFZParam[i]->aszName);
    i = ListView_GetNextItem (ghViewWnd, -1, LVNI_FOCUSED);
    if (LISTVIEW_GetZFZParam (ghViewWnd, i, &pZFZParam))
      SendDlgItemMessage (hDlg, IDC_ZFZ_NAME, CB_SETCURSEL, 
                          ZFZPARAM_SearchByName (pZFZParam->aszName), 0);
    else
      SendDlgItemMessage (hDlg, IDC_ZFZ_NAME, CB_SETCURSEL, 0, 0);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDOK:
      GetWindowText (GetDlgItem (hDlg, IDC_ZFZ_NAME),
                     gaszDlgZFZName, sizeof gaszDlgZFZName);
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
void ZFZ_DelDialog(void)
{
  DWORD           iItem, iIndex;
  LVFINDINFO      FindInfo;

  if (DialogBox (ghInst, MAKEINTRESOURCE (IDD_DELETE), ghMainWnd, ZFZ_DelDlgProc) == IDOK)
  {
    iIndex = ZFZPARAM_SearchByName (gaszDlgZFZName);
    if (iIndex != 0xffff)
    {
      FindInfo.flags  = LVFI_PARAM;
      FindInfo.lParam = (LPARAM)gapZFZParam[iIndex];
      iItem = ListView_FindItem (ghViewWnd, -1, &FindInfo);
      if (iItem != -1)
      {
        if (ListView_DeleteItem (ghViewWnd, iItem))
        {
          free (gapZFZParam[iIndex]);
          if (gdwTaskNodeIndex>iIndex) gdwTaskNodeIndex --;
          for (; iIndex<gdwZFZParams-1; iIndex++)
            gapZFZParam[iIndex] = gapZFZParam[iIndex+1];
          gdwZFZParams --;
          EnableZFZMenu ();
          SaveZFZList ();
        }
      }
    }
  }
}
/****************************************************************************************/
static BOOL CALLBACK ZFZ_PropertiesDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  WORD            iItem;
  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    GetWindowText (hDlg, gaszString2, sizeof gaszString2);
    wsprintf (gaszString, gaszString2, gpDlgZFZParam->aszName);
    SetWindowText (hDlg, gaszString);
    SetWindowText (GetDlgItem (hDlg, IDC_ZFZ_NAME), gpDlgZFZParam->aszName);
    SetWindowText (GetDlgItem (hDlg, IDC_ZFZ_PHONENO), gpDlgZFZParam->aszPhone);
    SetWindowText (GetDlgItem (hDlg, IDC_ZFZ_REMARK), gpDlgZFZParam->aszRemark);
    SendDlgItemMessage (hDlg, IDC_ZFZ_NAME, EM_LIMITTEXT, ZFZNAMELEN, 0);
    SendDlgItemMessage (hDlg, IDC_ZFZ_PHONENO, EM_LIMITTEXT, ZFZPHONELEN, 0);
    SendDlgItemMessage (hDlg, IDC_ZFZ_REMARK, EM_LIMITTEXT, ZFZREMARKLEN, 0);
    if ( !(gUserRight & UR_MODIFY) )
    {
      EnableWindow (GetDlgItem (hDlg, IDC_ZFZ_NAME), FALSE);
      EnableWindow (GetDlgItem (hDlg, IDC_ZFZ_PHONENO), FALSE);
      EnableWindow (GetDlgItem (hDlg, IDC_ZFZ_REMARK), FALSE);
    }
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDOK:
      GetWindowText (GetDlgItem (hDlg, IDC_ZFZ_NAME),
                     gaszDlgZFZName, sizeof gaszDlgZFZName);
      if (gaszDlgZFZName[0] == 0)
      {
        SetFocus (GetDlgItem (hDlg, IDC_ZFZ_NAME));
        return TRUE;
      }
      GetWindowText (GetDlgItem (hDlg, IDC_ZFZ_PHONENO),
                     gaszDlgRAMPhone, sizeof gaszDlgRAMPhone);
      if (gaszDlgRAMPhone[0] == 0)
      {
        SetFocus (GetDlgItem (hDlg, IDC_ZFZ_PHONENO));
        return TRUE;
      }
      GetWindowText (GetDlgItem (hDlg, IDC_ZFZ_REMARK),
                     gaszDlgZFZRemark, sizeof gaszDlgZFZRemark);
      iItem = ZFZPARAM_SearchByName (gaszDlgZFZName);
      if (iItem != 0xffff && gapZFZParam[iItem]!=gpDlgZFZParam)
      {
        LoadString (ghInst, IDS_DUPLICATEZFZNAME, gaszString, sizeof gaszString);
        LoadString (ghInst, IDS_PARAMETER_TITLE, gaszString2, sizeof gaszString2);
        MessageBox (NULL, gaszString, gaszString2, MB_OK|MB_ICONERROR);
        return TRUE;
      }
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
void ZFZ_PropertiesDialog(void)
{
  int             iItem;
  LVITEM          lvi;
  TASKID          TaskID;

  iItem = ListView_GetNextItem (ghViewWnd, -1, LVNI_FOCUSED);
  if (iItem == -1) return;
  if (!LISTVIEW_GetZFZParam (ghViewWnd, iItem, &gpDlgZFZParam)) return;

  if (DialogBox (ghInst, MAKEINTRESOURCE (IDD_PROPERTIES),
                 ghMainWnd, ZFZ_PropertiesDlgProc) == IDOK)
  {
    if (strcmp (gpDlgZFZParam->aszPhone, gaszDlgRAMPhone) != 0)
    {
      do
        TaskID=InterlockedExchange (&gpDlgZFZParam->TaskID, TASKID_SYNCHRONIZE);
      while (TaskID == TASKID_SYNCHRONIZE);
      TaskID |= TASKID_GETSTATUS | TASKID_SETPHONE2;
      InterlockedExchange (&gpDlgZFZParam->TaskID, TaskID);
      if (gdwTaskNodeIndex >= gdwZFZParams) gdwTaskNodeIndex = gdwZFZParams - 1;
      gfSkipIdleTickCount = TRUE;
    }
    strcpy (gpDlgZFZParam->aszName, gaszDlgZFZName);
    strcpy (gpDlgZFZParam->aszPhone, gaszDlgRAMPhone);
    strcpy (gpDlgZFZParam->aszRemark, gaszDlgZFZRemark);
    lvi.mask     = LVIF_TEXT;
    lvi.iItem    = iItem;
    lvi.iSubItem = 0;
    lvi.pszText  = LPSTR_TEXTCALLBACK;
    ListView_SetItem (ghViewWnd, &lvi);
    ListView_Update (ghViewWnd, iItem);
    SaveZFZList ();
  }
}
/****************************************************************************************/
void ZFZ_GetStatus(void)
{
  int             iItem;
  PZFZPARAM       pZFZParam;
  TASKID          TaskID;

  iItem = ListView_GetNextItem (ghViewWnd, -1, LVNI_FOCUSED);
  if (iItem == -1) return;
  if (!LISTVIEW_GetZFZParam (ghViewWnd, iItem, &pZFZParam)) return;
  pZFZParam->fPopupStatusDialog = TRUE;
  do
    TaskID=InterlockedExchange (&pZFZParam->TaskID, TASKID_SYNCHRONIZE);
  while (TaskID == TASKID_SYNCHRONIZE);
  TaskID |= TASKID_GETSTATUS;
  InterlockedExchange (&pZFZParam->TaskID, TaskID);
  gfSkipIdleTickCount = TRUE;
  if (gdwTaskNodeIndex >= gdwZFZParams) gdwTaskNodeIndex = gdwZFZParams - 1;
  SaveZFZList ();
}
/****************************************************************************************/
void ZFZ_ThrnOn(void)
{
  int             iItem;
  PZFZPARAM       pZFZParam;
  TASKID          TaskID;

  iItem = ListView_GetNextItem (ghViewWnd, -1, LVNI_FOCUSED);
  if (iItem == -1) return;
  if (!LISTVIEW_GetZFZParam (ghViewWnd, iItem, &pZFZParam)) return;
  do
    TaskID=InterlockedExchange (&pZFZParam->TaskID, TASKID_SYNCHRONIZE);
  while (TaskID == TASKID_SYNCHRONIZE);
  TaskID |= TASKID_TURNON | TASKID_GETSTATUS;
  InterlockedExchange (&pZFZParam->TaskID, TaskID);
  gfSkipIdleTickCount = TRUE;
  if (gdwTaskNodeIndex >= gdwZFZParams) gdwTaskNodeIndex = gdwZFZParams - 1;
  SaveZFZList ();
}
/****************************************************************************************/
void ZFZ_ThrnOff(void)
{
  int             iItem;
  PZFZPARAM       pZFZParam;
  TASKID          TaskID;

  iItem = ListView_GetNextItem (ghViewWnd, -1, LVNI_FOCUSED);
  if (iItem == -1) return;
  if (!LISTVIEW_GetZFZParam (ghViewWnd, iItem, &pZFZParam)) return;
  do
    TaskID=InterlockedExchange (&pZFZParam->TaskID, TASKID_SYNCHRONIZE);
  while (TaskID == TASKID_SYNCHRONIZE);
  TaskID |= TASKID_TURNOFF | TASKID_GETSTATUS;
  InterlockedExchange (&pZFZParam->TaskID, TaskID);
  gfSkipIdleTickCount = TRUE;
  if (gdwTaskNodeIndex >= gdwZFZParams) gdwTaskNodeIndex = gdwZFZParams - 1;
  SaveZFZList ();
}
/****************************************************************************************/
static LRESULT CALLBACK HistoryWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
  {
  case WM_SETFOCUS:
    SetFocus ((HWND)GetWindowLong (hWnd, GWL_USERDATA));
    return 1;
  case WM_SIZE:
    MoveWindow ((HWND)GetWindowLong (hWnd, GWL_USERDATA),
                0, 0, LOWORD (lParam), HIWORD (lParam), TRUE);
    return 1;
  }
  return DefWindowProc (hWnd, uMsg, wParam, lParam);
}
/****************************************************************************************/
HWND ZFZ_HistoryWindow(void)
{
  HWND            hWnd;
  HWND            hTextWnd;
  FILE           *fp;
  DWORD           dwTime;
  ZFZPARAM       *pZFZParam;
  BYTE            abBuf[sizeof HISTORYHEADER];
  struct tm      *ptm;
  WORD            i;
  DWORD           j;

  pZFZParam = ZFZPARAM_malloc (255);
  if (pZFZParam == NULL) return NULL;

  hWnd = CreateWindowEx (WS_EX_APPWINDOW, HISTORYWNDCLASSNAME, "History",
                         WS_OVERLAPPEDWINDOW,                     
                         20, 20,
                         GetSystemMetrics (SM_CXFULLSCREEN)-40,
                         GetSystemMetrics (SM_CYFULLSCREEN)-20,
                         ghMainWnd, NULL, ghInst, NULL);
  
  hTextWnd = HISWND_CreateWnd (ghInst, hWnd, STFLAG_ALWAYSFOCUS/*|STFLAG_IGNORETYPE*/);
  SetWindowLong (hWnd, GWL_USERDATA, (LONG)hTextWnd);

  fp = fopen (HISTORYFILENAME, "rb");
  if (fp)
  {
    if (fread (abBuf, 1, sizeof HISTORYHEADER, fp) == sizeof HISTORYHEADER &&
        memcmp (abBuf, HISTORYHEADER, sizeof HISTORYHEADER) == 0)
    {
      while (!feof (fp))
      {
        if (fread (&dwTime, 1, sizeof dwTime, fp) != sizeof dwTime) break;
        if (fread (&i, 1, sizeof i, fp) != sizeof i) break;
        if (fread (pZFZParam, 1, i, fp) != i) break;
        ptm = localtime (&dwTime);

        HISWND_AppendInfo (hTextWnd, OMPT_SUCCESS, "%04d-%02d-%02d %02d:%02d : %s",
                           1990+ptm->tm_year, ptm->tm_mon, ptm->tm_mday,
                           ptm->tm_hour, ptm->tm_min,
                           pZFZParam->aszName);

        GetTaskString (gaszString, pZFZParam->ExecutedTaskID);
        LoadString (ghInst, IDS_HIS_CMD, gaszString2, sizeof gaszString2);
        HISWND_AppendInfo (hTextWnd, OMPT_UPLOAD, gaszString2, gaszString);
        LoadString (ghInst, IDS_HIS_STATUS, gaszString2, sizeof gaszString2);
        HISWND_AppendInfo (hTextWnd, OMPT_UPLOAD, gaszString2);

        strcpy (gaszString, "  ");
        LoadString (ghInst, IDS_HIS_VOLTAGE, gaszString2, sizeof gaszString2);
        strcat (gaszString, gaszString2);
        i = pZFZParam->bVoltage*pZFZParam->bRatedVoltage*100/204+5;
        wsprintf (gaszString2, "%u.%u | ", i/100, i/10%10);
        strcat (gaszString, gaszString2);
        LoadString (ghInst, IDS_HIS_CURRENT, gaszString2, sizeof gaszString2);
        strcat (gaszString, gaszString2);
        i = pZFZParam->bCurrent*pZFZParam->bRatedCurrent*100/204+5;
        wsprintf (gaszString2, "%u.%u | ", i/100, i/10%10);
        strcat (gaszString, gaszString2);
        LoadString (ghInst, IDS_HIS_TEMPERATURE, gaszString2, sizeof gaszString2);
        strcat (gaszString, gaszString2);
        i = pZFZParam->bTemperature*pZFZParam->bRatedTempture*100/204+5;
        wsprintf (gaszString2, "%u.%u", i/100, i/10%10);
        strcat (gaszString, gaszString2);
        HISWND_AppendInfo (hTextWnd, OMPT_UPLOAD, gaszString);

        for (i=0; i<pZFZParam->bChannels; i++)
        {
          LoadString (ghInst, IDS_CHANNEL, gaszString2, sizeof gaszString2);
          wsprintf (gaszString, "  %s%d", gaszString2, i);
          LoadString (ghInst, IDS_HIS_UP, gaszString2, sizeof gaszString2);
          strcat (gaszString, gaszString2);
          strcat (gaszString, " - ");
#ifndef FIXEDFREQUENCE
          if (pZFZParam->ChannelData[i].bErrorCode & 0x10)
          {
            LoadString (ghInst, IDS_HIS_UNLOCK, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
          }
          else
          {
            LoadString (ghInst, IDS_HIS_FREQUENCY, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
            if (pZFZParam->ChannelData[i].wFrequency <= 333)
              j = pZFZParam->ChannelData[i].wFrequency;
            else if (pZFZParam->ChannelData[i].wFrequency >= 0x180 &&
                     pZFZParam->ChannelData[i].wFrequency < 0x180+125)
              j = pZFZParam->ChannelData[i].wFrequency - 0x180;
            else
              j = 0;
#ifdef CDMA
            wsprintf (gaszString2, "%u.%01u", j*3/100+825, j*3%100);
            strcat (gaszString, gaszString2);
#else
            wsprintf (gaszString2, "%u.%01u", j/5+890, j%5*2);
            strcat (gaszString, gaszString2);
#endif
            strcat (gaszString, "MHz | ");
#else
          {
#endif
            LoadString (ghInst, IDS_HIS_GAIN, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
            wsprintf (gaszString2, "%u | ", pZFZParam->ChannelData[i].bUpGain);
            strcat (gaszString, gaszString2);
            LoadString (ghInst, IDS_HIS_POWER, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
            wsprintf (gaszString2, "%u | ", pZFZParam->ChannelData[i].bUpPower);
            strcat (gaszString, gaszString2);
            LoadString (ghInst, IDS_HIS_ACTPOWER, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
            j = pZFZParam->ChannelData[i].bActualUpPower *
                (pZFZParam->bRatedPower1*10+pZFZParam->bRatedPower2) * 10 / 204 + 5;
            wsprintf (gaszString2, "%u.%u", j/100, j/10%10);
            strcat (gaszString, gaszString2);
          }
          HISWND_AppendInfo (hTextWnd, OMPT_UPLOAD, gaszString);

          LoadString (ghInst, IDS_CHANNEL, gaszString2, sizeof gaszString2);
          wsprintf (gaszString, "  %s%d", gaszString2, i);
          LoadString (ghInst, IDS_HIS_DOWN, gaszString2, sizeof gaszString2);
          strcat (gaszString, gaszString2);
          strcat (gaszString, " - ");
#ifndef FIXEDFREQUENCE
          if (pZFZParam->ChannelData[i].bErrorCode & 0x20)
          {
            LoadString (ghInst, IDS_HIS_UNLOCK, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
          }
          else
          {
            LoadString (ghInst, IDS_HIS_FREQUENCY, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
            if (pZFZParam->ChannelData[i].wFrequency <= 333)
              j = pZFZParam->ChannelData[i].wFrequency;
            else if (pZFZParam->ChannelData[i].wFrequency >= 0x180 &&
                     pZFZParam->ChannelData[i].wFrequency < 0x180+125)
              j = pZFZParam->ChannelData[i].wFrequency - 0x180;
            else
              j = 0;
#ifdef CDMA
            wsprintf (gaszString2, "%u.%01u", j*3/100+870, j*3%100);
            strcat (gaszString, gaszString2);
#else
            wsprintf (gaszString2, "%u.%01u", j/5+935, j%5*2);
            strcat (gaszString, gaszString2);
#endif
            strcat (gaszString, "MHz | ");
#else
          {
#endif
            LoadString (ghInst, IDS_HIS_GAIN, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
            wsprintf (gaszString2, "%u | ", pZFZParam->ChannelData[i].bDownGain);
            strcat (gaszString, gaszString2);
            LoadString (ghInst, IDS_HIS_POWER, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
            wsprintf (gaszString2, "%u | ", pZFZParam->ChannelData[i].bDownPower);
            strcat (gaszString, gaszString2);
            LoadString (ghInst, IDS_HIS_ACTPOWER, gaszString2, sizeof gaszString2);
            strcat (gaszString, gaszString2);
            j = pZFZParam->ChannelData[i].bActualDownPower *
                (pZFZParam->bRatedPower1*10+pZFZParam->bRatedPower2) * 10 / 204 + 5;
            wsprintf (gaszString2, "%u.%u", j/100, j/10%10);
            strcat (gaszString, gaszString2);
          }
          HISWND_AppendInfo (hTextWnd, OMPT_UPLOAD, gaszString);
        }
      }
    }

    fclose (fp);
  }

  ShowWindow (hWnd, SW_SHOW);
  UpdateWindow (hWnd);

  free (pZFZParam);
  return hWnd;
}
/****************************************************************************************/
static BOOL CALLBACK ZFZ_ChangeDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static WORD     wCurChannel;
  static DWORD    dwChanged;
  WORD            i;

  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    EnableAllWindow (FALSE);
    dwChanged = 0;
    GetWindowText (hDlg, gaszString2, sizeof gaszString2);
    wsprintf (gaszString, gaszString2, gpDlgZFZParam->aszName);
    SetWindowText (hDlg, gaszString);
    LoadString (ghInst, IDS_CHANNEL, gaszString, sizeof gaszString);
    for (i=1; i<=gpDlgZFZParam->bChannels; i++)
    {
      wsprintf (gaszString2, "%s %u", gaszString, i);
      SendDlgItemMessage (hDlg, IDC_CHANNEL, CB_ADDSTRING, 0, (LPARAM)gaszString2);
    }
    SendDlgItemMessage (hDlg, IDC_DISABLEALARM, BM_SETCHECK,
                        gpDlgZFZParam->fDisableAlarm ? BST_CHECKED : BST_UNCHECKED, 0);
    wCurChannel = 0;
  SetChannelData:
#ifndef FIXEDFREQUENCE
    if (gpDlgZFZParam->ChannelData[wCurChannel].wFrequency <= 333)
      i = gpDlgZFZParam->ChannelData[wCurChannel].wFrequency;
    else if (gpDlgZFZParam->ChannelData[wCurChannel].wFrequency >= 0x180 &&
             gpDlgZFZParam->ChannelData[wCurChannel].wFrequency < 0x180+125)
      i = gpDlgZFZParam->ChannelData[wCurChannel].wFrequency - 0x180;
    else
      i = 0;
    SendDlgItemMessage (hDlg, IDC_CHANNEL, CB_SETCURSEL, wCurChannel, 0);
    wsprintf (gaszString, "%u", i);
    SetWindowText (GetDlgItem (hDlg, IDC_FREQUENCY), gaszString);
#ifdef CDMA
    wsprintf (gaszString, "%u.%01u", i*3/100+825, i*3%100);
    SetWindowText (GetDlgItem (hDlg, IDC_UPFREQUENCY), gaszString);
    wsprintf (gaszString, "%u.%01u",  i*3/100+870, i*3%100);
    SetWindowText (GetDlgItem (hDlg, IDC_DOWNFREQUENCY), gaszString);
#else
    wsprintf (gaszString, "%u.%01u", i/5+890, i%5*2);
    SetWindowText (GetDlgItem (hDlg, IDC_UPFREQUENCY), gaszString);
    wsprintf (gaszString, "%u.%01u", i/5+935, i%5*2);
    SetWindowText (GetDlgItem (hDlg, IDC_DOWNFREQUENCY), gaszString);
#endif
#endif
    wsprintf (gaszString, "%u", gpDlgZFZParam->ChannelData[wCurChannel].bUpGain);
    SetWindowText (GetDlgItem (hDlg, IDC_UPGAIN), gaszString);
    wsprintf (gaszString, "%u", gpDlgZFZParam->ChannelData[wCurChannel].bDownGain);
    SetWindowText (GetDlgItem (hDlg, IDC_DOWNGAIN), gaszString);
    wsprintf (gaszString, "%u", gpDlgZFZParam->ChannelData[wCurChannel].bUpPower);
    SetWindowText (GetDlgItem (hDlg, IDC_UPPOWER), gaszString);
    wsprintf (gaszString, "%u", gpDlgZFZParam->ChannelData[wCurChannel].bDownPower);
    SetWindowText (GetDlgItem (hDlg, IDC_DOWNPOWER), gaszString);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
#ifndef FIXEDFREQUENCE
    case IDC_FREQUENCY:
      if (HIWORD (wParam) == EN_CHANGE)
      {
#ifdef CDMA
        if (ValidateField (hDlg, IDC_FREQUENCY, 0, 333))
        {
          GetWindowText (GetDlgItem (hDlg, IDC_FREQUENCY), gaszString, sizeof gaszString);
          i = atoi (gaszString);
          wsprintf (gaszString, "%u.%01u", i*3/100+825, i*3%100);
          SetWindowText (GetDlgItem (hDlg, IDC_UPFREQUENCY), gaszString);
          wsprintf (gaszString, "%u.%01u", i*3/100+870, i*3%100);
          SetWindowText (GetDlgItem (hDlg, IDC_DOWNFREQUENCY), gaszString);
        }
#else
        if (ValidateField (hDlg, IDC_FREQUENCY, 0, 125))
        {
          GetWindowText (GetDlgItem (hDlg, IDC_FREQUENCY), gaszString, sizeof gaszString);
          i = atoi (gaszString);
          wsprintf (gaszString, "%u.%01u", i/5+890, i%5*2);
          SetWindowText (GetDlgItem (hDlg, IDC_UPFREQUENCY), gaszString);
          wsprintf (gaszString, "%u.%01u", i/5+935, i%5*2);
          SetWindowText (GetDlgItem (hDlg, IDC_DOWNFREQUENCY), gaszString);
        }
#endif
      }
      return TRUE;
#endif
    case IDC_UPGAIN:
      if (HIWORD (wParam) == EN_CHANGE)
        ValidateField (hDlg, IDC_UPGAIN, 0, 31);
      return TRUE;
    case IDC_DOWNGAIN:
      if (HIWORD (wParam) == EN_CHANGE)
        ValidateField (hDlg, IDC_DOWNGAIN, 0, 31);
      return TRUE;
    case IDC_UPPOWER:
      if (HIWORD (wParam) == EN_CHANGE)
        ValidateField (hDlg, IDC_UPPOWER, 0, 255);
      return TRUE;
    case IDC_DOWNPOWER:
      if (HIWORD (wParam) == EN_CHANGE)
        ValidateField (hDlg, IDC_DOWNPOWER, 0, 255);
      return TRUE;
    case IDC_DISABLEALARM:
      if (HIWORD (wParam) == BN_CLICKED)
      {
        if (SendDlgItemMessage (hDlg, IDC_DISABLEALARM, BM_GETCHECK, 0, 0) == BST_CHECKED)
        {
          LoadString (ghInst, IDS_WARN_DISABLEALARM, gaszString, sizeof gaszString);
          LoadString (ghInst, IDS_WARNING_TITLE, gaszString2, sizeof gaszString2);
          if (MessageBox (hDlg, gaszString, gaszString2, MB_YESNO|MB_ICONWARNING) == IDNO)
            SendDlgItemMessage (hDlg, IDC_DISABLEALARM, BM_SETCHECK, BST_UNCHECKED, 0);
        }
      }
      return TRUE;
    case IDC_CHANNEL:
      if (HIWORD (wParam) != CBN_SELCHANGE) return TRUE;
    case IDOK:
#ifndef FIXEDFREQUENCE
#ifdef CDMA
      if (!ValidateField (hDlg, IDC_FREQUENCY, 0, 333)) return TRUE;
#else
      if (!ValidateField (hDlg, IDC_FREQUENCY, 0, 125)) return TRUE;
#endif
#endif
      if (!ValidateField (hDlg, IDC_UPGAIN, 0, 31)) return TRUE;
      if (!ValidateField (hDlg, IDC_DOWNGAIN, 0, 31)) return TRUE;
      if (!ValidateField (hDlg, IDC_UPPOWER, 0, 255)) return TRUE;
      if (!ValidateField (hDlg, IDC_DOWNPOWER, 0, 255)) return TRUE;
#ifndef FIXEDFREQUENCE
      GetWindowText (GetDlgItem (hDlg, IDC_FREQUENCY), gaszString, sizeof gaszString);
      gpDlgZFZParam->ChannelData[wCurChannel].wFrequency = (WORD)atoi (gaszString);
#ifdef GSM
      gpDlgZFZParam->ChannelData[wCurChannel].wFrequency += 0x180;
#endif
#else
      gpDlgZFZParam->ChannelData[wCurChannel].wFrequency = 0;
#endif
      GetWindowText (GetDlgItem (hDlg, IDC_UPGAIN), gaszString, sizeof gaszString);
      gpDlgZFZParam->ChannelData[wCurChannel].bUpGain = (BYTE)atoi (gaszString);
      GetWindowText (GetDlgItem (hDlg, IDC_DOWNGAIN), gaszString, sizeof gaszString);
      gpDlgZFZParam->ChannelData[wCurChannel].bDownGain =  (BYTE)atoi (gaszString);
      GetWindowText (GetDlgItem (hDlg, IDC_UPPOWER), gaszString, sizeof gaszString);
      gpDlgZFZParam->ChannelData[wCurChannel].bUpPower =  (BYTE)atoi (gaszString);
      GetWindowText (GetDlgItem (hDlg, IDC_DOWNPOWER), gaszString, sizeof gaszString);
      gpDlgZFZParam->ChannelData[wCurChannel].bDownPower = (BYTE)atoi (gaszString);
      if (LOWORD (wParam) == IDC_CHANNEL)
      {
        wCurChannel = (WORD)SendDlgItemMessage (hDlg, IDC_CHANNEL, CB_GETCURSEL, 0, 0);
        goto SetChannelData;
      }
      gpDlgZFZParam->fDisableAlarm = 
        SendDlgItemMessage (hDlg, IDC_DISABLEALARM, BM_GETCHECK, 0, 0) == BST_CHECKED;
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  case WM_DESTROY:
    EnableAllWindow (TRUE);
    break;
  }

  return FALSE;
}
/****************************************************************************************/
void ZFZ_ChangeDialog(PZFZPARAM pZFZParam)
{
  int             iItem;
  BYTE            i;
  TASKID          TaskID;
  BOOL            fAlarmChanged;

  if (pZFZParam == NULL)
  {
    iItem = ListView_GetNextItem (ghViewWnd, -1, LVNI_FOCUSED);
    if (iItem == -1) return;
    if (!LISTVIEW_GetZFZParam (ghViewWnd, iItem, &pZFZParam)) return;
  }

  if (pZFZParam->bChannels == 0)
  {
    LoadString (ghInst, IDS_HAVENOTZFZPARAM, gaszString2, sizeof gaszString2);
    wsprintf (gaszString, gaszString2, pZFZParam->aszName);
    LoadString (ghInst, IDS_PARAMETER_TITLE, gaszString2, sizeof gaszString2);
    MessageBox (ghMainWnd, gaszString, gaszString2, MB_OK|MB_ICONSTOP);
    return;
  }
  gpDlgZFZParam = ZFZPARAM_malloc (pZFZParam->bChannels);
  if (gpDlgZFZParam == NULL) return;
  memcpy (gpDlgZFZParam, pZFZParam, ZFZPARAM_SizeOf (pZFZParam->bChannels));
  if (DialogBox (ghInst, MAKEINTRESOURCE (IDD_CHANGE), ghMainWnd, ZFZ_ChangeDlgProc)==IDOK)
  {
    if (memcmp (gpDlgZFZParam, pZFZParam, ZFZPARAM_SizeOf (pZFZParam->bChannels)) != 0)
    {
      // Change status
      fAlarmChanged = (pZFZParam->fDisableAlarm != gpDlgZFZParam->fDisableAlarm);
      pZFZParam->fDisableAlarm = gpDlgZFZParam->fDisableAlarm;
      for (i=0; i<gpDlgZFZParam->bChannels; i++)
      {
        pZFZParam->ChannelData[i].wFrequency = gpDlgZFZParam->ChannelData[i].wFrequency;
        pZFZParam->ChannelData[i].bUpGain    = gpDlgZFZParam->ChannelData[i].bUpGain;
        pZFZParam->ChannelData[i].bDownGain  = gpDlgZFZParam->ChannelData[i].bDownGain;
        pZFZParam->ChannelData[i].bUpPower   = gpDlgZFZParam->ChannelData[i].bUpPower;
        pZFZParam->ChannelData[i].bDownPower = gpDlgZFZParam->ChannelData[i].bDownPower;
      }
      do
        TaskID=InterlockedExchange (&pZFZParam->TaskID, TASKID_SYNCHRONIZE);
      while (TaskID == TASKID_SYNCHRONIZE);
      TaskID |= TASKID_SETPARAM | TASKID_GETSTATUS;
      if (fAlarmChanged)
        TaskID |= TASKID_ALARM;
      InterlockedExchange (&pZFZParam->TaskID, TaskID);
      gfSkipIdleTickCount = TRUE;
      pZFZParam->fPopupStatusDialog = TRUE;
      if (gdwTaskNodeIndex >= gdwZFZParams) gdwTaskNodeIndex = gdwZFZParams - 1;
      SaveZFZList ();
    }
  }
  free (gpDlgZFZParam);
}
/****************************************************************************************/
static BOOL CALLBACK ZFZ_StatusDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  WORD            wCurChannel;
  DWORD           i;
  PZFZPARAM       pZFZParam;
  
  switch (uMsg)
  {
  case WM_INITDIALOG:
    pZFZParam = (PZFZPARAM)lParam;
    SetWindowLong (hDlg, GWL_USERDATA, lParam);
    AddWindowHandle (hDlg);
    CentreDialog (hDlg);
    GetWindowText (hDlg, gaszString2, sizeof gaszString2);
    wsprintf (gaszString, gaszString2, pZFZParam->aszName,
                                       pZFZParam->LastSamplingDate.wYear,
                                       pZFZParam->LastSamplingDate.bMonth,
                                       pZFZParam->LastSamplingDate.bDay,
                                       pZFZParam->LastSamplingDate.bHour,
                                       pZFZParam->LastSamplingDate.bMinute);
    SetWindowText (hDlg, gaszString);
    LoadString (ghInst, IDS_CHANNEL, gaszString, sizeof gaszString);
    for (i=1; i<=pZFZParam->bChannels; i++)
    {
      wsprintf (gaszString2, "%s %u", gaszString, i);
      SendDlgItemMessage (hDlg, IDC_CHANNEL, CB_ADDSTRING, 0, (LPARAM)gaszString2);
    }
    SendDlgItemMessage (hDlg, IDC_DISABLEALARM, BM_SETCHECK,
                        pZFZParam->fDisableAlarm ? BST_CHECKED : BST_UNCHECKED, 0);
    if (pZFZParam->bStatusFlag & 1)
      LoadString (ghInst, IDS_DOOROPEN, gaszString, sizeof gaszString);
    else
      LoadString (ghInst, IDS_DOORCLOSE, gaszString, sizeof gaszString);
    SetWindowText (GetDlgItem (hDlg, IDC_DOORSTATUS), gaszString);
    i = pZFZParam->bVoltage*pZFZParam->bRatedVoltage*100/204+5;
    wsprintf (gaszString, "%u.%u", i/100, i/10%10);
    SetWindowText (GetDlgItem (hDlg, IDC_VOLTAGE), gaszString);
    i = pZFZParam->bCurrent*pZFZParam->bRatedCurrent*100/204+5;
    wsprintf (gaszString, "%u.%u", i/100, i/10%10);
    SetWindowText (GetDlgItem (hDlg, IDC_CURRENT), gaszString);
    i = pZFZParam->bTemperature*pZFZParam->bRatedTempture*100/204+5;
    wsprintf (gaszString, "%u.%u", i/100, i/10%10);
    SetWindowText (GetDlgItem (hDlg, IDC_TEMPERATURE), gaszString);
    wCurChannel = 0;
  SetChannelData:
#ifndef FIXEDFREQUENCE
    if (pZFZParam->ChannelData[wCurChannel].wFrequency <= 333)
      i = pZFZParam->ChannelData[wCurChannel].wFrequency;
    else if (pZFZParam->ChannelData[wCurChannel].wFrequency >= 0x180 &&
             pZFZParam->ChannelData[wCurChannel].wFrequency < 0x180+125)
      i = pZFZParam->ChannelData[wCurChannel].wFrequency - 0x180;
    else
      i = 0;
    SendDlgItemMessage (hDlg, IDC_CHANNEL, CB_SETCURSEL, wCurChannel, 0);
    wsprintf (gaszString, "%u", i);
    SetWindowText (GetDlgItem (hDlg, IDC_FREQUENCY), gaszString);
    SendDlgItemMessage (hDlg, IDC_UPFLOCK, BM_SETCHECK,
                        pZFZParam->ChannelData[wCurChannel].bErrorCode & 0x10 ?
                        BST_UNCHECKED : BST_CHECKED, 0);
    SendDlgItemMessage (hDlg, IDC_DOWNFLOCK, BM_SETCHECK,
                        pZFZParam->ChannelData[wCurChannel].bErrorCode & 0x20 ?
                        BST_UNCHECKED : BST_CHECKED, 0);
#ifdef CDMA
    wsprintf (gaszString, "%u.%01u", i/10+825, i%10);
    SetWindowText (GetDlgItem (hDlg, IDC_UPFREQUENCY), gaszString);
    wsprintf (gaszString, "%u.%01u", i/10+870, i%10);
    SetWindowText (GetDlgItem (hDlg, IDC_DOWNFREQUENCY), gaszString);
#else
    wsprintf (gaszString, "%u.%01u", i/5+890, i%5*2);
    SetWindowText (GetDlgItem (hDlg, IDC_UPFREQUENCY), gaszString);
    wsprintf (gaszString, "%u.%01u", i/5+935, i%5*2);
    SetWindowText (GetDlgItem (hDlg, IDC_DOWNFREQUENCY), gaszString);
#endif
#endif
    wsprintf (gaszString, "%u", pZFZParam->ChannelData[wCurChannel].bUpGain);
    SetWindowText (GetDlgItem (hDlg, IDC_UPGAIN), gaszString);
    wsprintf (gaszString, "%u", pZFZParam->ChannelData[wCurChannel].bDownGain);
    SetWindowText (GetDlgItem (hDlg, IDC_DOWNGAIN), gaszString);
    wsprintf (gaszString, "%u", pZFZParam->ChannelData[wCurChannel].bUpPower);
    SetWindowText (GetDlgItem (hDlg, IDC_UPPOWER), gaszString);
    wsprintf (gaszString, "%u", pZFZParam->ChannelData[wCurChannel].bDownPower);
    SetWindowText (GetDlgItem (hDlg, IDC_DOWNPOWER), gaszString);
    i = pZFZParam->ChannelData[wCurChannel].bActualUpPower *
        (pZFZParam->bRatedPower1*10+pZFZParam->bRatedPower2) * 10 / 204 + 5;
    wsprintf (gaszString, "%u.%u", i/100, i/10%10);
    SetWindowText (GetDlgItem (hDlg, IDC_ACTUALUPPOWER), gaszString);
    i = pZFZParam->ChannelData[wCurChannel].bActualDownPower *
        (pZFZParam->bRatedPower1*10+pZFZParam->bRatedPower2) * 10 / 204 + 5;
    wsprintf (gaszString, "%u.%u", i/100, i/10%10);
    SetWindowText (GetDlgItem (hDlg, IDC_ACTUALDOWNPOWER), gaszString);
    if (! (gUserRight & UR_MODIFY))
      EnableWindow (GetDlgItem (hDlg, IDC_CHANGE), FALSE);
    return TRUE;
  case WM_ACTIVATE:
      if (wParam == WA_ACTIVE || wParam == WA_CLICKACTIVE)
        ghActivateDialog = hDlg;
      else
        ghActivateDialog = NULL;
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDC_CHANNEL:
      if (HIWORD (wParam) == CBN_SELCHANGE)
      {
        wCurChannel = (WORD)SendDlgItemMessage (hDlg, IDC_CHANNEL, CB_GETCURSEL, 0, 0);
        pZFZParam = (PZFZPARAM)GetWindowLong (hDlg, GWL_USERDATA);
        goto SetChannelData; 
      }
      return TRUE;
    case IDC_CHANGE:
      DestroyWindow (hDlg);
      pZFZParam = (PZFZPARAM)GetWindowLong (hDlg, GWL_USERDATA);
      ZFZ_ChangeDialog (pZFZParam);
      return TRUE;
    case IDOK:
    case IDCANCEL:
      DestroyWindow (hDlg);
      return TRUE;
    }
    break;
  case WM_DESTROY:
    RemoveWindowHandle (hDlg);
    break;
  }

  return FALSE;
}
/****************************************************************************************/
void ZFZ_StatusDialog(PZFZPARAM pZFZParam)
{
  int             iItem;
  HWND            hWnd;

  if (pZFZParam == NULL)
  {
    iItem = ListView_GetNextItem (ghViewWnd, -1, LVNI_FOCUSED);
    if (iItem == -1) return;
    if (!LISTVIEW_GetZFZParam (ghViewWnd, iItem, &pZFZParam)) return;
  }

  if (pZFZParam->bChannels == 0)
  {
    LoadString (ghInst, IDS_HAVENOTZFZPARAM, gaszString2, sizeof gaszString2);
    wsprintf (gaszString, gaszString2, pZFZParam->aszName);
    LoadString (ghInst, IDS_PARAMETER_TITLE, gaszString2, sizeof gaszString2);
    MessageBox (ghMainWnd, gaszString, gaszString2, MB_OK|MB_ICONSTOP);
    return;
  }
  hWnd = CreateDialogParam (ghInst, MAKEINTRESOURCE (IDD_STATUS), ghMainWnd,
                            ZFZ_StatusDlgProc, (LPARAM)pZFZParam);
  if (gfStatusDialogEnable)
    ShowWindow (hWnd, SW_SHOWNORMAL);
}
/****************************************************************************************/
static BOOL CALLBACK ModemSetupDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  CHAR            aszBuf[10];
  int             i;

  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    for (i=1; i<=4; i++)
    {
      wsprintf (aszBuf, "COM %u", i);
      SendDlgItemMessage (hDlg, IDC_SERIALPORT, CB_ADDSTRING, 0, (LPARAM)aszBuf);
    }
    SendDlgItemMessage (hDlg, IDC_SERIALPORT, CB_SETCURSEL, gModemParam.wCommPort-1, 0);
    SetWindowText (GetDlgItem (hDlg, IDC_MODEMINIT), gModemParam.aszInit);
    SendDlgItemMessage (hDlg, IDC_MODEMINIT, EM_LIMITTEXT, MAXATCMDLEN, 0);
    SetWindowText (GetDlgItem (hDlg, IDC_MODEMDIAL), gModemParam.aszDial);
    SendDlgItemMessage (hDlg, IDC_MODEMDIAL, EM_LIMITTEXT, MAXATCMDLEN, 0);
    SetWindowText (GetDlgItem (hDlg, IDC_MODEMANSWER), gModemParam.aszAnswer);
    SendDlgItemMessage (hDlg, IDC_MODEMANSWER, EM_LIMITTEXT, MAXATCMDLEN, 0);
    SetWindowText (GetDlgItem (hDlg, IDC_MODEMHANGUP), gModemParam.aszHangup);
    SendDlgItemMessage (hDlg, IDC_MODEMHANGUP, EM_LIMITTEXT, MAXATCMDLEN, 0);
    SetWindowText (GetDlgItem (hDlg, IDC_MODEMTOCMD), gModemParam.aszToCMD);
    SendDlgItemMessage (hDlg, IDC_MODEMTOCMD, EM_LIMITTEXT, MAXATCMDLEN, 0);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDOK:
      gModemParam.wCommPort =
        SendDlgItemMessage (hDlg, IDC_SERIALPORT, CB_GETCURSEL, 0, 0) + 1;
      GetWindowText (GetDlgItem (hDlg, IDC_MODEMINIT), gModemParam.aszInit, MAXATCMDLEN+1);
      GetWindowText (GetDlgItem (hDlg, IDC_MODEMDIAL), gModemParam.aszDial, MAXATCMDLEN+1);
      GetWindowText (GetDlgItem (hDlg, IDC_MODEMANSWER), gModemParam.aszAnswer, MAXATCMDLEN+1);
      GetWindowText (GetDlgItem (hDlg, IDC_MODEMHANGUP), gModemParam.aszHangup, MAXATCMDLEN+1);
      GetWindowText (GetDlgItem (hDlg, IDC_MODEMTOCMD), gModemParam.aszToCMD, MAXATCMDLEN+1);
      SetModemParam (&gModemParam);
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
BOOL ModemSetupDialog(void)
{
  STEC            ErrorCode;
  WORD            wOldCommPort = gModemParam.wCommPort;

  if (DialogBox (ghInst, MAKEINTRESOURCE (IDD_MODEMSETUP),
                 ghMainWnd, ModemSetupDlgProc) == IDOK)
  {
    if (wOldCommPort != gModemParam.wCommPort)
    {
      KillCommuicationThread ();
      while (1)
      {
        ErrorCode = StartCommunicationThread (gModemParam.wCommPort);
        if (ErrorCode == STEC_OPENCOMMPORT)
        {
          LoadString (ghInst, IDS_MODEMERR_TITLE, gaszString, sizeof gaszString);
          LoadString (ghInst, IDS_MODEMERR_01, gaszString2, sizeof gaszString2);
          MessageBox (ghMainWnd, gaszString2, gaszString, MB_OK|MB_ICONERROR);
          if (!ModemSetupDialog ())
          {
            DestroyWindow (ghMainWnd);
            break;
          }
        }
        else if (ErrorCode == STEC_STARTTHREAD)
        {
          LoadString (ghInst, IDS_SYSERR_TITLE, gaszString, sizeof gaszString);
          LoadString (ghInst, IDS_THREAD, gaszString2, sizeof gaszString2);
          MessageBox (ghMainWnd, gaszString2, gaszString, MB_OK|MB_ICONERROR);
          DestroyWindow (ghMainWnd);
          break;
        }
        else
          break;
      }
      STATUSBAR_ShowTaskStatus ("", TASKID_IDLE);
    }
    return TRUE;
  }
  else
    return FALSE;
}
/****************************************************************************************/
static BOOL CALLBACK CentrePhoneDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  CHAR              aszPhone[ZFZPHONELEN+1];

  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    SetWindowText (GetDlgItem (hDlg, IDC_CENTREPHONE), gaszCentrePhone);
    SendDlgItemMessage (hDlg, IDC_CENTREPHONE, EM_LIMITTEXT, ZFZPHONELEN, 0);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDOK:
      GetWindowText (GetDlgItem (hDlg, IDC_CENTREPHONE), aszPhone, sizeof aszPhone);
      if (aszPhone[0] == '\0')
      {
        SetFocus (GetDlgItem (hDlg, IDC_CENTREPHONE));
        return TRUE;
      }
      if (strcmp (gaszCentrePhone, aszPhone) != 0)
      {
        strcpy (gaszCentrePhone, aszPhone);
        EndDialog (hDlg, TRUE);
        return TRUE;
      }
    case IDCANCEL:
      EndDialog (hDlg, FALSE);
      return TRUE;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
BOOL CentrePhoneDialog(void)
{
  BOOL            fResult;
  DWORD           i;
  TASKID          TaskID;

  fResult = DialogBox (ghInst, MAKEINTRESOURCE (IDD_CENTREPHONE), ghMainWnd,
                       CentrePhoneDlgProc);
  if (fResult)
  {
    for (i=0; i<gdwZFZParams; i++)
    {
      do
        TaskID=InterlockedExchange (&gapZFZParam[i]->TaskID, TASKID_SYNCHRONIZE);
      while (TaskID == TASKID_SYNCHRONIZE);
      TaskID |= TASKID_GETSTATUS | TASKID_SETPHONE1;
      InterlockedExchange (&gapZFZParam[i]->TaskID, TaskID);
    }
    if (gdwTaskNodeIndex >= gdwZFZParams) gdwTaskNodeIndex = gdwZFZParams - 1;
  }
  return fResult;
}
/****************************************************************************************/
static BOOL CALLBACK AboutDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDOK:
    case IDCANCEL:
      EndDialog (hDlg, LOWORD (wParam));
      return TRUE;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
void AboutDialog(void)
{
  DialogBox (ghInst, MAKEINTRESOURCE (IDD_ABOUT), ghMainWnd, AboutDlgProc);
}
/****************************************************************************************/
static BOOL CALLBACK TerminateDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
  {
  case WM_INITDIALOG:
    CentreDialog (hDlg);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDC_STOPTHREAD:
      DestroyWindow (ghMainWnd);
      break;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
static BOOL CALLBACK OpeningDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static BMPINFO *pBmpInfo;
  static BYTE    *pbDIBBit;
  static int      w, h;
  int             x, y;
  HDC             hDC;
  PAINTSTRUCT     ps;

  switch (uMsg)
  {
  case WM_INITDIALOG:
    pBmpInfo = GetBmpInfo ("opening.bmp");
    if (pBmpInfo == NULL || pBmpInfo->BmpHeader.biBitCount != 24 ||
        pBmpInfo->BmpHeader.biCompression != BI_RGB)
    {
      EndDialog (hDlg, 0);
      return TRUE;
    }
    w = pBmpInfo->BmpHeader.biWidth;
    h = pBmpInfo->BmpHeader.biHeight;
    x = (GetSystemMetrics (SM_CXFULLSCREEN)-w)/2;
    y = (GetSystemMetrics (SM_CYFULLSCREEN)-h)/2;
    MoveWindow (hDlg, x, y, w, h, TRUE);
    pbDIBBit = malloc (pBmpInfo->FileHeader.bfSize - pBmpInfo->FileHeader.bfOffBits);
    if (pbDIBBit == NULL)
    {
      EndDialog (hDlg, 0);
      return TRUE;
    }
    fseek (pBmpInfo->BitmapFile, pBmpInfo->FileHeader.bfOffBits, SEEK_SET);
    fread (pbDIBBit, 1, pBmpInfo->FileHeader.bfSize - pBmpInfo->FileHeader.bfOffBits,
           pBmpInfo->BitmapFile);
    SetTimer (hDlg, 1, 2000, NULL);
    return TRUE;
  case WM_DESTROY:
    KillTimer (hDlg, 1);
    if (pbDIBBit) free (pbDIBBit);
    if (pBmpInfo) ReleaseBmpInfo (pBmpInfo);
    return FALSE;
  case WM_PAINT:
    hDC = BeginPaint (hDlg, &ps);
    StretchDIBits (hDC, 0, 0, w, h, 0, 0, w, h, pbDIBBit,
                   (BITMAPINFO*)&pBmpInfo->BmpHeader, DIB_RGB_COLORS, SRCCOPY);
    EndPaint (hDlg, &ps);
    return TRUE;
  case WM_LBUTTONDOWN:
    EndDialog (hDlg, 0);
    return TRUE;
  case WM_TIMER:
    EndDialog (hDlg, 0);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case IDOK:
    case IDCANCEL:
      EndDialog (hDlg, 0);
      return TRUE;
    }
    break;
  }

  return FALSE;
}
/****************************************************************************************/
/****************************************************************************************/
HWND CreateListViewControl(HINSTANCE hInst, HWND hWndParent)
{
  HWND            hWnd;
  HIMAGELIST      hLargeImageList, hSmallImageList;
  HICON           hIcon;
  LVCOLUMN        lvc;
  WORD            i;

  hWnd = CreateWindowEx (WS_EX_STATICEDGE, WC_LISTVIEW, "",
                         WS_CHILD|LVS_SINGLESEL|LVS_REPORT,
                         0, 0, 0, 0, hWndParent, NULL, hInst, NULL);
  if (hWnd == NULL) return NULL;
  SendMessage (hWnd, WM_COMMAND, ID_VIEW_DETAILMODE, 0);

  hLargeImageList = ImageList_Create (GetSystemMetrics(SM_CXICON), 
                                      GetSystemMetrics(SM_CYICON), ILC_COLOR8, 0, 0);
  hSmallImageList = ImageList_Create (GetSystemMetrics(SM_CXSMICON), 
                                      GetSystemMetrics(SM_CYSMICON), ILC_COLOR8, 0, 0);
  hIcon = LoadIcon (hInst, MAKEINTRESOURCE(IDI_NORMAL));
  ImageList_AddIcon (hLargeImageList, hIcon);
  ImageList_AddIcon (hSmallImageList, hIcon);
  DeleteObject(hIcon);
  hIcon = LoadIcon (hInst, MAKEINTRESOURCE(IDI_PROBLEM));
  ImageList_AddIcon (hLargeImageList, hIcon);
  ImageList_AddIcon (hSmallImageList, hIcon);
  DeleteObject(hIcon);
  hIcon = LoadIcon (hInst, MAKEINTRESOURCE(IDI_INSTALL));
  ImageList_AddIcon (hLargeImageList, hIcon);
  ImageList_AddIcon (hSmallImageList, hIcon);
  DeleteObject(hIcon);
  hIcon = LoadIcon (hInst, MAKEINTRESOURCE(IDI_CONNECT));
  ImageList_AddIcon (hLargeImageList, hIcon);
  ImageList_AddIcon (hSmallImageList, hIcon);
  DeleteObject(hIcon);
  hIcon = LoadIcon (hInst, MAKEINTRESOURCE(IDI_TURNOFF));
  ImageList_AddIcon (hLargeImageList, hIcon);
  ImageList_AddIcon (hSmallImageList, hIcon);
  DeleteObject(hIcon);
  ListView_SetImageList (hWnd, hLargeImageList, LVSIL_NORMAL);
  ListView_SetImageList (hWnd, hSmallImageList, LVSIL_SMALL);

  for (i=0; i<VIEWWNDCOL; i++)
  {
    lvc.mask = LVCF_FMT|LVCF_WIDTH|LVCF_TEXT|LVCF_SUBITEM; 
    lvc.fmt = VIEWWNDCOLFORMAT[i];
    lvc.cx = VIEWWNDCOLWIDTH[i];
    lvc.iSubItem = i;
    LoadString (ghInst, IDS_HEADERCOL_FIRST+i, gaszString, sizeof gaszString);
    lvc.pszText = gaszString;
    ListView_InsertColumn (hWnd, i, &lvc);
  }

  if (LoadZFZList ())
  {
    for (i=0; i<gdwZFZParams; i++)
    {
      gapZFZParam[i]->TaskID &= ~TASKID_LOCKNODE;
      LISTVIEW_AddItem (hWnd, i, gapZFZParam[i]);
    }
    if (gdwZFZParams)
      ListView_SetItemState (hWnd, 0, LVIS_FOCUSED, LVIS_FOCUSED);
  }
  else
  {
    gaszCentrePhone[0] = '\0';
    gdwZFZParams = 0;
    SaveZFZList ();
  }
  gdwTaskNodeIndex = gdwZFZParams - 1;

  ShowWindow (hWnd, SW_SHOW);
  SetFocus (hWnd);
  return hWnd;
}
/****************************************************************************************/
/****************************************************************************************/
static LRESULT CALLBACK WindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  POINT           Point;
  RECT            Rect;
  NMLVDISPINFO   *pViewWndInfo;
  LVHITTESTINFO   HitTestInfo;
  HMENU           hMenu, hNewMenu;
  POINT           Position;
  DWORD           dwViewWndStyle;
  DWORD           dwIndex;
  PZFZPARAM       pZFZParam;
  LVFINDINFO      FindInfo;
  LVITEM          lvi;
  SYSTEMTIME      LocalTime;
  WORD            i;
  DWORD           dwTime;
  URM_USERINFO    UserInfo;
  URM_PASSINFO    PassInfo;
  BOOL            fIsZoomed;

  switch (uMsg)
  {
  case WM_COMMAND:
    switch (LOWORD (wParam))
    {
    case ID_USER_LOGIN:
      if (UserLoginDlg (NULL, &UserInfo))
      {
        if (CheckUser (&UserInfo))
        {
          if (UserInfo.RightMask == UR_SHUTDOWN)
            PostMessage (hWnd, WM_CLOSE, 0, 0);
          else
          {
            gUserRight = UserInfo.RightMask;
            strcpy (gaszLoginUser, UserInfo.aszUserName);

            GetWindowPosition (&Rect, &fIsZoomed);
            ShowWindow (hWnd, fIsZoomed?SW_SHOWMAXIMIZED:SW_SHOWNORMAL);
            if (gUserRight & UR_MODIFY)
              hNewMenu = LoadMenu (ghInst, MAKEINTRESOURCE (IDR_MAINMENU_MODIFY));
            else
              hNewMenu = LoadMenu (ghInst, MAKEINTRESOURCE (IDR_MAINMENU_VIEW));
            CheckMenuRadioItem (GetSubMenu (hNewMenu, 2), 0, 3, 3, MF_BYPOSITION);
            SetMenu (hWnd, hNewMenu);
            if (hNewMenu)
            {
              hMenu = GetSubMenu (hNewMenu, 0);
 
              if (gUserRight & UR_USERMANAGER)
              {
                LoadString (ghInst, IDS_MENUITEM_USERMAN, gaszString, sizeof gaszString);
                InsertMenu (hNewMenu, ID_USER_CHANGEPASS, MF_BYCOMMAND | MF_STRING,
                            ID_USER_USERMANAGER, gaszString);
              }
#ifdef LOOPBACKDEBUG
              hMenu = CreatePopupMenu ();
              if (hMenu)
              {
                AppendMenu (hMenu, MF_STRING, ID_TEST_CONNECT, "Directly &Connect");
                AppendMenu (hMenu, MF_STRING, ID_TEST_DISCONNECT, "Directly &Disconnect");
                AppendMenu (hMenu, MF_STRING, ID_TEST_RING, "&Ring");
                AppendMenu (hNewMenu, MF_STRING | MF_POPUP, (UINT_PTR)hMenu, "(&T)");
                DrawMenuBar (hWnd);
              }
            }
#else
            }
#endif
          }
        }
        else
        {
          LoadString (ghInst, IDS_PASS_LOGINFAIL, gaszString, sizeof gaszString);
          MessageBox (hWnd, gaszString, "", MB_OK);
        }
      }
      break;
    case ID_USER_LOGOUT:
      hMenu = GetMenu (hWnd);
      SetMenu (hWnd, NULL);
      DestroyMenu (hMenu);
      gUserRight = UR_NONE;
      ShowWindow (hWnd, SW_HIDE);
      break;
    case ID_USER_CHANGEPASS:
      if (ChangePasswordDlg (hWnd, &PassInfo))
      {
        if (ChangeUserPass (gaszLoginUser, &PassInfo))
        {
        }
      }
      break;
    case ID_USER_USERMANAGER:
      UserManagerDlg (ghInst, ghMainWnd);
      break;
    case ID_VIEW_LARGEMODE:
      dwViewWndStyle = GetWindowLong (ghViewWnd, GWL_STYLE);
      dwViewWndStyle &= ~LVS_TYPEMASK;
      dwViewWndStyle |= LVS_ICON;
      SetWindowLong (ghViewWnd, GWL_STYLE, dwViewWndStyle);
      CheckMenuRadioItem (GetSubMenu (GetMenu (hWnd), 2), 0, 3, 0, MF_BYPOSITION);
      break;
    case ID_VIEW_SMALLMODE:
      dwViewWndStyle = GetWindowLong (ghViewWnd, GWL_STYLE);
      dwViewWndStyle &= ~LVS_TYPEMASK;
      dwViewWndStyle |= LVS_SMALLICON;
      SetWindowLong (ghViewWnd, GWL_STYLE, dwViewWndStyle);
      CheckMenuRadioItem (GetSubMenu (GetMenu (hWnd), 2), 0, 3, 1, MF_BYPOSITION);
      break;
    case ID_VIEW_LISTMODE:
      dwViewWndStyle = GetWindowLong (ghViewWnd, GWL_STYLE);
      dwViewWndStyle &= ~LVS_TYPEMASK;
      dwViewWndStyle |= LVS_LIST;
      SetWindowLong (ghViewWnd, GWL_STYLE, dwViewWndStyle);
      CheckMenuRadioItem (GetSubMenu (GetMenu (hWnd), 2), 0, 3, 2, MF_BYPOSITION);
      break;
    case ID_VIEW_DETAILMODE:
      dwViewWndStyle = GetWindowLong (ghViewWnd, GWL_STYLE);
      dwViewWndStyle &= ~LVS_TYPEMASK;
      dwViewWndStyle |= LVS_REPORT;
      SetWindowLong (ghViewWnd, GWL_STYLE, dwViewWndStyle);
      CheckMenuRadioItem (GetSubMenu (GetMenu (hWnd), 2), 0, 3, 3, MF_BYPOSITION);
      break;
    case ID_ZFZ_NEW:        ZFZ_NewDialog ();                     break;
    case ID_ZFZ_DELETE:     ZFZ_DelDialog ();                     break;
    case ID_ZFZ_PROPERTIES: ZFZ_PropertiesDialog ();              break;
    case ID_ZFZ_GETSTATUS:  ZFZ_GetStatus ();                     break;
    case ID_ZFZ_DISPSTATUS: ZFZ_StatusDialog (NULL);              break;
    case ID_ZFZ_RUNPARAM:   ZFZ_ChangeDialog (NULL);              break;
    case ID_ZFZ_TURNON:     ZFZ_ThrnOn ();                        break;
    case ID_ZFZ_TURNOFF:    ZFZ_ThrnOff ();                       break;
    case ID_ZFZ_HISTORY:    ZFZ_HistoryWindow ();                 break;
    case ID_EXIT:           PostMessage (hWnd, WM_CLOSE, 0, 0l);  break;
    case ID_SET_MODEM:      ModemSetupDialog ();                  break;
    case ID_SET_CENTRE:     CentrePhoneDialog ();                 break;
    case ID_HLP_MANUAL:     
      HtmlHelp (hWnd, "YBD.CHM::/index.htm", HH_DISPLAY_TOPIC, 0);
      break;
    case ID_HLP_ABOUT:      AboutDialog ();                       break;
#ifdef LOOPBACKDEBUG
    case ID_TEST_CONNECT:   gfEnableAT___ = TRUE;                 break;
    case ID_TEST_DISCONNECT:gfEnableAT___ = FALSE;                break;
    case ID_TEST_RING:      gfPutRing = TRUE;                     break;
#endif
    }
    return 0;
  case WM_NOTIFY:
    pViewWndInfo = (NMLVDISPINFO *)lParam;
    if (pViewWndInfo->hdr.hwndFrom == ghViewWnd)
    {
      switch (pViewWndInfo->hdr.code)
      {
      case NM_DBLCLK:
        GetCursorPos (&Position);
        ScreenToClient (ghViewWnd, &Position);
        HitTestInfo.pt    = Position;
        HitTestInfo.flags = LVHT_ONITEMICON|LVHT_ONITEMLABEL;
        if (ListView_HitTest (ghViewWnd, &HitTestInfo) != -1)
          ZFZ_StatusDialog (NULL);
        break;
      case NM_RCLICK:
        GetCursorPos (&Position);
        ScreenToClient (ghViewWnd, &Position);
        HitTestInfo.pt    = Position;
        HitTestInfo.flags = LVHT_ONITEMICON|LVHT_ONITEMLABEL;
        hMenu = LoadMenu (ghInst, MAKEINTRESOURCE (IDR_ITEMPOPUP));
        GetCursorPos (&Position);
        if (ListView_HitTest (ghViewWnd, &HitTestInfo) != -1)
        {
          if (gpDirectConnectZFZParam)
            EnableMenuItem (hMenu, ID_ZFZ_DELETE, MF_GRAYED);
          TrackPopupMenu (GetSubMenu (hMenu, 0), TPM_LEFTALIGN|TPM_TOPALIGN,
                          Position.x, Position.y, 0, hWnd, NULL);
        }
        else
        {
          if (gpDirectConnectZFZParam)
            EnableMenuItem (hMenu, ID_ZFZ_NEW, MF_GRAYED);
          TrackPopupMenu (GetSubMenu (hMenu, 1), TPM_LEFTALIGN|TPM_TOPALIGN,
                          Position.x, Position.y, 0, hWnd, NULL);
        }
        DestroyMenu (hMenu);
        break;
      case LVN_GETDISPINFO:
        pViewWndInfo->item.pszText =
          LISTVIEW_ItemContent (pViewWndInfo->item.iSubItem,
                                (PZFZPARAM)pViewWndInfo->item.lParam);
        break;
      case LVN_COLUMNCLICK:
        ListView_SortItems (ghViewWnd, (PFNLVCOMPARE)ZFZPARAM_Compare,
                            ((LPNMLISTVIEW)lParam)->iSubItem);
        break;
      }
      break;
    }
    return 0;
  case WM_SIZE:
    if (ghDebugWnd)
    {
      SendMessage (ghStatusWnd, WM_SIZE, 0, 0);
      GetWindowRect (ghStatusWnd, &Rect);
      MoveWindow (ghViewWnd, 0, 0, LOWORD (lParam), HIWORD (lParam)-Rect.bottom+Rect.top-150, TRUE);
      MoveWindow (ghDebugWnd, 0, HIWORD (lParam)-Rect.bottom+Rect.top-150, LOWORD (lParam), 150, TRUE);
      SendMessage (ghStatusWnd, SB_SETPARTS, STATUSBARPARTS, (LPARAM)STATUSBARPARTWIDTH);
    }
    else
    {
      SendMessage (ghStatusWnd, WM_SIZE, 0, 0);
      GetWindowRect (ghStatusWnd, &Rect);
      MoveWindow (ghViewWnd, 0, 0, LOWORD (lParam), HIWORD (lParam)-Rect.bottom+Rect.top, TRUE);
      SendMessage (ghStatusWnd, SB_SETPARTS, STATUSBARPARTS, (LPARAM)STATUSBARPARTWIDTH);
    }
    GetWindowRect (hWnd, &Rect);
    SetWindowPosition (&Rect, IsZoomed (hWnd));
    return 0;
  case WM_SETFOCUS:
    SetFocus (ghViewWnd);
    return 0;
  case WM_CREATE:
    SetTimer (hWnd, 2, 200, NULL);
    break;
  case WM_DESTROY:
    PostQuitMessage (0);
    break;
  case WM_CLOSE:
    LoadString (ghInst, IDS_QUITCONFIRM, gaszString, sizeof gaszString);
    if (MessageBox (hWnd, gaszString, "", MB_YESNO) != IDYES) return 0;
    StopCommunicationThread ();
    Sleep (50);
    if (IsStopCommunicationThread ())
      DestroyWindow (hWnd);
    else
    {
      ghTerminateWnd = CreateDialog (ghInst, MAKEINTRESOURCE (IDD_TERMINATETHREAD), hWnd,
                                     TerminateDlgProc);
      ShowWindow (ghTerminateWnd, SW_SHOW);
      SetTimer (hWnd, 1, 100, NULL);
    }
    return 0;
  case WM_TIMER:
    switch (wParam)
    {
    case 1:
      if (IsStopCommunicationThread ())
      {
        KillTimer (hWnd, 1);
        EndDialog (ghTerminateWnd, IDOK);
        DestroyWindow (hWnd);
      }
      break;
    case 2:
      if (gpAutoPopupParam)
      {
        pZFZParam = gpAutoPopupParam;
        gpAutoPopupParam = NULL;
        ZFZ_StatusDialog (pZFZParam);
      }
      break;
    }
    return 0;
  case WM_MODEMSTATUS:
    STATUSBAR_ShowModemStatus ((WORD)wParam, lParam);
    return 0;
  case WM_TASKSTATUS:
    STATUSBAR_ShowTaskStatus ((PCSTR)wParam, (TASKID)lParam);
    return 0;
  case WM_DIRECTCONNECT:
    if (gpDirectConnectZFZParam)
      pZFZParam = gpDirectConnectZFZParam;
    else
    {
      pZFZParam = ZFZPARAM_malloc ((BYTE)wParam);
      if (pZFZParam)
      {
        ListView_DeleteAllItems (ghViewWnd);
        gpDirectConnectZFZParam = pZFZParam;
        memset (pZFZParam, 0, ZFZPARAM_SizeOf ((BYTE)wParam));
        pZFZParam->bChannels = (BYTE)wParam;
        LoadString (ghInst, IDS_DIRECTCONNECTZFZNAME,
                    pZFZParam->aszName, sizeof pZFZParam->aszName);
        GetLocalTime (&LocalTime);
        pZFZParam->CreateDate.wYear   = LocalTime.wYear;
        pZFZParam->CreateDate.bMonth  = (BYTE)LocalTime.wMonth;
        pZFZParam->CreateDate.bDay    = (BYTE)LocalTime.wDay;
        pZFZParam->CreateDate.bHour   = (BYTE)LocalTime.wHour;
        pZFZParam->CreateDate.bMinute = (BYTE)LocalTime.wMinute;
        pZFZParam->VWImageIndex       = VWI_CONNECT;
        pZFZParam->TaskID             = TASKID_SETPHONE1 | TASKID_GETSTATUS |
                                        TASKID_ALARM;
        LISTVIEW_AddItem (ghViewWnd, 0, pZFZParam);
        ListView_SetItemState (ghViewWnd, 0, LVIS_FOCUSED, LVIS_FOCUSED);
        EnableZFZMenu ();
      }
    }
    return (LRESULT)pZFZParam;
  case WM_QUITDIRECTCONNECT:
    free (gpDirectConnectZFZParam);
    gpDirectConnectZFZParam = NULL;
    ListView_DeleteAllItems (ghViewWnd);
    for (i=0; i<gdwZFZParams; i++)
    {
      gapZFZParam[i]->TaskID &= ~TASKID_LOCKNODE;
      LISTVIEW_AddItem (ghViewWnd, i, gapZFZParam[i]);
    }
    if (gdwZFZParams)
      ListView_SetItemState (ghViewWnd, 0, LVIS_FOCUSED, LVIS_FOCUSED);
    EnableZFZMenu ();
    return 0;
  case WM_REMALLOCNODE:
    pZFZParam = ZFZPARAM_malloc ((BYTE)wParam);
    if (pZFZParam)
    {
      memcpy (pZFZParam, gapZFZParam[gdwTaskNodeIndex], ZFZPARAM_SizeOf (0));
      pZFZParam->bChannels = (BYTE)wParam;
      FindInfo.flags  = LVFI_PARAM;
      FindInfo.lParam = (LPARAM)gapZFZParam[gdwTaskNodeIndex];
      lvi.iItem    = ListView_FindItem (ghViewWnd, -1, &FindInfo);
      lvi.iSubItem = 0;
      lvi.mask     = LVIF_TEXT|LVIF_PARAM|LVIF_IMAGE;
      lvi.iImage   = pZFZParam->VWImageIndex = VWI_NORMAL;
      lvi.pszText  = LPSTR_TEXTCALLBACK;
      lvi.lParam   = (LPARAM)pZFZParam;
      ListView_SetItem (ghViewWnd, &lvi);
      ListView_Update (ghViewWnd, lvi.iItem);
      free (gapZFZParam[gdwTaskNodeIndex]);
      gapZFZParam[gdwTaskNodeIndex] = pZFZParam;
    }
    else
      pZFZParam = gapZFZParam[gdwTaskNodeIndex];
    return (LRESULT)pZFZParam;
  case WM_GETNEXTTASKNODE:
    if (gpDirectConnectZFZParam)
    {
      if (gpDirectConnectZFZParam->TaskID)
        return (LRESULT)gpDirectConnectZFZParam;
    }
    else
    {
      if (gdwTaskNodeIndex < gdwZFZParams)
      {
        dwIndex = gdwTaskNodeIndex;
        do
        {
          dwIndex ++;
          if (dwIndex >= gdwZFZParams) dwIndex = 0;
          if (gapZFZParam[dwIndex]->TaskID)
          {
            gdwTaskNodeIndex = dwIndex;
            gapZFZParam[dwIndex]->TaskID |= TASKID_LOCKNODE;
            return (LRESULT)gapZFZParam[dwIndex];
          }
        } while (dwIndex != gdwTaskNodeIndex);
        gdwTaskNodeIndex = gdwZFZParams;
      }
    }
    return (LRESULT)NULL;
  case WM_AUTOPOPUPSTATUSBOX:
    gpAutoPopupParam = (PZFZPARAM)lParam;
    return 0;
  case WM_UPDATETASKNODE:
    pZFZParam = gapZFZParam[gdwTaskNodeIndex];
    FindInfo.flags  = LVFI_PARAM;
    FindInfo.lParam = (LPARAM)pZFZParam;
    lvi.iItem    = ListView_FindItem (ghViewWnd, -1, &FindInfo);
    lvi.iSubItem = 0;
    lvi.mask     = LVIF_IMAGE;
    if (pZFZParam->bChannels == 0)
      lvi.iImage = pZFZParam->VWImageIndex = VWI_INSTALL;
    else if (pZFZParam->ErrorCode != ERRORCODE_NONE)
      lvi.iImage = pZFZParam->VWImageIndex = VWI_ERROR;
    else
      lvi.iImage = pZFZParam->VWImageIndex = VWI_NORMAL;
    ListView_SetItem (ghViewWnd, &lvi);
    ListView_Update (ghViewWnd, lvi.iItem);
    SaveControlHistory (time (&dwTime), pZFZParam);
    SaveZFZList ();
    return 0;
  case WM_TURNON:
    if (gpDirectConnectZFZParam)
      pZFZParam = gpDirectConnectZFZParam;
    else
      pZFZParam = gapZFZParam[gdwTaskNodeIndex];
    FindInfo.flags  = LVFI_PARAM;
    FindInfo.lParam = (LPARAM)pZFZParam;
    lvi.iItem    = ListView_FindItem (ghViewWnd, -1, &FindInfo);
    lvi.iSubItem = 0;
    lvi.mask     = LVIF_TEXT|LVIF_PARAM|LVIF_IMAGE;
    if (pZFZParam->bChannels == 0)
      lvi.iImage = pZFZParam->VWImageIndex = VWI_INSTALL;
    else if (pZFZParam->ErrorCode != ERRORCODE_NONE)
      lvi.iImage = pZFZParam->VWImageIndex = VWI_ERROR;
    else
      lvi.iImage = pZFZParam->VWImageIndex = VWI_NORMAL;
    lvi.pszText  = LPSTR_TEXTCALLBACK;
    lvi.lParam   = (LPARAM)pZFZParam;
    ListView_SetItem (ghViewWnd, &lvi);
    ListView_Update (ghViewWnd, lvi.iItem);
    return 0;
  case WM_TURNOFF:
    if (gpDirectConnectZFZParam)
      pZFZParam = gpDirectConnectZFZParam;
    else
      pZFZParam = gapZFZParam[gdwTaskNodeIndex];
    FindInfo.flags  = LVFI_PARAM;
    FindInfo.lParam = (LPARAM)pZFZParam;
    lvi.iItem    = ListView_FindItem (ghViewWnd, -1, &FindInfo);
    lvi.iSubItem = 0;
    lvi.mask     = LVIF_TEXT|LVIF_PARAM|LVIF_IMAGE;
    lvi.iImage   = pZFZParam->VWImageIndex = VWI_TURNOFF;
    lvi.pszText  = LPSTR_TEXTCALLBACK;
    lvi.lParam   = (LPARAM)pZFZParam;
    ListView_SetItem (ghViewWnd, &lvi);
    ListView_Update (ghViewWnd, lvi.iItem);
    return 0;
  case WM_SEARCHTASKNODE:
    for (i=0; i<gdwZFZParams; i++)
    {
      if (strcmp (gapZFZParam[i]->aszPhone, (LPCSTR)lParam)==0)
      {
        gapZFZParam[i]->TaskID |= TASKID_GETSTATUS | TASKID_LOCKNODE;
        gdwTaskNodeIndex = i;
        return (LRESULT)gapZFZParam[i];
      }
    }
    return (LRESULT)NULL;
  case WM_SHELLICONNOTIFY:
    switch (lParam)
    {
      case WM_RBUTTONDOWN:
	GetCursorPos (&Point); 
        if (gUserRight == UR_NONE)
          hMenu = LoadMenu (ghInst, MAKEINTRESOURCE (IDR_ICONMENU_OUT));
        else
          hMenu = LoadMenu (ghInst, MAKEINTRESOURCE (IDR_ICONMENU_IN));
	TrackPopupMenu(GetSubMenu (hMenu, 0), 0, Point.x, Point.y, 0, hWnd, NULL);
        DestroyMenu (hMenu);
	break;
      case WM_LBUTTONDOWN:
        if (gUserRight == UR_NONE)
          PostMessage (hWnd, WM_COMMAND, ID_USER_LOGIN, 0l);
        break;
    }
    return 0;
  }
  return DefWindowProc (hWnd, uMsg, wParam, lParam);
}
/****************************************************************************************/
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
  INITCOMMONCONTROLSEX  ics;
  NOTIFYICONDATA  NotifyIconData;
  WNDCLASSEX      WndClass;
  HWND            hWnd;
  HACCEL          hAccelView, hAccelModify;
  MSG             Msg;
  RECT            WndRect;
  BOOL            fIsZoomed;

  ghInst = hInst;

  if (!LoadUserList ())
    if (!AddAdmin ())
      return 0;

  ghDebugWndDll = LoadLibrary ("debugwnd.dll");
  if (ghDebugWndDll)
  {
    gpfunSTATWND_CreateWnd = (STATWND_CREATEWND)GetProcAddress (ghDebugWndDll,
                                                                "_STATWND_CreateWnd@12");
    gpfunSTATWND_StreamByte = (STATWND_STREAMBYTE)GetProcAddress (ghDebugWndDll,
                                                                  "_STATWND_StreamByte@12");
    if (gpfunSTATWND_CreateWnd == NULL || gpfunSTATWND_StreamByte == NULL)
    {
      FreeLibrary (ghDebugWndDll);
      ghDebugWndDll = NULL;
    }
  }

  ics.dwSize = sizeof ics;
  ics.dwICC  = ICC_LISTVIEW_CLASSES | ICC_BAR_CLASSES;
  if (!InitCommonControlsEx (&ics))
  {
    LoadString (ghInst, IDS_COMCTL32_ERR, gaszString, sizeof gaszString);
    LoadString (ghInst, IDS_SYSERR_TITLE, gaszString2, sizeof gaszString2);
    MessageBox (NULL, gaszString, gaszString2, MB_OK|MB_ICONERROR);
    return 0;
  }

  WndClass.cbSize        = sizeof (WNDCLASSEX);
  WndClass.style         = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
  WndClass.lpfnWndProc   = WindowProc;
  WndClass.cbClsExtra    = 0;
  WndClass.cbWndExtra    = 0;
  WndClass.hInstance     = hInst;
  WndClass.hIcon         = LoadIcon (hInst, MAKEINTRESOURCE (IDI_MAIN));
  WndClass.hCursor       = LoadCursor (NULL, IDC_ARROW);
  WndClass.hbrBackground = NULL;
  WndClass.lpszMenuName  = NULL;
  WndClass.lpszClassName = MAINWNDCLASSNAME;
  WndClass.hIconSm       = NULL;
  if (!RegisterClassEx (&WndClass)) return 0;

  WndClass.cbSize        = sizeof (WNDCLASSEX);
  WndClass.style         = CS_HREDRAW | CS_VREDRAW;
  WndClass.lpfnWndProc   = HistoryWindowProc;
  WndClass.cbClsExtra    = 0;
  WndClass.cbWndExtra    = 0;
  WndClass.hInstance     = ghInst;
  WndClass.hIcon         = LoadIcon (ghInst, MAKEINTRESOURCE (IDI_MAIN));
  WndClass.hCursor       = LoadCursor (NULL, IDC_ARROW);
  WndClass.hbrBackground = NULL;
  WndClass.lpszMenuName  = NULL;
  WndClass.lpszClassName = HISTORYWNDCLASSNAME;
  WndClass.hIconSm       = NULL;
  if (!RegisterClassEx (&WndClass)) return 0;

  WndClass.cbSize        = sizeof (WNDCLASSEX);
  WndClass.style         = 0;
  WndClass.lpfnWndProc   = HISWND_StatusWndProc;
  WndClass.cbClsExtra    = 0;
  WndClass.cbWndExtra    = 0;
  WndClass.hInstance     = hInst;
  WndClass.hIcon         = NULL;
  WndClass.hCursor       = LoadCursor (NULL, IDC_ARROW);
  WndClass.hbrBackground = GetStockObject(WHITE_BRUSH);
  WndClass.lpszMenuName  = NULL;
  WndClass.lpszClassName = STATUSWNDCLASSNAME;
  WndClass.hIconSm       = NULL;
  if (!RegisterClassEx (&WndClass)) return 0;

  GetWindowPosition (&WndRect, &fIsZoomed);
  LoadString (hInst, IDS_MAINWNDTITLE, gaszString, sizeof gaszString);
  ghMainWnd = hWnd = CreateWindowEx (WS_EX_APPWINDOW, MAINWNDCLASSNAME, gaszString,
                                     WS_OVERLAPPEDWINDOW, WndRect.left, WndRect.top,
                                     WndRect.right-WndRect.left,
                                     WndRect.bottom-WndRect.top,
                                     NULL, NULL, hInst, NULL);
  if (hWnd == NULL) return 0;

  NotifyIconData.cbSize  = sizeof (NotifyIconData);
  NotifyIconData.hWnd    = hWnd;
  NotifyIconData.uID     = 0;
  NotifyIconData.uFlags  = NIF_ICON | NIF_TIP | NIF_MESSAGE;
  NotifyIconData.hIcon   = LoadIcon (hInst, MAKEINTRESOURCE (IDI_MAIN));
  NotifyIconData.uCallbackMessage = WM_SHELLICONNOTIFY;
  strcpy (NotifyIconData.szTip, gaszString);

  DialogBox (hInst, MAKEINTRESOURCE (IDD_OPENING), hWnd, OpeningDlgProc);

  ghViewWnd = CreateListViewControl (hInst, hWnd);
  if (ghViewWnd == NULL)
  {
    DestroyWindow (hWnd);
    return 0;
  }
  EnableZFZMenu ();

  ghStatusWnd = CreateWindowEx (0, STATUSCLASSNAME, "", WS_CHILD, 0, 0, 0, 0,
                                hWnd, NULL, hInst, NULL);
  if (ghStatusWnd == NULL)
  {
    DestroyWindow (hWnd);
    return 0;
  }
  ShowWindow (ghStatusWnd, nCmdShow);

  hAccelView   = LoadAccelerators (hInst, MAKEINTRESOURCE (IDR_MAINACCEL_VIEW));
  hAccelModify = LoadAccelerators (hInst, MAKEINTRESOURCE (IDR_MAINACCEL_MODIFY));

  if (ghDebugWndDll)
    ghDebugWnd = gpfunSTATWND_CreateWnd (hInst, hWnd,
                                         STFLAG_INSPECTSTREAM|STFLAG_ALWAYSFOCUS);

  Shell_NotifyIcon (NIM_ADD, &NotifyIconData);
  PostMessage (hWnd, WM_COMMAND, ID_USER_LOGIN, 0l);

  ShowWindow (hWnd, SW_HIDE);
  UpdateWindow (hWnd);

  if (gaszCentrePhone[0] == '\0')
  {
    if (!CentrePhoneDialog ())
    {
      DestroyWindow (hWnd);
      return 0;
    }
    SaveZFZList ();
  }

  GetModemParam (&gModemParam);
StartThread:
  switch (StartCommunicationThread (gModemParam.wCommPort))
  {
  case STEC_OPENCOMMPORT:
    LoadString (hInst, IDS_MODEMERR_TITLE, gaszString, sizeof gaszString);
    LoadString (hInst, IDS_MODEMERR_01, gaszString2, sizeof gaszString2);
    MessageBox (hWnd, gaszString2, gaszString, MB_OK|MB_ICONERROR);
    if (!ModemSetupDialog ())
    {
      DestroyWindow (hWnd);
      return 0;
    }
    goto StartThread;
  case STEC_COMMPROTERR:
    break;
  case STEC_STARTTHREAD:
    LoadString (hInst, IDS_SYSERR_TITLE, gaszString, sizeof gaszString);
    LoadString (hInst, IDS_THREAD, gaszString2, sizeof gaszString2);
    MessageBox (hWnd, gaszString2, gaszString, MB_OK|MB_ICONERROR);
    DestroyWindow (hWnd);
    return 0;
  }
  STATUSBAR_ShowTaskStatus ("", TASKID_IDLE);

  while (GetMessage (&Msg, NULL, 0, 0))
  {
    if (ghActivateDialog == NULL || !IsDialogMessage (ghActivateDialog, &Msg))
    {
      if (gUserRight == UR_NONE)
        TranslateMessage (&Msg);
      else if (gUserRight & UR_MODIFY)
      {
        if (!TranslateAccelerator (hWnd, hAccelModify, &Msg)) TranslateMessage (&Msg);
      }
      else
      {
        if (!TranslateAccelerator (hWnd, hAccelView, &Msg)) TranslateMessage (&Msg);
      }
      DispatchMessage (&Msg);
    }
  }

  NotifyIconData.cbSize  = sizeof (NotifyIconData);
  NotifyIconData.hWnd    = hWnd;
  NotifyIconData.uID     = 0;
  NotifyIconData.uFlags  = 0;
  Shell_NotifyIcon (NIM_DELETE, &NotifyIconData);

  DestroyAcceleratorTable (hAccelModify);
  DestroyAcceleratorTable (hAccelView);
  ReleaseCommunicationThread ();

  if (ghDebugWndDll) FreeLibrary (ghDebugWndDll);

  return Msg.wParam;
}
/****************************************************************************************/
