/*****************************************************************************************
 * File name     : COMM.C
 * Author        : Han Zhi Qiang
 * Module        : 
 * Purpose       : 
 * Create Date   : June 30, 2001 - 23:42
 * Modify history: 
 *****************************************************************************************
 */

#define ALLOCATE_COMM_VARIABLE_MEMORY

#include <stdio.h>
#include <stdlib.h>

#include <windows.h>
#include "ui.h"
#include "comm.h"

/*****************************************************************************************
 * MACRO functions declare
 ****************************************************************************************/
#define MODEMCMD_INIT1                  "ATZ\r"
#define MODEMCMD_INIT2                  "ATV1\r"
#define MODEMCMD_INIT3                  "ATE0\r"
#define MODEMCMD_READMSG                "AT+CMGL=\"REC UNREAD\"\r"
#define MODEMCMD_DELETEMSG              "AT+CMGD=1,3\r"
#define MODEMCMD_SENDMSG                "AT+CMGS=\"%s\"\r"

#define NODEMNOFIY_OK                   "OK"
#define NODEMNOFIY_ERROR                "ERROR"
#define NODEMNOFIY_INCOMEMSG            "+CMTI: \"SM\""
#define NODEMNOFIY_MSGHEADER            "+CMGL:"

/*****************************************************************************************
 * Identifier constant declare
 ****************************************************************************************/

/*****************************************************************************************
 * Data structure declare
 ****************************************************************************************/
typedef enum
{
  MS_FREE,
  MS_INIT1,
  MS_INIT2,
  MS_INIT3,
  MS_DELETEMSG,
  MS_SENDPACKET,
  MS_MSGHEADER,
  MS_MSGBODY,
  MS_ANSWER
} MODEMSTATE;
/*****************************************************************************************
 * Local variable declare
 ****************************************************************************************/
static HANDLE     ghCommDev = INVALID_HANDLE_VALUE;

// Thread control variable
static HANDLE     ghCommunicationThreadHandle = NULL;
static BOOL       gfStopCommunicationThread;

static MODEMSTATE gModemState;

static const CHAR aszTranslateTable[] =
{
  'L', '9', 'I', '6', 'b', 'j', 'u', 'P',
  'R', 'K', '5', 'q', '4', 'X', 'C', 'N'
};

/*****************************************************************************************
 * Local functions declare
 ****************************************************************************************/

/****************************************************************************************/
static BOOL OpenCommPort(int PortNo)
{
  char          Buf[100];
  DCB           dcb;
  COMMTIMEOUTS  CommTimeouts;

  if (PortNo < 1 || PortNo > 2) return FALSE;

  wsprintf (Buf, "COM%d", PortNo);

  ghCommDev = CreateFile (Buf, GENERIC_WRITE|GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
  if (ghCommDev != INVALID_HANDLE_VALUE)
  {
    wsprintf (Buf, "COM%d:baud=9600 parity=N data=8 stop=1", PortNo);
    if  (GetCommState (ghCommDev, &dcb))
    {
      //dcb.fDtrControl = RTS_CONTROL_ENABLE;
      //dcb.fRtsControl = RTS_CONTROL_DISABLE;
      if (BuildCommDCB (Buf, &dcb))
      {
        if (SetCommState (ghCommDev, &dcb))
        {
          memset (&CommTimeouts, 0, sizeof(COMMTIMEOUTS));
          CommTimeouts.ReadTotalTimeoutMultiplier = 2;
          if (SetCommTimeouts (ghCommDev, &CommTimeouts))
          {
            return TRUE;
          }
        }
      }
    }
    CloseHandle (ghCommDev);
  }
  return FALSE;
}
/****************************************************************************************/
static int CloseCommPort(void)
{
  HANDLE hHandle = ghCommDev;
  ghCommDev = INVALID_HANDLE_VALUE;
  return CloseHandle (hHandle);
}
/****************************************************************************************/
/****************************************************************************************/
BOOL DebugReadFile(HANDLE hFile, PBYTE pbBuf, DWORD dwBytes, PDWORD pdwActBytes,
                      LPOVERLAPPED lpOverlapped)
{
  BOOL            fResult;
  DWORD           i;
  
  fResult = ReadFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped);
  if (ghDebugWnd)
  {
    for (i=0; i<*pdwActBytes; i++)
      gpfunSTATWND_StreamByte (ghDebugWnd, OMPT_DOWNLOAD, pbBuf[i]);
  }
  return fResult;
}
/****************************************************************************************/
BOOL DebugWriteFile(HANDLE hFile, const BYTE *pbBuf, DWORD dwBytes, PDWORD pdwActBytes,
                        LPOVERLAPPED lpOverlapped)
{
  BOOL            fResult;
  DWORD           i;
  
  fResult = WriteFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped);
  if (ghDebugWnd)
  {
    for (i=0; i<*pdwActBytes; i++)
      gpfunSTATWND_StreamByte (ghDebugWnd, OMPT_UPLOAD, pbBuf[i]);
  }
  return fResult;
}
/****************************************************************************************/
#define ReadFile(hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped) \
        DebugReadFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped)
#define WriteFile(hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped) \
        DebugWriteFile (hFile, pbBuf, dwBytes, pdwActBytes, lpOverlapped)
/****************************************************************************************/
/****************************************************************************************/
static BOOL SendModemCommand(const CHAR *pcszModemCommand)
{
  DWORD           dwActualBytes;
  DWORD           dwBytes;

  dwBytes = strlen (pcszModemCommand);
  return WriteFile (ghCommDev, pcszModemCommand, dwBytes, &dwActualBytes, NULL) &&
         dwActualBytes == dwBytes;
}
/****************************************************************************************/
static BOOL SendPacket(const char *pcszPhoneNo, const BYTE *pcbData, BYTE bDataSize)
{
  CHAR            aszBuf[1024];
  WORD            wPos, i;
  BYTE            bRand, bCheckSum;

  wsprintf (aszBuf, MODEMCMD_SENDMSG, pcszPhoneNo);
  if (!SendModemCommand (aszBuf)) return FALSE;

  bRand = (BYTE)(GetTickCount () & 0xf);
  wPos = 0;
  bCheckSum = 0;
  
  aszBuf[wPos++] = aszTranslateTable[bRand];
  aszBuf[wPos++] = aszTranslateTable[(bDataSize&0xf)^bRand];
  aszBuf[wPos++] = aszTranslateTable[(bDataSize>>4)^bRand];
  for (i=0; i<bDataSize; i++)
  {
    bCheckSum += pcbData[i];
    aszBuf[wPos++] = aszTranslateTable[(pcbData[i]&0xf)^bRand];
    aszBuf[wPos++] = aszTranslateTable[(pcbData[i]>>4)^bRand];
  }
  aszBuf[wPos++] = aszTranslateTable[(bCheckSum&0xf)^bRand];
  aszBuf[wPos++] = aszTranslateTable[(bCheckSum>>4)^bRand];
  aszBuf[wPos++] = '\x1a';
  aszBuf[wPos++] = '\x0d';
  aszBuf[wPos]   = '\x00';
  if (!SendModemCommand (aszBuf)) return FALSE;
  return TRUE;
}
/***************************************************************************************/
static BYTE GetCharValue(CHAR Char)
{
  BYTE            i;
  for (i=0; i<16; i++)
    if (Char == aszTranslateTable[i]) break;
  return i;
}
/****************************************************************************************/
WORD DecodePacket(const char *pcszSource, unsigned char *pbResult)
{
  BYTE            bRand, bLen, bCheckSum;
  BYTE            bHigh, bLow;
  WORD            i;

  if (pcszSource == NULL && *pcszSource == 0) return 0;
  bRand = GetCharValue (*(pcszSource++));
  if (bRand >= 16) return 0;
  
  if (*pcszSource == 0 || *(pcszSource+1) == 0) return 0;
  bLow  = GetCharValue (*(pcszSource++));
  bHigh = GetCharValue (*(pcszSource++));
  if (bHigh >= 16 || bLow >= 16) return 0;
  bLen = (bHigh ^ bRand) * 16 + (bLow ^ bRand);

  bCheckSum = 0;
  for (i=0; i<bLen; i++)
  {
    if (*pcszSource == 0 || *(pcszSource+1) == 0) return 0;
    bLow  = GetCharValue (*(pcszSource++));
    bHigh = GetCharValue (*(pcszSource++));
    if (bHigh >= 16 || bLow >= 16) return 0;
    pbResult[i] = (bHigh ^ bRand) * 16 + (bLow ^ bRand);
    bCheckSum += pbResult[i];
  }

  if (*pcszSource == 0 || *(pcszSource+1) == 0) return 0;
  bLow  = GetCharValue (*(pcszSource++));
  bHigh = GetCharValue (*(pcszSource++));
  if (bHigh >= 16 || bLow >= 16) return 0;
  
  if (bCheckSum == ((bHigh ^ bRand) * 16 + (bLow ^ bRand)))
    return bLen;
  else
    return 0;

  return bLen;
}
/****************************************************************************************/
/****************************************************************************************/
static DWORD WINAPI CommunicationThread(LPVOID lpParameter)
{
  static WORD     wLineBufPointer = 0;
  static CHAR     aszLineBuf[1024];
  static BYTE     abPacket[256];
  static CHAR     aszPhoneNo[256];
  static WORD     wRecvMsgs;
  BYTE            bData;
  DWORD           dwActualBytes;
  int             i, j;
  RECVMSGDATA     RecvMsgData;
  DWORD           dwLookForDelay = GetTickCount ();
  PZFZPARAM       pZFZParam;
  WORD            wRetryTimes;

  gModemState = MS_INIT1;
  wRetryTimes = 1;
  while (1)
  {
    if (!ReadFile (ghCommDev, &bData, 1, &dwActualBytes, NULL)) goto CommError;
    if (dwActualBytes == 1)
    {
      if (bData == '\n')
      {
        if (wLineBufPointer)
        {
          for (; wLineBufPointer>0; wLineBufPointer--)
            if (aszLineBuf[wLineBufPointer-1] != '\r' &&
                !isspace (aszLineBuf[wLineBufPointer-1]))
              break;
          aszLineBuf[wLineBufPointer] = '\0';
          printf ("%s\n", aszLineBuf[wLineBufPointer]);

          switch (gModemState)
          {
          case MS_FREE:
            if (memcmp (aszLineBuf, NODEMNOFIY_INCOMEMSG, strlen (NODEMNOFIY_INCOMEMSG)) == 0)
            {
              if (!SendModemCommand (MODEMCMD_READMSG)) goto CommError;
              wRecvMsgs = 0;
              gModemState = MS_MSGHEADER;
            }
            break;
          case MS_INIT1:
            if (memcmp (aszLineBuf, NODEMNOFIY_OK, strlen (NODEMNOFIY_OK)) == 0)
            {
              if (!SendModemCommand (MODEMCMD_INIT2)) goto CommError;
              wRetryTimes = 1;
              gModemState = MS_INIT2;
            }
            else if (memcmp (aszLineBuf, NODEMNOFIY_ERROR, strlen (NODEMNOFIY_ERROR)) == 0)
            {
              wRetryTimes ++;
              if (wRetryTimes>3) goto CommError;
              if (!SendModemCommand (MODEMCMD_INIT1)) goto CommError;
            }
            break;         
          case MS_INIT2:
            if (memcmp (aszLineBuf, NODEMNOFIY_OK, strlen (NODEMNOFIY_OK)) == 0)
            {
              if (!SendModemCommand (MODEMCMD_INIT3)) goto CommError;
              wRetryTimes = 1;
              gModemState = MS_INIT3;
            }
            else if (memcmp (aszLineBuf, NODEMNOFIY_ERROR, strlen (NODEMNOFIY_ERROR)) == 0)
            {
              wRetryTimes ++;
              if (wRetryTimes>3) goto CommError;
              if (!SendModemCommand (MODEMCMD_INIT2)) goto CommError;
            }
            break;         
          case MS_INIT3:
            if (memcmp (aszLineBuf, NODEMNOFIY_OK, strlen (NODEMNOFIY_OK)) == 0)
            {
              wRetryTimes = 0;
              gModemState = MS_FREE;
            }
            else if (memcmp (aszLineBuf, NODEMNOFIY_ERROR, strlen (NODEMNOFIY_ERROR)) == 0)
            {
              wRetryTimes ++;
              if (wRetryTimes>3) goto CommError;
              if (!SendModemCommand (MODEMCMD_INIT2)) goto CommError;
            }
            break;         
          case MS_MSGHEADER:
            if (memcmp (aszLineBuf, NODEMNOFIY_MSGHEADER, strlen (NODEMNOFIY_MSGHEADER)) == 0)
            {
              for (i=strlen (NODEMNOFIY_MSGHEADER), j=0; aszLineBuf[i] && j<3; i++)
              {
                if (aszLineBuf[i] == '"') j ++;
              }
              for (j=0; aszLineBuf[i] && aszLineBuf[i] != '"'; i++)
                aszPhoneNo[j++] = aszLineBuf[i];
              aszPhoneNo[j] = 0;
              wRecvMsgs ++;
              gModemState = MS_MSGBODY;
            }
            else if (memcmp (aszLineBuf, NODEMNOFIY_OK, strlen (NODEMNOFIY_OK)) == 0 ||
                     memcmp (aszLineBuf, NODEMNOFIY_ERROR, strlen (NODEMNOFIY_ERROR)) == 0)
            {
              if (wRecvMsgs) 
              {
                if (!SendModemCommand (MODEMCMD_DELETEMSG)) goto CommError;
                gModemState = MS_DELETEMSG;
                wRetryTimes = 1;
              }
              else
              {
                wRetryTimes = 0;
                gModemState = MS_FREE;
              }
            }
            break;
          case MS_MSGBODY:
            gModemState = MS_MSGHEADER;

            RecvMsgData.pcszPhoneNo   = aszPhoneNo;
            RecvMsgData.pbContent     = abPacket;
            RecvMsgData.dwContentSize = DecodePacket (aszLineBuf, abPacket);
            
            if (RecvMsgData.dwContentSize != 0)
              SendMessage (ghMainWnd, WM_RECEIVEDSMS, 0, (LPARAM)&RecvMsgData);
            break;
          case MS_SENDPACKET:
            if (memcmp (aszLineBuf, NODEMNOFIY_OK, strlen (NODEMNOFIY_OK)) == 0 ||
                memcmp (aszLineBuf, NODEMNOFIY_ERROR, strlen (NODEMNOFIY_ERROR)) == 0)
            {
              gModemState = MS_FREE;
              wRetryTimes = 0;
            }
            break;
          case MS_DELETEMSG:
            if (memcmp (aszLineBuf, NODEMNOFIY_OK, strlen (NODEMNOFIY_OK)) == 0 ||
                memcmp (aszLineBuf, NODEMNOFIY_ERROR, strlen (NODEMNOFIY_ERROR)) == 0)
            {
              wRetryTimes = 0;
              gModemState = MS_FREE;
            }
            break;
          }
          wLineBufPointer = 0;
        }
      }
      else
        aszLineBuf[wLineBufPointer++] = bData;
    }
    
    if (gfStopCommunicationThread) break;

    if (GetTickCount () - dwLookForDelay > 2000)
    {
      switch (gModemState)
      {
      case MS_FREE:
        dwLookForDelay = GetTickCount ();
        pZFZParam = (PZFZPARAM)SendMessage (ghMainWnd, WM_GETNEXTTASKNODE, 0, 0);
        if (pZFZParam)
        {
          if (pZFZParam->TaskID & TASKID_GETPROPERTIES)
          {
            abPacket[0] = 0x01;
            strcpy (abPacket+1, gaszCentrePhone);
            if (!SendPacket (pZFZParam->aszPhone, abPacket, 
                             (BYTE)(strlen (gaszCentrePhone)+1))) goto CommError;
            gModemState = MS_SENDPACKET;
          }
          else if (pZFZParam->TaskID & TASKID_GETPARAM)
          {
            abPacket[0] = 0x04;
            if (!SendPacket (pZFZParam->aszPhone, abPacket, 1)) goto CommError;
            gModemState = MS_SENDPACKET;
          }
          else
          {
            abPacket[0] = 0x03;
                                            abPacket[1]  = 0x00;
            if (pZFZParam->fDisableAlarm)   abPacket[1] |= 0x01;
            if (!pZFZParam->fPowerOnStatus) abPacket[1] |= 0x02;
            for (i=0, j=2; i<pZFZParam->bChannels; i++)
            {
#if defined (CDMA) || defined (FIXEDFREQUENCY)
              abPacket[j++] = pZFZParam->ChannelData[i].wFrequency & 0xff;
              abPacket[j++] = (pZFZParam->ChannelData[i].wFrequency >> 8) & 0xff;
#else
              abPacket[j++] = (pZFZParam->ChannelData[i].wFrequency+0x180) & 0xff;
              abPacket[j++] = ((pZFZParam->ChannelData[i].wFrequency+0x180) >> 8) & 0xff;
#endif
              abPacket[j++] = pZFZParam->ChannelData[i].bUpGain;
              abPacket[j++] = pZFZParam->ChannelData[i].bDownGain;
              abPacket[j++] = pZFZParam->ChannelData[i].bUpPower;
              abPacket[j++] = pZFZParam->ChannelData[i].bDownPower;
            }
            if (!SendPacket (pZFZParam->aszPhone, abPacket, (BYTE)j)) goto CommError;
            gModemState = MS_SENDPACKET;
          }
        }
        break;
      case MS_INIT1:
        wRetryTimes ++;
        if (wRetryTimes>3) goto CommError;
        if (!SendModemCommand (MODEMCMD_INIT1)) goto CommError;
        break;
      case MS_INIT2:
        wRetryTimes ++;
        if (wRetryTimes>3) goto CommError;
        if (!SendModemCommand (MODEMCMD_INIT2)) goto CommError;
        break;
      case MS_INIT3:
        wRetryTimes ++;
        if (wRetryTimes>3) goto CommError;
        if (!SendModemCommand (MODEMCMD_INIT3)) goto CommError;
        break;
      }
    }

#ifdef LOOPBACKDEBUG
    if (gfSendPacket82)
    {
      abPacket[ 0] = 0x82;       // Command
      abPacket[ 1] = 0xc0;       // Voltage
      abPacket[ 2] = 0xc0;       // Current
      abPacket[ 3] = 0xc0;       // Temperature
      abPacket[ 4] = 0x02;       // Status

      abPacket[ 5] = 0x00;       // Error Code
      abPacket[ 6] = 0x10;       // Channel
      abPacket[ 7] = 0x00;       //
      abPacket[ 8] = 0x31;       // Up Gain
      abPacket[ 9] = 0x30;       // Down Gain
      abPacket[10] = 0xf0;       // Up Power
      abPacket[11] = 0xf1;       // Down Poer
      abPacket[12] = 0xf0;       // Up Actual Poer
      abPacket[13] = 0xf1;       // Down Actual Poer

      abPacket[14] = 0x00;       // Error Code
      abPacket[15] = 0x10;       // Channel
      abPacket[16] = 0x00;       //
      abPacket[17] = 0x21;       // Up Gain
      abPacket[18] = 0x20;       // Down Gain
      abPacket[19] = 0xe0;       // Up Power
      abPacket[20] = 0xe1;       // Down Poer
      abPacket[21] = 0xe0;       // Up Actual Poer
      abPacket[22] = 0xe1;       // Down Actual Poer

      RecvMsgData.pcszPhoneNo   = "13620224364";
      RecvMsgData.pbContent     = abPacket;
      RecvMsgData.dwContentSize = 23;
      SendMessage (ghMainWnd, WM_RECEIVEDSMS, 0, (LPARAM)&RecvMsgData);
      gfSendPacket82 = FALSE;
    }

    if (gfSendPacket83)
    {
      abPacket[ 0] = 0x83;       // Command
      abPacket[ 1] = 0x02;       // Channels
      abPacket[ 2] = 0x03;       // Version
      abPacket[ 3] = 0x20;       // Voltage
      abPacket[ 4] = 0x10;       // Power 1
      abPacket[ 5] = 0x00;       // Power 2
      abPacket[ 6] = 0x30;       // Current
      abPacket[ 7] = 0x48;       // Temperature
      abPacket[ 8] = 0x50;       // 1 Up Gain
      abPacket[ 9] = 0x50;       // 1 Down Gain
      abPacket[10] = 0x50;       // 2 Up Gain
      abPacket[11] = 0x50;       // 2 Down Gain

      RecvMsgData.pcszPhoneNo   = "13620224364";
      RecvMsgData.pbContent     = abPacket;
      RecvMsgData.dwContentSize = 12;
      SendMessage (ghMainWnd, WM_RECEIVEDSMS, 0, (LPARAM)&RecvMsgData);
      gfSendPacket83 = FALSE;
    }
#endif
  }
CommError:
  return 0;
}
/****************************************************************************************/
/****************************************************************************************/
STEC StartCommunicationThread(WORD wPort)
{
  DWORD           dwThreadID;
  if (ghCommunicationThreadHandle) return STEC_ALREADYSTART;
  if (!OpenCommPort (wPort)) return STEC_OPENCOMMPORT;

  // Init Modem
  if (!SendModemCommand (MODEMCMD_INIT1))
  {
    CloseCommPort ();
    return STEC_COMMPROTERR;
  }
  
  gfStopCommunicationThread = FALSE;

  ghCommunicationThreadHandle = CreateThread (NULL, 0, CommunicationThread,
                                              NULL, 0, &dwThreadID);
  if (ghCommunicationThreadHandle)
    return STEC_SUCCESSFUL;
  else
  {
    CloseCommPort ();
    return STEC_STARTTHREAD;
  }
}
/****************************************************************************************/
void StopCommunicationThread(void)
{
  gfStopCommunicationThread = TRUE;
}
/****************************************************************************************/
BOOL IsStopCommunicationThread(void)
{
  DWORD           dwExitCode;
  if (ghCommunicationThreadHandle == NULL) return TRUE;
  GetExitCodeThread (ghCommunicationThreadHandle, &dwExitCode);
  return dwExitCode!=STILL_ACTIVE;
}
/****************************************************************************************/
void ReleaseCommunicationThread(void)
{
  CloseHandle (ghCommunicationThreadHandle);
  ghCommunicationThreadHandle = NULL;
  ghCommDev = INVALID_HANDLE_VALUE;
}
/****************************************************************************************/
void KillCommuicationThread(void)
{
  DWORD           dwTickCount;
  
  StopCommunicationThread ();
  dwTickCount = GetTickCount ();
  while (GetTickCount () - dwTickCount < 5000l)
    if (IsStopCommunicationThread ()) break;
  if (!IsStopCommunicationThread ())
    TerminateThread (ghCommunicationThreadHandle, 0);
  CloseCommPort ();
  ReleaseCommunicationThread ();
}
/****************************************************************************************/
