/*****************************************************************************************
 * File name     : STATWND.C
 * Author        : Han Zhi Qiang
 * Module        : SOCKTOOL
 * Purpose       : The status window
 * Create Date   : May 16, 2001 - 11:26
 * Modify history: 
 *****************************************************************************************
 */

#include <windows.h>
#include <stdio.h>
#include "ui.h"
#include "hiswnd.h"

/*****************************************************************************************
 * MACRO functions declare
 ****************************************************************************************/

/*****************************************************************************************
 * Identifier constant declare
 ****************************************************************************************/
#define WHEELDELTAPERLINE               40

/*****************************************************************************************
 * Data structure declare
 ****************************************************************************************/
typedef struct tagINFOITEM
{
  struct tagINFOITEM *pPrevItem;
  struct tagINFOITEM *pNextItem;
  OMPT                Type;
  PCHAR               pszContent;
  DWORD               dwLineNo;
} INFOITEM;

typedef struct tagSTATWNDINFO
{
  INFOITEM       *pStatusInfo;
  INFOITEM       *pCurStatusInfo;

  HFONT           hFont;
  WORD            wFontHeight;

  WORD            wStatusWndWidth, wStatusWndHeight;
  WORD            wStatusWndTextLines;
  DWORD           dwTotalLine;
  STFLAG          Flags;

  OMPT            LineType;
  BYTE            bCacheSize;
  BYTE            abStreamCache[16];
} STATWNDINFO, *PSTATWNDINFO;

/*****************************************************************************************
 * Local variable declare
 ****************************************************************************************/

/*****************************************************************************************
 * Local functions declare
 ****************************************************************************************/

/****************************************************************************************/
static void SetStatusWndTextPos(HWND hWnd, PSTATWNDINFO pStatusWndInfo, DWORD dwPos)
{
  INFOITEM       *pStatusInfo;
  SCROLLINFO      ScrollInfo;

  pStatusInfo = pStatusWndInfo->pCurStatusInfo;
  if (pStatusInfo == NULL) return;

  while (pStatusInfo->dwLineNo != dwPos)
  {
    if (pStatusInfo->dwLineNo > dwPos)
      pStatusInfo = pStatusInfo->pPrevItem;
    else
      pStatusInfo = pStatusInfo->pNextItem;
  }

  if (pStatusInfo != pStatusWndInfo->pCurStatusInfo)
  {
    pStatusWndInfo->pCurStatusInfo = pStatusInfo;
    ScrollInfo.cbSize = sizeof (SCROLLINFO);
    ScrollInfo.fMask  = SIF_PAGE | SIF_POS | SIF_RANGE;
    ScrollInfo.nMin   = 0;
    ScrollInfo.nMax   = pStatusWndInfo->dwTotalLine - pStatusWndInfo->wStatusWndTextLines;
    ScrollInfo.nPage  = 1;
    ScrollInfo.nPos   = dwPos;
    SetScrollInfo (hWnd, SB_VERT, &ScrollInfo, TRUE);
    InvalidateRect (hWnd, NULL, FALSE);
  }
}
/****************************************************************************************/
LRESULT CALLBACK HISWND_StatusWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  PSTATWNDINFO    pStatusWndInfo;
  WORD            i;
  SHORT           iDelta;
  HDC             hDC;
  PAINTSTRUCT     ps;
  CREATESTRUCT   *pcs;
  INFOITEM       *pStatusInfo;
  HFONT           hOldFont;
  HPEN            hPen, hOldPen;
  HBRUSH          hBrush, hOldBrush;
  RECT            TextRect;
  COLORREF        TextBkColor, TextFgColor;
  SCROLLINFO      ScrollInfo;
  TEXTMETRIC      TextMetric;
  BOOLEAN         fFocusWnd;

  switch (uMsg)
  {
  case WM_PAINT:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    hDC = BeginPaint (hWnd, &ps);
    if (pStatusWndInfo->pCurStatusInfo)
    {
      if (pStatusWndInfo->hFont) hOldFont = SelectObject (hDC, pStatusWndInfo->hFont);
      hPen = CreatePen (PS_SOLID, 0, RGB (192, 192, 192));
      hOldPen = SelectObject (hDC, hPen);
    
      pStatusInfo = pStatusWndInfo->pCurStatusInfo;
      fFocusWnd = (pStatusWndInfo->Flags & STFLAG_ALWAYSFOCUS) || (GetFocus () == hWnd);
      for (i=0; i<pStatusWndInfo->wStatusWndHeight; i+=pStatusWndInfo->wFontHeight+1)
      {
        if (pStatusInfo)
        {
          switch (pStatusInfo->Type)
          {
          case OMPT_SUCCESS:
            TextBkColor = fFocusWnd ? RGB (223, 255, 223) : RGB (200, 200, 200);
            TextFgColor = RGB (0, 0, 0);
            break;
          case OMPT_ERROR:
            TextBkColor = fFocusWnd ? RGB (255, 223, 223) : RGB (200, 200, 200);
            TextFgColor = RGB (0, 0, 0);
            break;
          case OMPT_INFO:
            TextBkColor = fFocusWnd ? RGB (223, 239, 255) : RGB (255, 255, 255);
            TextFgColor = RGB (0, 0, 0);
            break;
          case OMPT_UPLOAD:
            TextBkColor = fFocusWnd ? RGB (233, 255, 255) : RGB (218, 218, 218);
            TextFgColor = RGB (0, 0, 0);
            break;
          case OMPT_DOWNLOAD:
            TextBkColor = fFocusWnd ? RGB (255, 255, 239) : RGB (236, 236, 236);
            TextFgColor = RGB (0, 0, 0);
            break;
          }
          SetTextColor (hDC, TextFgColor);
          SetBkColor (hDC, TextBkColor);
          TextRect.top    = i;
          TextRect.left   = 0;
          TextRect.bottom = i + pStatusWndInfo->wFontHeight;
          TextRect.right  = pStatusWndInfo->wStatusWndWidth;
          DrawText (hDC, pStatusInfo->pszContent, -1, &TextRect, DT_SINGLELINE|DT_NOPREFIX);
          DrawText (hDC, pStatusInfo->pszContent, -1, &TextRect, 
                    DT_SINGLELINE|DT_CALCRECT|DT_NOPREFIX);
          TextRect.left = TextRect.right;
          TextRect.right = pStatusWndInfo->wStatusWndWidth;
          
          hBrush = CreateSolidBrush (TextBkColor);
          hOldBrush = SelectObject (hDC, hBrush);
          FillRect (hDC, &TextRect, hBrush);
          SelectObject (hDC, hOldBrush);
          DeleteObject (hBrush);

          pStatusInfo = pStatusInfo->pNextItem;
          MoveToEx (hDC, 0, i+pStatusWndInfo->wFontHeight, NULL);
          LineTo (hDC, pStatusWndInfo->wStatusWndWidth, i+pStatusWndInfo->wFontHeight);
        }
        else
          break;
      }
      hBrush = CreateSolidBrush (RGB (0xff, 0xff, 0xff));
      hOldBrush = SelectObject (hDC, hBrush);
      TextRect.top    = i;
      TextRect.left   = 0;
      TextRect.bottom = pStatusWndInfo->wStatusWndHeight;
      TextRect.right  = pStatusWndInfo->wStatusWndWidth;
      FillRect (hDC, &TextRect, hBrush);
      SelectObject (hDC, hOldBrush);
      DeleteObject (hBrush);

      SelectObject (hDC, hOldPen);
      DeleteObject (hPen);
      if (pStatusWndInfo->hFont) SelectObject (hDC, hOldFont);
    }
    else
    {
      hBrush = CreateSolidBrush (RGB (0xff, 0xff, 0xff));
      hOldBrush = SelectObject (hDC, hBrush);
      TextRect.top    = 0;
      TextRect.left   = 0;
      TextRect.bottom = pStatusWndInfo->wStatusWndHeight;
      TextRect.right  = pStatusWndInfo->wStatusWndWidth;
      FillRect (hDC, &TextRect, hBrush);
      SelectObject (hDC, hOldBrush);
      DeleteObject (hBrush);
    }
    EndPaint (hWnd, &ps);
    return 0;
  case WM_VSCROLL:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    ScrollInfo.cbSize = sizeof (SCROLLINFO);
    ScrollInfo.fMask  = SIF_POS | SIF_RANGE;
    GetScrollInfo (hWnd, SB_VERT, &ScrollInfo);
    switch (LOWORD(wParam))
    {
    case SB_LINEUP:
      if (ScrollInfo.nPos > ScrollInfo.nMin)
      {
        ScrollInfo.nPos --;
        SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
      }
      break;
    case SB_LINEDOWN:
      if (ScrollInfo.nPos < ScrollInfo.nMax)
      {
        ScrollInfo.nPos ++;
        SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
      }
      break;
    case SB_PAGEUP:
      if (ScrollInfo.nPos > ScrollInfo.nMin)
      {
        ScrollInfo.nPos -= pStatusWndInfo->wStatusWndTextLines;
        if (ScrollInfo.nPos < ScrollInfo.nMin) ScrollInfo.nPos = ScrollInfo.nMin;
        SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
      }
      break;
    case SB_PAGEDOWN:
      if (ScrollInfo.nPos < ScrollInfo.nMax)
      {
        ScrollInfo.nPos += pStatusWndInfo->wStatusWndTextLines;
        if (ScrollInfo.nPos > ScrollInfo.nMax) ScrollInfo.nPos = ScrollInfo.nMax;
        SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
      }
      break;
    case SB_THUMBTRACK:
      SetStatusWndTextPos (hWnd, pStatusWndInfo, HIWORD(wParam));
      break;
    }
    return 0;
  case WM_SIZE:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    pStatusWndInfo->wStatusWndWidth     = LOWORD(lParam);
    pStatusWndInfo->wStatusWndHeight    = HIWORD(lParam);
    if (pStatusWndInfo->pCurStatusInfo)
      i = pStatusWndInfo->dwTotalLine - pStatusWndInfo->wStatusWndTextLines ==
          pStatusWndInfo->pCurStatusInfo->dwLineNo;
    else
      i = 0;
    pStatusWndInfo->wStatusWndTextLines = pStatusWndInfo->wStatusWndHeight /
                                          (pStatusWndInfo->wFontHeight+1);
    if (pStatusWndInfo->wStatusWndTextLines<1) pStatusWndInfo->wStatusWndTextLines = 10;
    if (i)
    {
      if (pStatusWndInfo->dwTotalLine <= pStatusWndInfo->wStatusWndTextLines)
        i = 0;
      else
        i = (WORD)(pStatusWndInfo->dwTotalLine - pStatusWndInfo->wStatusWndTextLines);
      if (i>=pStatusWndInfo->pCurStatusInfo->dwLineNo)
      {
        while (i>pStatusWndInfo->pCurStatusInfo->dwLineNo)
          pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pCurStatusInfo->pNextItem;
      }
      else
      {
        while (i<pStatusWndInfo->pCurStatusInfo->dwLineNo)
          pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pCurStatusInfo->pPrevItem;
      }
    }
    if (pStatusWndInfo->wStatusWndHeight<0) pStatusWndInfo->wStatusWndHeight = 0;
    ScrollInfo.cbSize = sizeof (SCROLLINFO);
    ScrollInfo.fMask  = SIF_PAGE | SIF_POS | SIF_RANGE;
    ScrollInfo.nMin   = 0;
    ScrollInfo.nMax   = pStatusWndInfo->dwTotalLine-pStatusWndInfo->wStatusWndTextLines;
    ScrollInfo.nPage  = 1;
    ScrollInfo.nPos   = pStatusWndInfo->pCurStatusInfo ? 
                          pStatusWndInfo->pCurStatusInfo->dwLineNo : 0;
    SetScrollInfo (hWnd, SB_VERT, &ScrollInfo, TRUE);
    return 0;
  case WM_KEYDOWN:
    switch (wParam)
    {
    case VK_UP:    PostMessage (hWnd, WM_VSCROLL, SB_LINEUP, 0);   break;
    case VK_DOWN:  PostMessage (hWnd, WM_VSCROLL, SB_LINEDOWN, 0); break;
    case VK_PRIOR: PostMessage (hWnd, WM_VSCROLL, SB_PAGEUP, 0);   break;
    case VK_NEXT:  PostMessage (hWnd, WM_VSCROLL, SB_PAGEDOWN, 0); break;
    }
    return 0;
  case WM_MOUSEWHEEL:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    ScrollInfo.cbSize = sizeof (SCROLLINFO);
    ScrollInfo.fMask  = SIF_POS | SIF_RANGE;
    GetScrollInfo (hWnd, SB_VERT, &ScrollInfo);
    iDelta = HIWORD(wParam);

    if (iDelta < 0)
    {
      iDelta /= -WHEELDELTAPERLINE;
      if (iDelta == 0) iDelta = 1;
      ScrollInfo.nPos += iDelta;
      if (ScrollInfo.nPos > ScrollInfo.nMax) ScrollInfo.nPos = ScrollInfo.nMax;
      SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
    }
    else
    {
      iDelta /= WHEELDELTAPERLINE;
      if (iDelta == 0) iDelta = 1;
      ScrollInfo.nPos -= iDelta;
      if (ScrollInfo.nPos < ScrollInfo.nMin) ScrollInfo.nPos = ScrollInfo.nMin;
      SetStatusWndTextPos (hWnd, pStatusWndInfo, ScrollInfo.nPos);
    }
    return 0;
  case WM_CREATE:
    pStatusWndInfo = malloc (sizeof *pStatusWndInfo);
    if (pStatusWndInfo == NULL) return -1;
    SetWindowLong (hWnd, GWL_USERDATA, (LONG)pStatusWndInfo);

    memset (pStatusWndInfo, 0, sizeof *pStatusWndInfo);
    pcs = (CREATESTRUCT *)lParam;
    pStatusWndInfo->Flags = (STFLAG)pcs->lpCreateParams;
    
    pStatusWndInfo->hFont = GetStockObject (OEM_FIXED_FONT);
    if (pStatusWndInfo->hFont)
    {
      hDC = GetDC (hWnd);
      hOldFont = SelectObject (hDC, pStatusWndInfo->hFont);
      GetTextMetrics (hDC, &TextMetric);
      pStatusWndInfo->wFontHeight = (WORD)TextMetric.tmHeight;
      SelectObject (hDC, hOldFont);
      ReleaseDC (hWnd, hDC);
    }
    else
      pStatusWndInfo->wFontHeight = 16;
    break;
  case WM_DESTROY:
    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
    while (pStatusWndInfo->pStatusInfo)
    {
      pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pStatusInfo;
      pStatusWndInfo->pStatusInfo = pStatusWndInfo->pStatusInfo->pNextItem;
      free (pStatusWndInfo->pCurStatusInfo->pszContent);
      free (pStatusWndInfo->pCurStatusInfo);
    }
    if (pStatusWndInfo->hFont) DeleteObject (pStatusWndInfo->hFont);
    free (pStatusWndInfo);
    break;
  case WM_SETFOCUS:
  case WM_KILLFOCUS:
    InvalidateRect (hWnd, NULL, FALSE);
    return 0;
  case WM_GETDLGCODE:
    return DLGC_WANTARROWS;
  case WM_LBUTTONDOWN:
    SetFocus (hWnd);
    return 0;
  }
  return DefWindowProc (hWnd, uMsg, wParam, lParam);
}
/****************************************************************************************/
HWND PASCAL HISWND_CreateWnd(HINSTANCE hInst, HWND hWndParent, STFLAG Flags)
{
  HWND            hWnd;

  hWnd = CreateWindowEx (0, STATUSWNDCLASSNAME, NULL,
                         WS_CHILD|WS_VSCROLL|(Flags&STFLAG_TABSTOP ? WS_TABSTOP : 0),
                         0, 0, 0, 0, hWndParent, NULL, hInst, (LPVOID)Flags);
  if (hWnd == NULL) return NULL;
  ShowWindow (hWnd, SW_SHOW);

  return hWnd;
}
/****************************************************************************************/
DWORD PASCAL HISWND_SetFlag(HWND hWnd, STFLAG Flags)
{
  PSTATWNDINFO    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);

  if (!(pStatusWndInfo->Flags&STFLAG_INSPECTSTREAM) && (Flags&STFLAG_INSPECTSTREAM))
    pStatusWndInfo->bCacheSize = 0;
  pStatusWndInfo->Flags |= Flags;

  return pStatusWndInfo->Flags;
}
/****************************************************************************************/
DWORD PASCAL HISWND_ClearFlag(HWND hWnd, STFLAG Flags)
{
  PSTATWNDINFO    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
  pStatusWndInfo->Flags &= ~Flags;
  return pStatusWndInfo->Flags;
}
/****************************************************************************************/
static void AppendStatusInfoSub(HWND hWnd, PSTATWNDINFO pStatusWndInfo, OMPT Type,
                                PCSTR pcszStatusInfo)
{
  INFOITEM       *pStatusInfo, *pTempInfo;
  PCHAR           pszContent;
  SCROLLINFO      ScrollInfo;

  pStatusInfo = malloc (sizeof (INFOITEM));
  if (pStatusInfo == NULL) return;
  pszContent = malloc (strlen (pcszStatusInfo)+1);
  if (pszContent == NULL)
  {
    free (pStatusInfo);
    return;
  }
  strcpy (pszContent, pcszStatusInfo);

  pStatusInfo->pNextItem  = NULL;
  pStatusInfo->pPrevItem  = NULL;
  pStatusInfo->Type       = Type;
  pStatusInfo->pszContent = pszContent;

  if (pStatusWndInfo->pStatusInfo == NULL)
  {
    pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pStatusInfo = pStatusInfo;
    pStatusInfo->dwLineNo = 0;
    pStatusWndInfo->dwTotalLine = 1;
  }
  else
  {
    pTempInfo = pStatusWndInfo->pCurStatusInfo;
    while (pTempInfo->pNextItem) pTempInfo = pTempInfo->pNextItem;
    pTempInfo->pNextItem = pStatusInfo;
    pStatusInfo->pPrevItem = pTempInfo;
    pStatusInfo->dwLineNo = pTempInfo->dwLineNo + 1;
    pStatusWndInfo->dwTotalLine ++;
  }

  if (pStatusInfo->dwLineNo - pStatusWndInfo->pCurStatusInfo->dwLineNo ==
      pStatusWndInfo->wStatusWndTextLines)
    pStatusWndInfo->pCurStatusInfo = pStatusWndInfo->pCurStatusInfo->pNextItem;
  ScrollInfo.cbSize = sizeof (SCROLLINFO);
  ScrollInfo.fMask  = SIF_PAGE | SIF_POS | SIF_RANGE;
  ScrollInfo.nMin   = 0;
  ScrollInfo.nMax   = pStatusWndInfo->dwTotalLine - pStatusWndInfo->wStatusWndTextLines;
  ScrollInfo.nPage  = 1;
  ScrollInfo.nPos   = pStatusWndInfo->pCurStatusInfo->dwLineNo;
  SetScrollInfo (hWnd, SB_VERT, &ScrollInfo, TRUE);
}
/****************************************************************************************/
WORD PASCAL HISWND_AppendInfoV(HWND hWnd, OMPT Type, PCSTR pcszFormatString,
                                            va_list ArgList)
{
  PSTATWNDINFO    pStatusWndInfo = (PSTATWNDINFO)GetWindowLong (hWnd, GWL_USERDATA);
  CHAR            aszBuf[1024];
  DWORD           i, j;

  if (pStatusWndInfo->Flags & STFLAG_INSPECTSTREAM) return 0;

  if (pcszFormatString)
  {
    vsprintf (aszBuf, pcszFormatString, ArgList);
    if (pStatusWndInfo->Flags & STFLAG_EXPANDNEWLINE)
    {
      for (i=0, j=0; aszBuf[j]; j++)
      {
        if (aszBuf[j] == '\n')
        {
          if (j>i && aszBuf[j] == '\r') aszBuf[j-1] = '\0'; else aszBuf[j] = '\0';
          AppendStatusInfoSub (hWnd, pStatusWndInfo, Type, aszBuf+i);
          i = j + 1;
        }
      }
      if (i < j)
        AppendStatusInfoSub (hWnd, pStatusWndInfo, Type, aszBuf+i);
    }
    else
      AppendStatusInfoSub (hWnd, pStatusWndInfo, Type, aszBuf);
  }
  InvalidateRect (hWnd, NULL, FALSE);
  return 0;
}
/****************************************************************************************/
WORD HISWND_AppendInfo(HWND hWnd, OMPT Type, PCSTR pcszFormatString, ...)
{
  WORD            wResult;
  va_list         ArgList;
  
  va_start (ArgList, pcszFormatString);   
  wResult = HISWND_AppendInfoV(hWnd, Type, pcszFormatString, ArgList);
  va_end (ArgList);

  return wResult;
}
/****************************************************************************************/
